Type: Package
Package: bigD
Title: Flexibly Format Dates and Times to a Given Locale
Version: 0.3.1
Description: Format dates and times flexibly and to whichever locales
    make sense. Parses dates, times, and date-times in various formats
    (including string-based ISO 8601 constructions). The formatting syntax gives
    the user many options for formatting the date and time output in a precise
    manner. Time zones in the input can be expressed in multiple ways and there
    are many options for formatting time zones in the output as well. Several of
    the provided helper functions allow for automatic generation of locale-aware
    formatting patterns based on date/time skeleton formats and standardized
    date/time formats with varying specificity.
Authors@R: c(
    person("Richard", "Iannone", , "rich@posit.co", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-3925-190X")),
    person("Olivier", "Roy", role = "ctb"),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
License: MIT + file LICENSE
URL: https://rstudio.github.io/bigD/, https://github.com/rstudio/bigD
BugReports: https://github.com/rstudio/bigD/issues
Encoding: UTF-8
RoxygenNote: 7.3.2
Depends: R (>= 3.6.0)
Suggests: testthat (>= 3.0.0), vctrs (>= 0.5.0)
Config/testthat/edition: 3
Config/testthat/parallel: true
NeedsCompilation: no
Packaged: 2025-04-03 13:52:40 UTC; riannone
Author: Richard Iannone [aut, cre] (<https://orcid.org/0000-0003-3925-190X>),
  Olivier Roy [ctb],
  Posit Software, PBC [cph, fnd]
Maintainer: Richard Iannone <rich@posit.co>
Repository: CRAN
Date/Publication: 2025-04-03 14:20:02 UTC
Built: R 4.6.0; ; 2026-04-25 15:59:07 UTC; unix

# Generated by roxygen2: do not edit by hand

export(fdt)
export(fdt_locales_lst)
export(fdt_locales_vec)
export(first_day_of_week)
export(flex_d_lst)
export(flex_d_vec)
export(flex_t12_lst)
export(flex_t12_vec)
export(flex_t24_lst)
export(flex_t24_vec)
export(names_months)
export(names_wkdays)
export(standard_date)
export(standard_date_time)
export(standard_time)

get_date_pattern <- function() {
  "([\\+-]?\\d{4}-\\d\\d-\\d\\d)"
}

get_time_pattern <- function() {
  "((T| )((\\d|\\d\\d):\\d\\d[0-9:.]*)|^((\\d|\\d\\d):\\d\\d[0-9:.]*))"
}

get_tz_pattern <- function() {
  "(Z|(\\+|-)\\d\\d|(\\+|-)\\d\\d:\\d\\d|(\\+|-)\\d\\d\\d\\d)"
}

get_tz_pattern_z <- function() {
  paste0(get_time_pattern(), "(Z)")
}

get_tz_pattern_hh <- function() {
  paste0(get_time_pattern(), "((\\+|-)\\d\\d)")
}

get_tz_pattern_hh_mm <- function() {
  paste0(get_time_pattern(), "((\\+|-)\\d\\d:\\d\\d)")
}

get_tz_pattern_hhmm <- function() {
  paste0(get_time_pattern(), "((\\+|-)\\d\\d\\d\\d)")
}

get_wrapped_iana_pattern <- function() {
  "(\\[|\\()([a-zA-Z0-9\\/ -\\+_]*?)(?:\\]|\\))"
}

get_attached_iana_pattern <- function() {
  "(\\s-\\s|\\s|\\/|\\|)([a-zA-Z0-9\\/ -\\+_]*?)"
}

is_date_present <- function(input) {
  grepl(get_date_pattern(), input)
}

is_time_present <- function(input) {
  grepl(get_time_pattern(), input)
}

# returns a vector of length of input
is_tz_present <- function(input) {
  regex <- paste(
    get_tz_pattern_z(),
    get_tz_pattern_hh(),
    get_tz_pattern_hh_mm(),
    get_tz_pattern_hhmm(),
    sep = "|"
  )
  grepl(regex, input)
}

is_long_tzid_present <- function(input) {

  if (!grepl("^.*\\([^']*\\)$", input)) {
    return(FALSE)
  }

  long_tzid <- extract_long_tzid(input = input)

  validate_long_tzid(long_tzid = long_tzid)

  TRUE
}

get_long_tzid_str <- function(input) {

  if (!is_long_tzid_present(input = input)) {
    return(NA_character_)
  }

  long_tzid <- extract_long_tzid(input = input)

  normalize_long_tzid(long_tzid = long_tzid)
}

extract_long_tzid <- function(input) {
  gsub("^.*\\((.*)\\)$", "\\1", input)
}

validate_long_tzid <- function(long_tzid) {

  known_long_tzid_vals <-
    c(unique(tz_name_resolution$tz_canonical), tz_name_resolution$tz_alt)

  if (!(long_tzid %in% known_long_tzid_vals)) {
    stop("The long time zone provided is not valid.", call. = FALSE)
  }

  invisible(TRUE)
}

normalize_long_tzid <- function(long_tzid) {

  if (long_tzid %in% tz_name_resolution$tz_alt) {

    long_tzid <-
      tz_name_resolution$tz_canonical[tz_name_resolution$tz_alt == long_tzid]
  }

  long_tzid
}

# A long tzid can be provided in parentheses as `(America/Vancouver)` at the
# end of a datetime string; the `get_long_tzid_str()` function will remove the
# parentheses (and should only be invoked if the datetime input contains a
# long tzid)
# The `long_tzid_to_tz_str` will convert a long tzid to a valid tz string in
# the (+/-)hhmm form
long_tzid_to_tz_str <- function(long_tzid, input_dt) {

  if (startsWith(long_tzid, "Etc/") && long_tzid %in% c("Etc/GMT", "Etc/UTC")) {
    return("+0000")
  }

  tzdb_entries_tzid <- tzdb[
    tzdb$zone_name == long_tzid, c("date_start", "gmt_offset_h")]

  if (nrow(tzdb_entries_tzid) == 0L) {
    return(NULL)
  }

  input_date <- as.Date(input_dt)

  tzdb_idx <- rle(tzdb_entries_tzid$date_start >= input_date)$lengths[1]

  tz_offset <- tzdb_entries_tzid[[tzdb_idx, "gmt_offset_h"]]

  minutes <- formatC(round((abs(tz_offset) %% 1) * 60, 0), width = 2, flag = "0")
  hours <- formatC(trunc(abs(tz_offset)), width = 2, flag = "0")
  sign <- ifelse(tz_offset < 0, "-", "+")

  paste0(sign, hours, minutes)
}

get_tz_str <- function(input) {

  if (is.null(input) || !is_tz_present(input = input)) {
    return("")
  }

  if (grepl("+", input, fixed = TRUE)) {

    out <- unlist(strsplit(input, split = "+", fixed = TRUE))[[2]]
    out <- paste0("+", out)

  } else if (grepl("Z", input, fixed = TRUE)) {

    out <- "Z"

  } else {

    out <- unlist(strsplit(input, split = "-", fixed = TRUE))
    out <- out[length(out)]
    out <- paste0("-", out)
  }

  # If there is an attached tzid string then remove it
  if (is_long_tzid_present(input = out)) {
    out <- gsub("\\s*\\([^']*\\)$", "", out)
  }

  out
}

get_tz_offset_val_from_tz_str <- function(tz_str) {

  if (tz_str == "Z") {

    offset_val <- 0

  } else {

    offset_val <-
      as.numeric(substr(tz_str, 2, 3)) +
      as.numeric(substr(tz_str, nchar(tz_str) - 1, nchar(tz_str))) / 60

    if (grepl("-", tz_str, fixed = TRUE)) {
      offset_val <- offset_val * (-1)
    }
  }

  offset_val
}

get_tz_offset_val <- function(input, tz_str = NULL) {

  tz_str <- tz_str %||% get_tz_str(input = input)

  get_tz_offset_val_from_tz_str(tz_str = tz_str)
}

strip_tz <- function(input) {

  tz_str <- get_tz_str(input = input)

  out <- gsub(tz_str, "", input, fixed = TRUE)
  out <- gsub("\\s*\\([^']*\\)$", "", out)

  out
}

strip_long_tzid <- function(input) {

  long_tzid <- get_long_tzid_str(input = input)

  input_str <- gsub(paste0("(", long_tzid, ")"), "", input, fixed = TRUE)
  input_str <- strip_surrounding_whitespace(input = input_str)

  input_str
}

is_iana_present <- function(input) {

  grepl(
    paste0(
      get_time_pattern(), get_tz_pattern(),
      "(", get_wrapped_iana_pattern(), "|", get_attached_iana_pattern(), ")"),
    input
  )
}

which_tz_pattern <- function(input) {

  if (grepl(get_tz_pattern_z(), input)) {
    return("z")
  }

  if (grepl(get_tz_pattern_hh_mm(), input)) {
    return("hh_mm")
  }

  if (grepl(get_tz_pattern_hhmm(), input)) {
    return("hhmm")
  }

  if (grepl(get_tz_pattern_hh(), input)) {
    return("hh")
  }

  NA
}

which_iana_pattern <- function(input) {

  if (grepl(paste0(get_tz_pattern(), get_wrapped_iana_pattern()), input)) {
    return("wrapped")
  } else if (grepl(paste0(get_tz_pattern(), get_attached_iana_pattern()), input)) {
    return("attached")
  }
  NA_character_
}

get_tz_offset <- function(input) {

  tz_pattern <- which_tz_pattern(input)

  switch(
    tz_pattern,
    z = {
      tz_offset <- 0.0
    },
    hh = {
      iso_tz_component <- gsub(paste0(".*", get_tz_pattern_hh(), ".*"), "\\4", input)
      tz_offset <- as.numeric(iso_tz_component)
    },
    hh_mm = {
      iso_tz_component <- gsub(paste0(".*", get_tz_pattern_hh_mm(), ".*"), "\\4", input)
      offset_sign <- ifelse(startsWith(iso_tz_component, "-"), -1L, 1L)
      offset_h <- as.numeric(substr(iso_tz_component, 1, 3))
      offset_min <- as.numeric(substr_right(iso_tz_component, 2)) / 60.0
      tz_offset <- (abs(offset_h) + offset_min) * offset_sign
    },
    hhmm = {
      iso_tz_component <- gsub(paste0(".*", get_tz_pattern_hhmm(), ".*"), "\\4", input)
      offset_sign <- ifelse(startsWith(iso_tz_component, "-"), -1L, 1L)
      offset_h <- as.numeric(substr(iso_tz_component, 1, 3))
      offset_min <- as.numeric(substr_right(iso_tz_component, 2)) / 60.0
      tz_offset <- (abs(offset_h) + offset_min) * offset_sign
    })

  tz_offset
}

get_exemplar_city <- function(long_tzid) {

  exemplar_city <- unlist(strsplit(long_tzid, "/", fixed = TRUE)[[1]])[2]

  exemplar_city
}

get_localized_exemplar_city <- function(
    long_tzid,
    locale,
    yield_unknown = TRUE
) {

  exemplar_city <- unlist(strsplit(long_tzid, "/", fixed = TRUE)[[1]])[2]

  # TODO: Resolve links of exemplar cities to a canonical exemplar city
  #       This will require a separate lookup table

  if (!(exemplar_city %in% colnames(tz_exemplar)[-1])) {

    # Get localized variant of 'Unknown City'
    if (yield_unknown) {
      return(tz_exemplar[tz_exemplar$locale == locale, "Unknown"])
    } else {
      return(NA_character_)
    }
  }

  exemplar_city_localized <-
    tz_exemplar[tz_exemplar$locale == locale, exemplar_city]

  exemplar_city_localized
}

# The short specific non-location format (e.g., 'PST') from a `long_tzid`
get_tz_short_specific <- function(long_tzid, input_dt) {

  tzdb_entries_tzid <- tzdb[tzdb$zone_name == long_tzid, ]
  if (nrow(tzdb_entries_tzid) == 0) {
    return(NA_character_)
  }

  input_date <- as.Date(input_dt)

  tzdb_idx <- rle(tzdb_entries_tzid$date_start >= input_date)$lengths[1]

  # TODO: add check to ensure that the `abbrev` value is a valid
  # short specific non-location time zone
  tzdb_entries_tzid[tzdb_idx, "abbrev"]
}

# The long specific non-location format (e.g., 'Pacific Standard Time') from
# a `long_tzid`
get_tz_long_specific <- function(long_tzid, input_dt, locale) {

  input_date <- as.Date(input_dt)

  tzdb_entries_tzid <- tzdb[tzdb$zone_name == long_tzid, ]

  if (nrow(tzdb_entries_tzid) == 0L) {
    return(NA_character_)
  }

  tzdb_idx <- rle(tzdb_entries_tzid$date_start >= input_date)$lengths[1]

  tzdb_entries_tzid_ln <- tzdb_entries_tzid[tzdb_idx, ]

  if (tzdb_entries_tzid_ln$dst) {
    pattern_col_tz_formats <- "region_format_daylight"
  } else {
    pattern_col_tz_formats <- "region_format"
  }

  tz_long_specific_pattern <-
    tz_formats[tz_formats$locale == locale, ][[pattern_col_tz_formats]]

  # Get the metazone in its long ID format
  metazone_long_id <- long_tz_id_to_metazone_long_id(long_tzid = long_tzid)

  if (is.na(metazone_long_id)) {
    return(NA_character_)
  }

  # Get the row of the `tz_metazone_names` table based on the supplied locale
  tz_metazone_names_row <- tz_metazone_names[tz_metazone_names$locale == locale, ]

  if (nrow(tz_metazone_names_row) == 0) {
    return(NA_character_)
  }

  # Get the list entry corresponding to the metazone and the locale
  tz_metazone_names_entry <-
    unlist(
      tz_metazone_names_row[, colnames(tz_metazone_names_row) == metazone_long_id][[1]]
    )

  if (tzdb_entries_tzid_ln$dst) {
    daylight_standard <- "daylight"
  } else {
    daylight_standard <- "standard"
  }

  tz_metazone_names_filtered <-
    tz_metazone_names_entry[
      grepl(daylight_standard, names(tz_metazone_names_entry))]

  tz_metazone_name <-
    unname(tz_metazone_names_filtered[
      grepl("long", names(tz_metazone_names_filtered), fixed = TRUE)])

  tz_metazone_name
}

get_tz_bcp_id <- function(long_tzid) {

  # If the supplied `long_tzid` value is NA, return NA
  if (is.na(long_tzid)) {
    return(NA_character_)
  }

  tz_name <- tz_bcp_id[tz_bcp_id$tz_canonical == long_tzid, ][["tz_bcp_id"]]

  tz_name
}

# Get the non-location formatted time zone names (e.g., 'Pacific Time', 'PT')
# from a `long_tzid` (canonical tz name)
#
# The non-location format reflects "wall time" (what is on
# a clock on the wall). It's used for recurring events, meetings, or anywhere
# people do not want to be overly specific. For example, "10 am Pacific Time"
# will be GMT-8 in the winter, and GMT-7 in the summer.
get_tz_non_location <- function(
    long_tzid,
    locale,
    short_long,
    type
) {

  short_long <- match.arg(short_long, c("long", "short"))
  type <- match.arg(type, c("generic", "standard", "daylight"))

  # If the supplied `long_tzid` value is NA, return NA
  if (is.na(long_tzid)) {
    return(NA_character_)
  }

  # Get the metazone in its long ID format
  metazone_long_id <- long_tz_id_to_metazone_long_id(long_tzid = long_tzid)

  # Check if metazone is NA and return NA if that is so
  if (is.na(metazone_long_id)) {
    return(NA_character_)
  }

  # Get the row of the `tz_metazone_names` table based on the supplied locale
  tz_metazone_names_row <- tz_metazone_names[tz_metazone_names$locale == locale, ]

  # Get the list entry corresponding to the metazone and the locale
  tz_metazone_names_entry <-
    unlist(
      tz_metazone_names_row[, colnames(tz_metazone_names_row) == metazone_long_id][[1]]
    )

  target_item <- paste0(short_long, ".", type)
  available_items <- names(tz_metazone_names_entry)
  has_long_items <- any(grepl("long", names(tz_metazone_names_entry), fixed = TRUE))
  has_short_items <- any(grepl("short", names(tz_metazone_names_entry), fixed = TRUE))

  if (length(available_items) == 1) {
    tz_name <- tz_metazone_names_entry[[available_items]]
  } else if (target_item %in% available_items) {
    tz_name <- tz_metazone_names_entry[[target_item]]
  } else if (!has_short_items && short_long == "short") {
    if (any(grepl(type, available_items))) {
      tz_name <- tz_metazone_names_entry[[paste0("long.", type)]]
    } else {
      tz_name <- tz_metazone_names_entry[["long.standard"]]
    }
  } else {
    tz_name <- tz_metazone_names_entry[[available_items[1]]]
  }

  tz_name
}

long_tz_id_to_metazone_long_id <- function(long_tzid) {

  tzid_in_tz_metazone_users <- long_tzid %in% tz_metazone_users$canonical_tz_name

  if (!tzid_in_tz_metazone_users) {

    if (long_tzid %in% unique(tz_name_resolution$tz_canonical)) {

      alt_names <-
        tz_name_resolution$tz_alt[tz_name_resolution$tz_canonical == long_tzid]

      if (!any(alt_names %in% tz_metazone_users$canonical_tz_name)) {
        return(NA_character_)
      }
      long_tzid <- alt_names[1]

    } else {
      return(NA_character_)
    }
  }

  rows <- which(tz_metazone_users$canonical_tz_name == long_tzid)
  tz_metazone_users_rows <-
    tz_metazone_users[[rows, "metazone_long_id"]]

  # Return NA if number of rows in `tz_metazone_users_rows` is zero
  if (length(tz_metazone_users_rows) == 0) {
    return(NA_character_)
  }

  # TODO: develop routine to further filter multirow `tz_metazone_users_rows`
  # to a single row based on `locale`; for now, obtain the first metazone
  metazone <- tz_metazone_users_rows[1]

  metazone
}

get_long_local_gmt <- function(tz_offset) {

  minutes <- formatC(round((abs(tz_offset) %% 1) * 60, 0), width = 2, flag = "0")
  hours <- formatC(trunc(abs(tz_offset)), width = 2, flag = "0")
  sign <- ifelse(tz_offset < 0, "-", "+")

  paste0("GMT", sign, hours, ":", minutes)
}

get_short_local_gmt <- function(tz_offset) {

  minutes <- formatC(round((abs(tz_offset) %% 1) * 60, 0), width = 2, flag = "0")

  if (minutes == "00") {
    minutes <- ""
  } else {
    minutes <- paste0(":", minutes)
  }

  hours <- as.character(trunc(abs(tz_offset)))
  sign <- ifelse(tz_offset < 0, "-", "+")

  paste0("GMT", sign, hours, minutes)
}

get_date_component <- function(input) {

  if (!is_date_present(input)) {
    return(NA_character_)
  }

  gsub(paste0(".*", get_date_pattern(), ".*"), "\\1", input)
}

get_time_component <- function(input) {

  if (!is_time_present(input)) {
    return(NA_character_)
  }

  gsub(paste0(".*", get_time_pattern(), ".*"), "\\2", input)
}

get_iana_tz <- function(input) {

  iana_pattern <- which_iana_pattern(input)

  value <- switch(
    iana_pattern,
    wrapped = {
      tz_name <-
        gsub(
          paste0(".*", get_time_pattern(), get_tz_pattern()),
          "", input)
      gsub("(\\(|\\[|\\)|\\])", "", tz_name)
    },
    attached = gsub(
      paste0(".*", get_time_pattern(), get_tz_pattern(), "(\\s-\\s|\\s|\\/|\\|)"),
      "",
      input
    ),
    # default
    NA_character_
   )

  tz_name
}

strip_surrounding_whitespace <- function(input) {
  gsub("(^[[:space:]]*)|([[:space:]]*$)", "", input)
}

#' Get a vector of all supported locales
#'
#' @description
#' The `fdt_locales_vec()` function produces a vector of all supported locale
#' IDs in the **bigD** package.
#'
#' @return
#' A character vector of supported locale IDs.
#'
#' @examples
#' # Let's get all the `ar` locales that exist
#' # in the vector produced by `fdt_locales_vec()`
#' grep("^ar", fdt_locales_vec(), value = TRUE)
#'
#' # Let's get all the locales that pertain to the
#' # `CH` territory in the vector produced by
#' # `fdt_locales_vec()`
#' grep("CH", fdt_locales_vec(), value = TRUE)
#'
#' @export
fdt_locales_vec <- function() {
  c(
    "af", "af_NA", "agq", "ak", "am", "ar", "ar_AE", "ar_BH", "ar_DJ",
    "ar_DZ", "ar_EG", "ar_EH", "ar_ER", "ar_IL", "ar_IQ", "ar_JO",
    "ar_KM", "ar_KW", "ar_LB", "ar_LY", "ar_MA", "ar_MR", "ar_OM",
    "ar_PS", "ar_QA", "ar_SA", "ar_SD", "ar_SO", "ar_SS", "ar_SY",
    "ar_TD", "ar_TN", "ar_YE", "as", "asa", "ast", "az", "az_Cyrl",
    "az_Latn", "bas", "be", "be_tarask", "bem", "bez", "bg", "bm",
    "bn", "bn_IN", "bo", "bo_IN", "br", "brx", "bs", "bs_Cyrl", "bs_Latn",
    "ca", "ca_AD", "ca_ES_valencia", "ca_FR", "ca_IT", "ccp", "ccp_IN",
    "ce", "ceb", "cgg", "chr", "ckb", "ckb_IR", "cs", "cy", "da",
    "da_GL", "dav", "de", "de_AT", "de_BE", "de_CH", "de_IT", "de_LI",
    "de_LU", "dje", "doi", "dsb", "dua", "dyo", "dz", "ebu", "ee",
    "ee_TG", "el", "el_CY", "en", "en_001", "en_150", "en_AE", "en_AG",
    "en_AI", "en_AS", "en_AT", "en_AU", "en_BB", "en_BE", "en_BI",
    "en_BM", "en_BS", "en_BW", "en_BZ", "en_CA", "en_CC", "en_CH",
    "en_CK", "en_CM", "en_CX", "en_CY", "en_DE", "en_DG", "en_DK",
    "en_DM", "en_ER", "en_FI", "en_FJ", "en_FK", "en_FM", "en_GB",
    "en_GD", "en_GG", "en_GH", "en_GI", "en_GM", "en_GU", "en_GY",
    "en_HK", "en_IE", "en_IL", "en_IM", "en_IN", "en_IO", "en_JE",
    "en_JM", "en_KE", "en_KI", "en_KN", "en_KY", "en_LC", "en_LR",
    "en_LS", "en_MG", "en_MH", "en_MO", "en_MP", "en_MS", "en_MT",
    "en_MU", "en_MV", "en_MW", "en_MY", "en_NA", "en_NF", "en_NG",
    "en_NL", "en_NR", "en_NU", "en_NZ", "en_PG", "en_PH", "en_PK",
    "en_PN", "en_PR", "en_PW", "en_RW", "en_SB", "en_SC", "en_SD",
    "en_SE", "en_SG", "en_SH", "en_SI", "en_SL", "en_SS", "en_SX",
    "en_SZ", "en_TC", "en_TK", "en_TO", "en_TT", "en_TV", "en_TZ",
    "en_UG", "en_UM", "en_VC", "en_VG", "en_VI", "en_VU", "en_WS",
    "en_ZA", "en_ZM", "en_ZW", "eo", "es", "es_419", "es_AR", "es_BO",
    "es_BR", "es_BZ", "es_CL", "es_CO", "es_CR", "es_CU", "es_DO",
    "es_EA", "es_EC", "es_GQ", "es_GT", "es_HN", "es_IC", "es_MX",
    "es_NI", "es_PA", "es_PE", "es_PH", "es_PR", "es_PY", "es_SV",
    "es_US", "es_UY", "es_VE", "et", "eu", "ewo", "fa", "fa_AF",
    "ff", "ff_Adlm", "ff_Adlm_BF", "ff_Adlm_CM", "ff_Adlm_GH", "ff_Adlm_GM",
    "ff_Adlm_GW", "ff_Adlm_LR", "ff_Adlm_MR", "ff_Adlm_NE", "ff_Adlm_NG",
    "ff_Adlm_SL", "ff_Adlm_SN", "ff_Latn", "ff_Latn_BF", "ff_Latn_CM",
    "ff_Latn_GH", "ff_Latn_GM", "ff_Latn_GN", "ff_Latn_GW", "ff_Latn_LR",
    "ff_Latn_MR", "ff_Latn_NE", "ff_Latn_NG", "ff_Latn_SL", "fi",
    "fil", "fo", "fo_DK", "fr", "fr_BE", "fr_BF", "fr_BI", "fr_BJ",
    "fr_BL", "fr_CA", "fr_CD", "fr_CF", "fr_CG", "fr_CH", "fr_CI",
    "fr_CM", "fr_DJ", "fr_DZ", "fr_GA", "fr_GF", "fr_GN", "fr_GP",
    "fr_GQ", "fr_HT", "fr_KM", "fr_LU", "fr_MA", "fr_MC", "fr_MF",
    "fr_MG", "fr_ML", "fr_MQ", "fr_MR", "fr_MU", "fr_NC", "fr_NE",
    "fr_PF", "fr_PM", "fr_RE", "fr_RW", "fr_SC", "fr_SN", "fr_SY",
    "fr_TD", "fr_TG", "fr_TN", "fr_VU", "fr_WF", "fr_YT", "fur",
    "fy", "ga", "ga_GB", "gd", "gl", "gsw", "gsw_FR", "gsw_LI", "gu",
    "guz", "gv", "ha", "ha_GH", "ha_NE", "haw", "he", "hi", "hi_Latn",
    "hr", "hr_BA", "hsb", "hu", "hy", "ia", "id", "ig", "ii", "is",
    "it", "it_CH", "it_SM", "it_VA", "ja", "jgo", "jmc", "jv", "ka",
    "kab", "kam", "kde", "kea", "kgp", "khq", "ki", "kk", "kkj",
    "kl", "kln", "km", "kn", "ko", "ko_KP", "kok", "ks", "ks_Arab",
    "ks_Deva", "ksb", "ksf", "ksh", "ku", "kw", "ky", "lag", "lb",
    "lg", "lkt", "ln", "ln_AO", "ln_CF", "ln_CG", "lo", "lrc", "lrc_IQ",
    "lt", "lu", "luo", "luy", "lv", "mai", "mas", "mas_TZ", "mer",
    "mfe", "mg", "mgh", "mgo", "mi", "mk", "ml", "mn", "mni", "mni_Beng",
    "mr", "ms", "ms_BN", "ms_ID", "ms_SG", "mt", "mua", "my", "mzn",
    "naq", "nb", "nb_SJ", "nd", "nds", "nds_NL", "ne", "ne_IN", "nl",
    "nl_AW", "nl_BE", "nl_BQ", "nl_CW", "nl_SR", "nl_SX", "nmg",
    "nn", "nnh", "no", "nus", "nyn", "om", "om_KE", "or", "os", "os_RU",
    "pa", "pa_Arab", "pa_Guru", "pcm", "pl", "ps", "ps_PK", "pt",
    "pt_AO", "pt_CH", "pt_CV", "pt_GQ", "pt_GW", "pt_LU", "pt_MO",
    "pt_MZ", "pt_PT", "pt_ST", "pt_TL", "qu", "qu_BO", "qu_EC", "rm",
    "rn", "ro", "ro_MD", "rof", "ru", "ru_BY", "ru_KG", "ru_KZ",
    "ru_MD", "ru_UA", "rw", "rwk", "sa", "sah", "saq", "sat", "sat_Olck",
    "sbp", "sc", "sd", "sd_Arab", "sd_Deva", "se", "se_FI", "se_SE",
    "seh", "ses", "sg", "shi", "shi_Latn", "shi_Tfng", "si", "sk",
    "sl", "smn", "sn", "so", "so_DJ", "so_ET", "so_KE", "sq", "sq_MK",
    "sq_XK", "sr", "sr_Cyrl", "sr_Cyrl_BA", "sr_Cyrl_ME", "sr_Cyrl_XK",
    "sr_Latn", "sr_Latn_BA", "sr_Latn_ME", "sr_Latn_XK", "su", "su_Latn",
    "sv", "sv_AX", "sv_FI", "sw", "sw_CD", "sw_KE", "sw_UG", "ta",
    "ta_LK", "ta_MY", "ta_SG", "te", "teo", "teo_KE", "tg", "th",
    "ti", "ti_ER", "tk", "to", "tr", "tr_CY", "tt", "twq", "tzm",
    "ug", "uk", "und", "ur", "ur_IN", "uz", "uz_Arab", "uz_Cyrl",
    "uz_Latn", "vai", "vai_Latn", "vai_Vaii", "vi", "vun", "wae",
    "wo", "xh", "xog", "yav", "yi", "yo", "yo_BJ", "yrl", "yrl_CO",
    "yrl_VE", "yue", "yue_Hans", "yue_Hant", "zgh", "zh", "zh_Hans",
    "zh_Hans_HK", "zh_Hans_MO", "zh_Hans_SG", "zh_Hant", "zh_Hant_HK",
    "zh_Hant_MO", "zu"
  )
}

#' A list of all supported locales
#'
#' @description
#' The `fdt_locales_lst` object is a list of all supported locales. This is
#' useful when used with the [fdt()] function as the list can be auto-completed
#' with a locale identifier and this generates valid input for the `locale`
#' argument.
#'
#' @return A list where each element corresponds to a supported locale ID.
#'
#' @section Examples:
#'
#' The `fdt_locales_lst` object can be incredibly useful when choosing one of
#' supported locales. You can avoid typing errors and every element of the list
#' is meant to work. In this example, we'll use the `"da"` locale through
#' use of the list.
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05",
#'   format = "yy-MMMM-d",
#'   locale = fdt_locales_lst$da
#' )
#' ```
#' ```
#' #> [1] "18-juli-4"
#' ````
#'
#' @export
fdt_locales_lst <-
  stats::setNames(as.list(fdt_locales_vec()), as.list(fdt_locales_vec()))

test_that("A vector can be obtained from the `fdt_locales_vec()` function", {

  expect_vector(fdt_locales_vec())
  expect_type(fdt_locales_vec(), "character")
  expect_length(fdt_locales_vec(), 574)
})

test_that("The `fdt_locales_lst` object has a particular structure", {

  expect_type(fdt_locales_lst, "list")
  expect_named(fdt_locales_lst, fdt_locales_vec())
  expect_equal(unlist(fdt_locales_lst, use.names = FALSE), fdt_locales_vec())
})

test_that("The `fdt_locales_lst` object can be used with `fdt()`", {

  for (i in seq_along(fdt_locales_vec())) {

      fmt_val <-
        fdt(
          input = "2018-07-04 22:05",
          locale = fdt_locales_lst[[i]]
        )

      expect_vector(fmt_val)
      expect_length(fmt_val, 1)
      expect_type(fmt_val, "character")
  }
})

#' Get a vector of all flexible date types
#'
#' @description
#' The `flex_d_vec()` function produces a vector of all supported flexible date
#' types in the **bigD** package. These types are essentially identifiers for
#' classes of cross-locale date formatting, so, none of these should be used
#' directly in the `format` argument of the `fdt()` function (use the
#' [flex_d_lst] object for that).
#'
#' @return
#' A character vector of supported flexible date types.
#'
#' @export
flex_d_vec <- function() {

  c(
    "yMd",
    "yMEd",
    "yMMM",
    "yMMMM",
    "yMMMd",
    "yMMMEd",
    "GyMd",
    "GyMMMd",
    "GyMMMEd",
    "yM",
    "Md",
    "MEd",
    "MMMd",
    "MMMEd",
    "MMMMd",
    "GyMMM",
    "yQQQ",
    "yQQQQ",
    "Gy",
    "y",
    "M",
    "MMM",
    "d",
    "Ed",
    #"E",    # TODO: need to determine local day of week for this to work
    "MMMMW", # as "MMMMW-count-*" `one`, `other`, `few`, `many`, `two`, `zero`
    "yw"     # as "yw-*"          `one`, `other`, `few`, `many`, `two`, `zero`
  )
}

#' Get a vector of all 12-hour flexible time types
#'
#' @description
#' The `flex_t12_vec()` function produces a vector of all supported flexible
#' 12-hour time types in the **bigD** package. These types are essentially
#' identifiers for classes of cross-locale time formatting, so, none of these
#' should be used directly in the `format` argument of the `fdt()` function (use
#' the [flex_t12_lst] object for that).
#'
#' @return
#' A character vector of supported 12-hour flexible time types.
#'
#' @export
flex_t12_vec <- function() {

  c(
    "hms",
    "hm",
    "h",
    "Ehm",
    "Ehms",
    "EBhms",
    "Bhms",
    "EBhm",
    "Bhm",
    "Bh",
    "hmsv",
    "hmv"
  )
}

#' Get a vector of all 24-hour flexible time types
#'
#' @description
#' The `flex_t24_vec()` function produces a vector of all supported flexible
#' 24-hour time types in the **bigD** package. These types are essentially
#' identifiers for classes of cross-locale time formatting, so, none of these
#' should be used directly in the `format` argument of the `fdt()` function (use
#' the [flex_t24_lst] object for that).
#'
#' @return
#' A character vector of supported 24-hour flexible time types.
#'
#' @export
flex_t24_vec <- function() {

  c(
    "Hms",
    "Hm",
    "H",
    "EHm",
    "EHms",
    "Hmsv",
    "Hmv",
    "ms"
  )
}

#' A list of all flexible date types
#'
#' @description
#' The `flex_d_lst` object is a list of widely supported flexible date types.
#' Flexible date types are classes of date formatting which can be translated
#' across locales. There are `r length(flex_d_lst)` flexible date types in
#' `flex_d_lst`.
#'
#' @return A list where each element corresponds to a classifier for a flexible
#' date type.
#'
#' @section Examples:
#'
#' The `flex_d_lst` object can be incredibly useful when you need to get a
#' format for date formatting that works across all locales. You can avoid
#' typing errors by using this list and every flexible date type from this list
#' is guaranteed to work across all supported locales. In this example, we'll
#' use the `"yMMMEd"` flexible date type by accessing it from the `flex_d_lst`
#' object.
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05",
#'   format = flex_d_lst$yMMMEd,
#'   locale = "en"
#' )
#' ```
#' ```
#' #> [1] "Wed, Jul 4, 2018"
#' ```
#'
#' If we wanted this in a different locale, the locale-specific `format` pattern
#' behind the flexible date identifier would ensure consistency while moving to
#' that locale. Let's use the `fdt_locales_lst` object in the same spirit to
#' specify the French (Canada) locale.
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05",
#'   format = flex_d_lst$yMMMEd,
#'   locale = fdt_locales_lst$fr_CA
#' )
#' ```
#' ```
#' #> [1] "mer. 4 juill. 2018"
#' ```
#'
#' @export
flex_d_lst <-
  lapply(
    stats::setNames(as.list(flex_d_vec()), as.list(flex_d_vec())),
    FUN = function(x) {
      class(x) <- c("date_time_pattern", "flex_d")
      x
    }
  )

#' A list of all 12-hour flexible time types
#'
#' @description
#' The `flex_t12_lst` object is a list of the 12-hour flexible time types which
#' are widely supported. Flexible time types are classes of time formatting
#' which can be translated across locales. There are `r length(flex_t12_lst)`
#' flexible time types of the 12-hour variety in `flex_t12_lst`.
#'
#' @return A list where each element corresponds to a classifier for a 12-hour
#' flexible time type.
#'
#' @section Examples:
#'
#' The `flex_t12_lst` object can be incredibly useful when you need to get a
#' format for 12-hour time formatting that works across all locales. You can
#' avoid typing errors by using this list and every flexible time type from this
#' list is guaranteed to work across all supported locales. In this example,
#' we'll use the `"Ehms"` flexible time type by accessing it from the
#' `flex_t12_lst` object.
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05",
#'   format = flex_t12_lst$Bhms,
#'   locale = "en"
#' )
#' ```
#' ```
#' #> [1] "10:05:00 at night"
#' ```
#'
#' If we wanted this in a different locale, the locale-specific `format` pattern
#' behind the flexible date identifier would ensure consistency while moving to
#' that locale. Let's use the `fdt_locales_lst` object in the same spirit to
#' specify the German (Austria) locale.
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05",
#'   format = flex_t12_lst$Bhms,
#'   locale = fdt_locales_lst$de_AT
#' )
#' ```
#' ```
#' #> [1] "10:05:00 abends"
#' ```
#'
#' @export
flex_t12_lst <-
  lapply(
    stats::setNames(as.list(flex_t12_vec()), as.list(flex_t12_vec())),
    FUN = function(x) {
      class(x) <- c("date_time_pattern", "flex_t12")
      x
    }
  )

#' A list of all 24-hour flexible time types
#'
#' @description
#' The `flex_t24_lst` object is a list of the 24-hour flexible time types which
#' are widely supported. Flexible time types are classes of time formatting
#' which can be translated across locales. There are `r length(flex_t24_lst)`
#' flexible time types of the 24-hour variety in `flex_t24_lst`.
#'
#' @return A list where each element corresponds to a classifier for a 24-hour
#' flexible time type.
#'
#' @section Examples:
#'
#' The `flex_t24_lst` object can be incredibly useful when you need to get a
#' format for 24-hour time formatting that works across all locales. You can
#' avoid typing errors by using this list and every flexible time type from this
#' list is guaranteed to work across all supported locales. In this example,
#' we'll use the `"EHms"` flexible time type by accessing it from the
#' `flex_t24_lst` object.
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05",
#'   format = flex_t24_lst$EHms,
#'   locale = "en"
#' )
#' ```
#' ```
#' #> [1] "Wed 22:05:00"
#' ```
#'
#' If we wanted this in a different locale, the locale-specific `format` pattern
#' behind the flexible date identifier would ensure consistency while moving to
#' that locale. Let's use the `fdt_locales_lst` object in the same spirit to
#' specify the German locale.
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05",
#'   format = flex_t24_lst$EHms,
#'   locale = fdt_locales_lst$de
#' )
#' ```
#' ```
#' #> [1] "Mi, 22:05:00"
#' ```
#'
#' @export
flex_t24_lst <-
  lapply(
    stats::setNames(as.list(flex_t24_vec()), as.list(flex_t24_vec())),
    FUN = function(x) {
      class(x) <- c("date_time_pattern", "flex_t24")
      x
    }
  )

#' Get a vector of all the short weekday names
#'
#' @description
#' The `names_wkdays()` function produces a vector of all short weekday names
#' used by the **bigD** package.
#'
#' @return
#' A character vector of short weekday names.
#'
#' @examples
#' # Let's get all the short weekday names with
#' # the `names_wkdays()` function
#' names_wkdays()
#'
#' @export
names_wkdays <- function() {
  c(
    "sun", "mon", "tue", "wed", "thu", "fri", "sat"
  )
}

#' Get a vector of all the short month names
#'
#' @description
#' The `names_months()` function produces a vector of all short month names
#' used by the **bigD** package.
#'
#' @return
#' A character vector of short month names.
#'
#' @examples
#' # Let's get all the short month names with
#' # the `names_months()` function
#' names_months()
#'
#' @export
names_months <- function() {
  c(
    "jan", "feb", "mar", "apr", "may", "jun",
    "jul", "aug", "sep", "oct", "nov", "dec"
  )
}

#' Get a named vector of all first-day-of-the-week names for different regions
#'
#' @description
#' The `names_months()` function produces a vector of all short month names
#' used by the **bigD** package.
#'
#' @return
#' A character vector of short month names.
#'
#' @examples
#' # Let's get a vector of regions where the
#' # first day of the week is Saturday
#' names(first_day_of_week()[first_day_of_week() == "sat"])
#'
#' @export
first_day_of_week <- function() {
  c(
    "001" = "mon",
    "AD" = "mon",
    "AE" = "sat",
    "AF" = "sat",
    "AG" = "sun",
    "AI" = "mon",
    "AL" = "mon",
    "AM" = "mon",
    "AN" = "mon",
    "AR" = "mon",
    "AS" = "sun",
    "AT" = "mon",
    "AU" = "mon",
    "AX" = "mon",
    "AZ" = "mon",
    "BA" = "mon",
    "BD" = "sun",
    "BE" = "mon",
    "BG" = "mon",
    "BH" = "sat",
    "BM" = "mon",
    "BN" = "mon",
    "BR" = "sun",
    "BS" = "sun",
    "BT" = "sun",
    "BW" = "sun",
    "BY" = "mon",
    "BZ" = "sun",
    "CA" = "sun",
    "CH" = "mon",
    "CL" = "mon",
    "CM" = "mon",
    "CN" = "sun",
    "CO" = "sun",
    "CR" = "mon",
    "CY" = "mon",
    "CZ" = "mon",
    "DE" = "mon",
    "DJ" = "sat",
    "DK" = "mon",
    "DM" = "sun",
    "DO" = "sun",
    "DZ" = "sat",
    "EC" = "mon",
    "EE" = "mon",
    "EG" = "sat",
    "ES" = "mon",
    "ET" = "sun",
    "FI" = "mon",
    "FJ" = "mon",
    "FO" = "mon",
    "FR" = "mon",
    "GB" = "mon",
    "GE" = "mon",
    "GF" = "mon",
    "GP" = "mon",
    "GR" = "mon",
    "GT" = "sun",
    "GU" = "sun",
    "HK" = "sun",
    "HN" = "sun",
    "HR" = "mon",
    "HU" = "mon",
    "ID" = "sun",
    "IE" = "mon",
    "IL" = "sun",
    "IN" = "sun",
    "IQ" = "sat",
    "IR" = "sat",
    "IS" = "mon",
    "IT" = "mon",
    "JM" = "sun",
    "JO" = "sat",
    "JP" = "sun",
    "KE" = "sun",
    "KG" = "mon",
    "KH" = "sun",
    "KR" = "sun",
    "KW" = "sat",
    "KZ" = "mon",
    "LA" = "sun",
    "LB" = "mon",
    "LI" = "mon",
    "LK" = "mon",
    "LT" = "mon",
    "LU" = "mon",
    "LV" = "mon",
    "LY" = "sat",
    "MC" = "mon",
    "MD" = "mon",
    "ME" = "mon",
    "MH" = "sun",
    "MK" = "mon",
    "MM" = "sun",
    "MN" = "mon",
    "MO" = "sun",
    "MQ" = "mon",
    "MT" = "sun",
    "MV" = "fri",
    "MX" = "sun",
    "MY" = "mon",
    "MZ" = "sun",
    "NI" = "sun",
    "NL" = "mon",
    "NO" = "mon",
    "NP" = "sun",
    "NZ" = "mon",
    "OM" = "sat",
    "PA" = "sun",
    "PE" = "sun",
    "PH" = "sun",
    "PK" = "sun",
    "PL" = "mon",
    "PR" = "sun",
    "PT" = "sun",
    "PY" = "sun",
    "QA" = "sat",
    "RE" = "mon",
    "RO" = "mon",
    "RS" = "mon",
    "RU" = "mon",
    "SA" = "sun",
    "SD" = "sat",
    "SE" = "mon",
    "SG" = "sun",
    "SI" = "mon",
    "SK" = "mon",
    "SM" = "mon",
    "SV" = "sun",
    "SY" = "sat",
    "TH" = "sun",
    "TJ" = "mon",
    "TM" = "mon",
    "TR" = "mon",
    "TT" = "sun",
    "TW" = "sun",
    "UA" = "mon",
    "UM" = "sun",
    "US" = "sun",
    "UY" = "mon",
    "UZ" = "mon",
    "VA" = "mon",
    "VE" = "sun",
    "VI" = "sun",
    "VN" = "mon",
    "WS" = "sun",
    "XK" = "mon",
    "YE" = "sun",
    "ZA" = "sun",
    "ZW" = "sun"
  )
}

braced <- function(value) {
  paste0("{", value, "}")
}

substr_right <- function(x, n) {
  substr(x, nchar(x) - n + 1, nchar(x))
}

#
# Taken from the i18n package
#
dates_elements_bigd <-
  list(
    months_format_abbrev = "months_format_abbrev",
    months_format_narrow = "months_format_narrow",
    months_format_wide = "months_format_wide",
    months_standalone_abbrev = "months_standalone_abbrev",
    months_standalone_narrow = "months_standalone_narrow",
    months_standalone_wide = "months_standalone_wide",
    days_format_abbrev = "days_format_abbrev",
    days_format_narrow = "days_format_narrow",
    days_format_short = "days_format_short",
    days_format_wide = "days_format_wide",
    days_standalone_abbrev = "days_standalone_abbrev",
    days_standalone_narrow = "days_standalone_narrow",
    days_standalone_short = "days_standalone_short",
    days_standalone_wide = "days_standalone_wide",
    quarters_format_abbrev = "quarters_format_abbrev",
    quarters_format_narrow = "quarters_format_narrow",
    quarters_format_wide = "quarters_format_wide",
    quarters_standalone_abbrev = "quarters_standalone_abbrev",
    quarters_standalone_narrow = "quarters_standalone_narrow",
    quarters_standalone_wide = "quarters_standalone_wide",
    dayperiods_format_abbrev = "dayperiods_format_abbrev",
    dayperiods_format_narrow = "dayperiods_format_narrow",
    dayperiods_format_wide = "dayperiods_format_wide",
    dayperiods_standalone_abbrev = "dayperiods_standalone_abbrev",
    dayperiods_standalone_narrow = "dayperiods_standalone_narrow",
    dayperiods_standalone_wide = "dayperiods_standalone_wide",
    eras_abbrev = "eras_abbrev",
    eras_names = "eras_names",
    eras_narrow = "eras_narrow",
    date_formats = "date_formats",
    date_skeletons = "date_skeletons",
    time_formats = "time_formats",
    time_skeletons = "time_skeletons",
    date_time_patterns = "date_time_patterns",
    date_time_available_formats = "date_time_available_formats",
    date_time_append_items = "date_time_append_items",
    date_time_interval_formats = "date_time_interval_formats"
  )

#
# Taken from the i18n package
#
cldr_dates_bigd <- function(locale = "en", element) {

  rows <- which(dates$locale == locale)
  values <- dates[[rows, element]]
  values <- unlist(values, use.names = TRUE)

  names(values) <- sub("^value\\.", "", names(values))

  as.list(values)
}

# TODO Remove this when depending on R 4.4
`%||%` <- function(x, y) {
  if (is.null(x)) y else x
}

#' Format a datetime with a formatting string
#'
#' With `fdt()`, we can format datetime values with the greatest of ease, and,
#' with great power. There is a lot of leniency in what types of input
#' date/time/datetime values can be passed in. The formatting string allows for
#' a huge array of possibilities when formatting. Not only that, we can set a
#' `locale` value and get the formatted values localized in the language/region
#' of choice. There's plenty of ways to represent time zone information, and
#' this goes along with the option to enrich the input values with a precise
#' time zone identifier (like `"America/Los_Angeles"`). The choices are ample
#' here, with the goal being a comprehensiveness and ease-of-use in date/time
#' formatting.
#'
#' @section Valid Input Values:
#'
#' The `input` argument of the `fdt()` function allows for some flexibility on
#' what can be passed in. This section describes the kinds of inputs that are
#' understandable by `fdt()`. A vector of strings is allowed, as are vectors of
#' `Date` or `POSIXct` values.
#'
#' If using strings, a good option is to use those that adhere to the ISO
#' 8601:2004 standard. For a datetime this can be of the form
#' `YYYY-MM-DDThh:mm:ss.s<TZD>`. With this, `YYYY-MM-DD` corresponds to the
#' date, the literal `"T"` is optional, `hh:mm:ss` is the time (where seconds,
#' `ss`, is optional as are `.s` for fractional seconds), and `<TZD>` refers to
#' an optional time-zone designation (more on time zones in the next paragraph).
#' You can provide just the date part, and this assumes midnight as an implicit
#' time. It's also possible to provide just the time part, and this internally
#' assembles a datetime that uses the current date. When formatting standalone
#' dates or times, you'll probably just format the explicit parts but `fdt()`
#' won't error if you format the complementary parts.
#'
#' The time zone designation on string-based datetimes is completely optional.
#' If not provided then `"UTC"` is assumed. If you do want to supply time zone
#' information, it can be given as an offset value with the following
#' constructions:
#'
#' - `<time>Z`
#' - `<time>(+/-)hh:mm`
#' - `<time>(+/-)hhmm`
#' - `<time>(+/-)hh`
#'
#' The first, `<time>Z`, is zone designator for the zero UTC offset; it's
#' equivalent to `"+00:00"`. The next two are formats for providing the time
#' offsets from UTC with hours and minutes fields. Examples are `"-05:00"` (New
#' York, standard time), `"+0200"` (Cairo), and `"+05:30"` (Mumbai). Note that
#' the colon is optional but leading zeros to maintain two-digit widths are
#' essential. The final format, `<time>(+/-)hh`, omits the minutes field and as
#' so many offsets have `"00"` minutes values, this can be convenient.
#'
#' We can also supply an Olson/IANA-style time zone identifier (tzid) in
#' parentheses within the string, or, as a value supplied to `use_tz` (should a
#' tzid apply to all date/time/datetime values in the `input` vector). By
#' extension, this would use the form: `YYYY-MM-DDThh:mm:ss.s<TZD>(<tzid>)`.
#' Both a `<TZD>` (UTC offset value) and a `<tzid>` shouldn't really be used
#' together but if that occurs the `<tzid>` overrides the UTC offset. Here are
#' some examples:
#'
#' - `"2018-07-04T22:05 (America/Vancouver)"` (preferable)
#' - `"2018-07-04T22:05-0800(America/Vancouver)"` (redundant, but still okay)
#'
#' A tzid contains much more information about the time zone than a UTC offset
#' value since it is tied to some geographical location and the timing of
#' Standard Time (STD) and Daylight Saving Time (DST) is known. In essence we
#' can derive UTC offset values from a tzid and also a host of other identifiers
#' (time zone names, their abbreviations, etc.). Here are some examples of valid
#' tzid values that can be used:
#'
#' - `"America/Jamaica"` (the official time in Jamaica, or, `"Jamaica Time"`)
#' - `"Australia/Perth"` (`"+08:00"` year round in Western Australia)
#' - `"Europe/Dublin"` (IST/GMT time: `"+01:00"`/`"+00:00"`)
#'
#' The tz database (a compilation of information about the world's time zones)
#' consists of canonical zone names (those that are primary and preferred) and
#' alternative names (less preferred in modern usage, and was either discarded
#' or more commonly replaced by a canonical zone name). The `fdt()` function can
#' handle both types and what occurs is that non-canonical tzid values are
#' internally mapped onto canonical zone names. Here's a few examples:
#'
#' - `"Africa/Luanda"` (in Angola) maps to `"Africa/Lagos"`
#' - `"America/Indianapolis"` maps to `"America/Indiana/Indianapolis"`
#' - `"Asia/Calcutta"` maps to `"Asia/Kolkata"`
#' - `"Pacific/Midway"` maps to `"Pacific/Pago_Pago"`
#' - `"Egypt"` maps to `"Africa/Cairo"`
#'
#' For the most part, the Olson-format tzid follows the form `"{region}/{city}"`
#' where the region is usually a continent, the city is considered an 'exemplar
#' city', and the exemplar city itself belongs in a country.
#'
#' @section Date/Time Format Syntax:
#'
#' A formatting pattern as used in **bigD** consists of a string of characters,
#' where certain strings are replaced with date and time data that are derived
#' from the parsed input.
#'
#' The characters used in patterns are tabulated below to show which specific
#' strings produce which outputs (e.g., `"y"` for the year). A common pattern is
#' characters that are used consecutively to produce variations on a date, time,
#' or timezone output. Say that the year in the input is 2015. If using `"yy"`
#' you'll get `"15"` but with `"yyyy"` the output becomes `"2015"`. There's a
#' whole lot of this, so the following subsections try to illustrate as best as
#' possible what each string will produce. All of the examples will use this
#' string-based datetime input unless otherwise indicated:
#'
#' `"2018-07-04T22:05:09.2358(America/Vancouver)"`
#'
#' ## Year
#'
#' ### Calendar Year (little y)
#'
#' This yields the calendar year, which is always numeric. In most cases the
#' length of the `"y"` field specifies the minimum number of digits to display,
#' zero-padded as necessary. More digits will be displayed if needed to show the
#' full year. There is an exception: `"yy"` gives use just the two low-order
#' digits of the year, zero-padded as necessary. For most use cases, `"y"` or
#' `"yy"` should be good enough.
#'
#' | Field Patterns                 | Output                                 |
#' |------------------------------- |----------------------------------------|
#' | `"y"`                          | `"2018"`                               |
#' | `"yy"`                         | `"18"`                                 |
#' | `"yyy"` to `"yyyyyyyyy"`       | `"2018"` to `"000002018"`              |
#'
#' ### Year in the Week in Year Calendar (big Y)
#'
#' This is the year in 'Week of Year' based calendars in which the year
#' transition occurs on a week boundary. This may differ from calendar year
#' `"y"` near a year transition. This numeric year designation is used in
#' conjunction with pattern character `"w"` in the ISO year-week calendar as
#' defined by ISO 8601.
#'
#' | Field Patterns                 | Output                                 |
#' |--------------------------------|----------------------------------------|
#' | `"Y"`                          | `"2018"`                               |
#' | `"YY"`                         | `"18"`                                 |
#' | `"YYY"` to `"YYYYYYYYY"`       | `"2018"` to `"000002018"`              |
#'
#' ## Quarter
#'
#' ### Quarter of the Year: formatting (big Q) and standalone (little q)
#'
#' The quarter names are identified numerically, starting at `1` and ending at
#' `4`. Quarter names may vary along two axes: the width and the context. The
#' context is either 'formatting' (taken as a default), which the form used
#' within a complete date format string, or, 'standalone', the form for date
#' elements used independently (such as in calendar headers). The standalone
#' form may be used in any other date format that shares the same form of the
#' name. Here, the formatting form for quarters of the year consists of some run
#' of `"Q"` values whereas the standalone form uses `"q"`.
#'
#' | Field Patterns    | Output          | Notes                             |
#' |-------------------|-----------------|-----------------------------------|
#' | `"Q"`/`"q"`       | `"3"`           | Numeric, one digit                |
#' | `"QQ"`/`"qq"`     | `"03"`          | Numeric, two digits (zero padded) |
#' | `"QQQ"`/`"qqq"`   | `"Q3"`          | Abbreviated                       |
#' | `"QQQQ"`/`"qqqq"` | `"3rd quarter"` | Wide                              |
#' | `"QQQQQ"`/`"qqqqq"` | `"3"`         | Narrow                            |
#'
#' ## Month
#'
#' ### Month: formatting (big M) and standalone (big L)
#'
#' The month names are identified numerically, starting at `1` and ending at
#' `12`. Month names may vary along two axes: the width and the context. The
#' context is either 'formatting' (taken as a default), which the form used
#' within a complete date format string, or, 'standalone', the form for date
#' elements used independently (such as in calendar headers). The standalone
#' form may be used in any other date format that shares the same form of the
#' name. Here, the formatting form for months consists of some run of `"M"`
#' values whereas the standalone form uses `"L"`.
#'
#' | Field Patterns    | Output          | Notes                             |
#' |-------------------|-----------------|-----------------------------------|
#' | `"M"`/`"L"`       | `"7"`           | Numeric, minimum digits           |
#' | `"MM"`/`"LL"`     | `"07"`          | Numeric, two digits (zero padded) |
#' | `"MMM"`/`"LLL"`   | `"Jul"`         | Abbreviated                       |
#' | `"MMMM"`/`"LLLL"` | `"July"`        | Wide                              |
#' | `"MMMMM"`/`"LLLLL"` | `"J"`         | Narrow                            |
#'
#' ## Week
#'
#' ### Week of Year (little w)
#'
#' Values calculated for the week of year range from `1` to `53`. Week `1` for a
#' year is the first week that contains at least the specified minimum number of
#' days from that year. Weeks between week `1` of one year and week `1` of the
#' following year are numbered sequentially from `2` to `52` or `53` (if
#' needed).
#'
#' There are two available field lengths. Both will display the week of year
#' value but the `"ww"` width will always show two digits (where weeks `1` to
#' `9` are zero padded).
#'
#' | Field Patterns   | Output    | Notes                                    |
#' |------------------|-----------|------------------------------------------|
#' | `"w"`            | `"27"`    | Minimum digits                           |
#' | `"ww"`           | `"27"`    | Two digits (zero padded)                 |
#'
#' ### Week of Month (big W)
#'
#' The week of a month can range from `1` to `5`. The first day of every month
#' always begins at week `1` and with every transition into the beginning of a
#' week, the week of month value is incremented by `1`.
#'
#' | Field Pattern    | Output                                               |
#' |------------------|------------------------------------------------------|
#' | `"W"`            | `"1"`                                                |
#'
#' ## Day
#'
#' ### Day of Month (little d)
#'
#' The day of month value is always numeric and there are two available field
#' length choices in its formatting. Both will display the day of month value
#' but the `"dd"` formatting will always show two digits (where days `1` to `9`
#' are zero padded).
#'
#' | Field Patterns | Output    | Notes                                      |
#' |----------------|-----------|--------------------------------------------|
#' | `"d"`          | `"4"`     | Minimum digits                             |
#' | `"dd"`         | `"04"`    | Two digits, zero padded                    |
#'
#' ### Day of Year (big D)
#'
#' The day of year value ranges from `1` (January 1) to either `365` or `366`
#' (December 31), where the higher value of the range indicates that the year is
#' a leap year (29 days in February, instead of 28). The field length specifies
#' the minimum number of digits, with zero-padding as necessary.
#'
#' | Field Patterns  | Output   | Notes                                      |
#' |-----------------|----------|--------------------------------------------|
#' | `"D"`           | `"185"`  |                                            |
#' | `"DD"`          | `"185"`  | Zero padded to minimum width of 2          |
#' | `"DDD"`         | `"185"`  | Zero padded to minimum width of 3          |
#'
#' ### Day of Week in Month (big F)
#'
#' The day of week in month returns a numerical value indicating the number of
#' times a given weekday had occurred in the month (e.g., '2nd Monday in
#' March'). This conveniently resolves to predicable case structure where ranges
#' of day of the month values return predictable day of week in month values:
#'
#' - days `1` - `7` -> `1`
#' - days `8` - `14` -> `2`
#' - days `15` - `21` -> `3`
#' - days `22` - `28` -> `4`
#' - days `29` - `31` -> `5`
#'
#' | Field Pattern                  | Output                                 |
#' |--------------------------------|----------------------------------------|
#' | `"F"`                          | `"1"`                                  |
#'
#' ### Modified Julian Date (little g)
#'
#' The modified version of the Julian date is obtained by subtracting
#' 2,400,000.5 days from the Julian date (the number of days since January 1,
#' 4713 BC). This essentially results in the number of days since midnight
#' November 17, 1858. There is a half day offset (unlike the Julian date, the
#' modified Julian date is referenced to midnight instead of noon).
#'
#' | Field Patterns                 | Output                                 |
#' |--------------------------------|----------------------------------------|
#' | `"g"` to `"ggggggggg"`         | `"58303"` -> `"000058303"`             |
#'
#' ## Weekday
#'
#' ### Day of Week Name (big E)
#'
#' The name of the day of week is offered in four different widths.
#'
#' | Field Patterns             | Output         | Notes                     |
#' |----------------------------|----------------|---------------------------|
#' | `"E"`, `"EE"`, or `"EEE"`  | `"Wed"`        | Abbreviated               |
#' | `"EEEE"`                   | `"Wednesday"`  | Wide                      |
#' | `"EEEEE"`                  | `"W"`          | Narrow                    |
#' | `"EEEEEE"`                 | `"We"`         | Short                     |
#'
#' ## Periods
#'
#' ### AM/PM Period of Day (little a)
#'
#' This denotes before noon and after noon time periods. May be upper or
#' lowercase depending on the locale and other options. The wide form may be
#' the same as the short form if the 'real' long form (e.g. 'ante meridiem') is
#' not customarily used. The narrow form must be unique, unlike some other
#' fields.
#'
#' | Field Patterns                 | Output   | Notes                       |
#' |--------------------------------|----------|-----------------------------|
#' | `"a"`, `"aa"`, or `"aaa"`      | `"PM"`   | Abbreviated                 |
#' | `"aaaa"`                       | `"PM"`   | Wide                        |
#' | `"aaaaa"`                      | `"p"`    | Narrow                      |
#'
#' ### AM/PM Period of Day Plus Noon and Midnight (little b)
#'
#' Provide AM and PM as well as phrases for exactly noon and midnight. May be
#' upper or lowercase depending on the locale and other options. If the locale
#' doesn't have the notion of a unique 'noon' (i.e., 12:00), then the PM form
#' may be substituted. A similar behavior can occur for 'midnight' (00:00) and
#' the AM form. The narrow form must be unique, unlike some other fields.
#'
#' (a) `input_midnight`: `"2020-05-05T00:00:00"`
#' (b) `input_noon`: `"2020-05-05T12:00:00"`
#'
#' | Field Patterns                 | Output             | Notes             |
#' |--------------------------------|--------------------|-------------------|
#' | `"b"`, `"bb"`, or `"bbb"`      | (a) `"midnight"`   | Abbreviated       |
#' |                                | (b) `"noon"`       |                   |
#' | `"bbbb"`                       | (a) `"midnight"`   | Wide              |
#' |                                | (b) `"noon"`       |                   |
#' | `"bbbbb"`                      | (a) `"mi"`         | Narrow            |
#' |                                | (b) `"n"`          |                   |
#'
#' ### Flexible Day Periods (big B)
#'
#' Flexible day periods denotes things like 'in the afternoon', 'in the
#' evening', etc., and the flexibility comes from a locale's language and
#' script. Each locale has an associated rule set that specifies when the day
#' periods start and end for that locale.
#'
#' (a) `input_morning`: `"2020-05-05T00:08:30"`
#' (b) `input_afternoon`: `"2020-05-05T14:00:00"`
#'
#' | Field Patterns             | Output                   | Notes           |
#' |----------------------------|--------------------------|-----------------|
#' | `"B"`, `"BB"`, or `"BBB"`  | (a) `"in the morning"`   | Abbreviated     |
#' |                            | (b) `"in the afternoon"` |                 |
#' | `"BBBB"`                   | (a) `"in the morning"`   | Wide            |
#' |                            | (b) `"in the afternoon"` |                 |
#' | `"BBBBB"`                  | (a) `"in the morning"`   | Narrow          |
#' |                            | (b) `"in the afternoon"` |                 |
#'
#' ## Hours, Minutes, and Seconds
#'
#' ### Hour 0-23 (big H)
#'
#' Hours from `0` to `23` are for a standard 24-hour clock cycle (midnight plus
#' 1 minute is `00:01`) when using `"HH"` (which is the more common width that
#' indicates zero-padding to 2 digits).
#'
#' Using `"2015-08-01T08:35:09"`:
#'
#' | Field Patterns         | Output  | Notes                                |
#' |------------------------|---------|--------------------------------------|
#' | `"H"`                  | `"8"`   | Numeric, minimum digits              |
#' | `"HH"`                 | `"08"`  | Numeric, 2 digits (zero padded)      |
#'
#' ### Hour 1-12 (little h)
#'
#' Hours from `1` to `12` are for a standard 12-hour clock cycle (midnight plus
#' 1 minute is `12:01`) when using `"hh"` (which is the more common width that
#' indicates zero-padding to 2 digits).
#'
#' Using `"2015-08-01T08:35:09"`:
#'
#' | Field Patterns         | Output  | Notes                                |
#' |------------------------|---------|--------------------------------------|
#' | `"h"`                  | `"8"`   | Numeric, minimum digits              |
#' | `"hh"`                 | `"08"`  | Numeric, 2 digits (zero padded)      |
#'
#' ### Hour 1-24 (little k)
#'
#' Using hours from `1` to `24` is a less common way to express a 24-hour clock
#' cycle (midnight plus 1 minute is `24:01`) when using `"kk"` (which is the
#' more common width that indicates zero-padding to 2 digits).
#'
#' Using `"2015-08-01T08:35:09"`:
#'
#' | Field Patterns         | Output  | Notes                                |
#' |------------------------|---------|--------------------------------------|
#' | `"k"`                  | `"9"`   | Numeric, minimum digits              |
#' | `"kk"`                 | `"09"`  | Numeric, 2 digits (zero padded)      |
#'
#' ### Hour 0-11 (big K)
#'
#' Using hours from `0` to `11` is a less common way to express a 12-hour clock
#' cycle (midnight plus 1 minute is `00:01`) when using `"KK"` (which is the
#' more common width that indicates zero-padding to 2 digits).
#'
#' Using `"2015-08-01T08:35:09"`:
#'
#' | Field Patterns         | Output  | Notes                                |
#' |------------------------|---------|--------------------------------------|
#' | `"K"`                  | `"7"`   | Numeric, minimum digits              |
#' | `"KK"`                 | `"07"`  | Numeric, 2 digits (zero padded)      |
#'
#' ### Minute (little m)
#'
#' The minute of the hour which can be any number from `0` to `59`. Use `"m"` to
#' show the minimum number of digits, or `"mm"` to always show two digits
#' (zero-padding, if necessary).
#'
#' | Field Patterns         | Output  | Notes                                |
#' |------------------------|---------|--------------------------------------|
#' | `"m"`                  | `"5"`   | Numeric, minimum digits              |
#' | `"mm"`                 | `"06"`  | Numeric, 2 digits (zero padded)      |
#'
#' ### Seconds (little s)
#'
#' The second of the minute which can be any number from `0` to `59`. Use `"s"`
#' to show the minimum number of digits, or `"ss"` to always show two digits
#' (zero-padding, if necessary).
#'
#' | Field Patterns         | Output  | Notes                                |
#' |------------------------|---------|--------------------------------------|
#' | `"s"`                  | `"9"`   | Numeric, minimum digits              |
#' | `"ss"`                 | `"09"`  | Numeric, 2 digits (zero padded)      |
#'
#' ### Fractional Second (big S)
#'
#' The fractional second truncates (like other time fields) to the width
#' requested (i.e., count of letters). So using pattern `"SSSS"` will display
#' four digits past the decimal (which, incidentally, needs to be added manually
#' to the pattern).
#'
#' | Field Patterns                 | Output                                 |
#' |--------------------------------|----------------------------------------|
#' | `"S"` to `"SSSSSSSSS"`         | `"2"` -> `"235000000"`                 |
#'
#' ### Milliseconds Elapsed in Day (big A)
#'
#' There are 86,400,000 milliseconds in a day and the `"A"` pattern will provide
#' the whole number. The width can go up to nine digits with `"AAAAAAAAA"` and
#' these higher field widths will result in zero padding if necessary.
#'
#' Using `"2011-07-27T00:07:19.7223"`:
#'
#' | Field Patterns                 | Output                                 |
#' |--------------------------------|----------------------------------------|
#' | `"A"` to `"AAAAAAAAA"`         | `"439722"` -> `"000439722"`            |
#'
#' ## Era
#'
#' ### The Era Designator (big G)
#'
#' This provides the era name for the given date. The Gregorian calendar has two
#' eras: AD and BC. In the AD year numbering system, AD 1 is immediately
#' preceded by 1 BC, with nothing in between them (there was no year zero).
#'
#' | Field Patterns                 | Output          | Notes                |
#' |--------------------------------|-----------------|----------------------|
#' | `"G"`, `"GG"`, or `"GGG"`      | `"AD"`          | Abbreviated          |
#' | `"GGGG"`                       | `"Anno Domini"` | Wide                 |
#' | `"GGGGG"`                      | `"A"`           | Narrow               |
#'
#' ## Time Zones
#'
#' ### TZ // Short and Long Specific non-Location Format (little z)
#'
#' The short and long specific non-location formats for time zones are suggested
#' for displaying a time with a user friendly time zone name. Where the short
#' specific format is unavailable, it will fall back to the short localized GMT
#' format (`"O"`). Where the long specific format is unavailable, it will fall
#' back to the long localized GMT format (`"OOOO"`).
#'
#' | Field Patterns             | Output                    | Notes          |
#' |----------------------------|---------------------------|----------------|
#' | `"z"`, `"zz"`, or `"zzz"`  | `"PDT"`                   | Short Specific |
#' | `"zzzz"`                   | `"Pacific Daylight Time"` | Long Specific  |
#'
#' ### TZ // Common UTC Offset Formats (big Z)
#'
#' The ISO8601 basic format with hours, minutes and optional seconds fields is
#' represented by `"Z"`, `"ZZ"`, or `"ZZZ"`. The format is equivalent to RFC 822
#' zone format (when the optional seconds field is absent). This is equivalent
#' to the `"xxxx"` specifier. The field pattern `"ZZZZ"` represents the long
#' localized GMT format. This is equivalent to the `"OOOO"` specifier. Finally,
#' `"ZZZZZ"` pattern yields the ISO8601 extended format with hours, minutes and
#' optional seconds fields. The ISO8601 UTC indicator `Z` is used when local
#' time offset is `0`. This is equivalent to the `"XXXXX"` specifier.
#'
#' | Field Patterns             | Output       | Notes                       |
#' |----------------------------|--------------|-----------------------------|
#' | `"Z"`, `"ZZ"`, or `"ZZZ"`  | `"-0700"`    | ISO 8601 basic format       |
#' | `"ZZZZ"`                   | `"GMT-7:00"` | Long localized GMT format   |
#' | `"ZZZZZ"`                  | `"-07:00"`   | ISO 8601 extended format    |
#'
#' ### TZ // Short and Long Localized GMT Formats (big O)
#'
#' The localized GMT formats come in two widths `"O"` (which removes the minutes
#' field if it's `0`) and `"OOOO"` (which always contains the minutes field).
#' The use of the `GMT` indicator changes according to the locale.
#'
#' | Field Patterns          | Output        | Notes                         |
#' |-------------------------|---------------|-------------------------------|
#' | `"O"`                   | `"GMT-7"`     | Short localized GMT format    |
#' | `"OOOO"`                | `"GMT-07:00"` | Long localized GMT format     |
#'
#' ### TZ // Short and Long Generic non-Location Formats (little v)
#'
#' The generic non-location formats are useful for displaying a recurring wall
#' time (e.g., events, meetings) or anywhere people do not want to be overly
#' specific. Where either of these is unavailable, there is a fallback to the
#' generic location format (`"VVVV"`), then the short localized GMT format as
#' the final fallback.
#'
#' | Field Patterns  | Output           | Notes                              |
#' |-----------------|------------------|------------------------------------|
#' | `"v"`           | `"PT"`           | Short generic non-location format  |
#' | `"vvvv"`        | `"Pacific Time"` | Long generic non-location format   |
#'
#' ### TZ // Short Time Zone IDs and Exemplar City Formats (big V)
#'
#' These formats provide variations of the time zone ID and often include the
#' exemplar city. The widest of these formats, `"VVVV"`, is useful for
#' populating a choice list for time zones, because it supports 1-to-1 name/zone
#' ID mapping and is more uniform than other text formats.
#'
#' | Field Patterns     | Output                | Notes                      |
#' |--------------------|-----------------------|----------------------------|
#' | `"V"`              | `"cavan"`             | Short time zone ID         |
#' | `"VV"`             | `"America/Vancouver"` | Long time zone ID          |
#' | `"VVV"`            | `"Vancouver"`         | The tz exemplar city       |
#' | `"VVVV"`           | `"Vancouver Time"`    | Generic location format    |
#'
#' ### TZ // ISO 8601 Formats with Z for +0000 (big X)
#'
#' The `"X"`-`"XXX"` field patterns represent valid ISO 8601 patterns for time
#' zone offsets in datetimes. The final two widths, `"XXXX"` and `"XXXXX"` allow
#' for optional seconds fields. The seconds field is *not* supported by the ISO
#' 8601 specification. For all of these, the ISO 8601 UTC indicator `Z` is used
#' when the local time offset is `0`.
#'
#' | Field Patterns | Output     | Notes                                     |
#' |----------------|------------|-------------------------------------------|
#' | `"X"`          | `"-07"`    | ISO 8601 basic format (h, optional m)     |
#' | `"XX"`         | `"-0700"`  | ISO 8601 basic format (h & m)             |
#' | `"XXX"`        | `"-07:00"` | ISO 8601 extended format (h & m)          |
#' | `"XXXX"`       | `"-0700"`  | ISO 8601 basic format (h & m, optional s) |
#' | `"XXXXX"`      | `"-07:00"` | ISO 8601 extended format (h & m, optional s) |
#'
#' ### TZ // ISO 8601 Formats (no use of Z for +0000) (little x)
#'
#' The `"x"`-`"xxxxx"` field patterns represent valid ISO 8601 patterns for time
#' zone offsets in datetimes. They are similar to the `"X"`-`"XXXXX"` field
#' patterns except that the ISO 8601 UTC indicator `Z` *will not* be used when
#' the local time offset is `0`.
#'
#' | Field Patterns | Output     | Notes                                     |
#' |----------------|------------|-------------------------------------------|
#' | `"x"`          | `"-07"`    | ISO 8601 basic format (h, optional m)     |
#' | `"xx"`         | `"-0700"`  | ISO 8601 basic format (h & m)             |
#' | `"xxx"`        | `"-07:00"` | ISO 8601 extended format (h & m)          |
#' | `"xxxx"`       | `"-0700"`  | ISO 8601 basic format (h & m, optional s) |
#' | `"xxxxx"`      | `"-07:00"` | ISO 8601 extended format (h & m, optional s) |
#'
#' @param input A vector of date, time, or datetime values. Several
#'   representations are acceptable here including strings, `Date` objects, or
#'   `POSIXct` objects. Refer to the *Valid Input Values* section for more
#'   information.
#' @param format The formatting string to apply to all `input` values. If not
#'   provided, the inputs will be formatted to ISO 8601 datetime strings. The
#'   *Date/Time Format Syntax* section has detailed information on how to create
#'   a formatting string.
#' @param use_tz A tzid (e.g., `"America/New_York"`) time-zone designation for
#'   precise formatting of related outputs. This overrides any time zone
#'   information available in `character`-based input values and is applied to
#'   all vector components.
#' @param locale The output locale to use for formatting the input value
#'   according to the specified locale's rules. Example locale names include
#'   `"en"` for English (United States) and `"es-EC"` for Spanish (Ecuador). If
#'   a locale isn't provided the `"en"` locale will be used. The
#'   [fdt_locales_vec] vector contains the valid locales and [fdt_locales_lst]
#'   list provides an easy way to obtain a valid locale.
#'
#' @return A character vector of formatted dates, times, or datetimes.
#'
#' @section Examples:
#'
#' ## Basics with `input` datetimes, formatting strings, and localization
#'
#' With an input datetime of `"2018-07-04 22:05"` supplied as a string, we can
#' format to get just a date with the full year first, the month abbreviation
#' second, and the day of the month last (separated by hyphens):
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05",
#'   format = "y-MMM-dd"
#' )
#' ```
#' ```
#' #> [1] "2018-Jul-04"
#' ```
#'
#' There are sometimes many options for each time part. Instead of using
#' `"y-MMM-dd"`, let's try a variation on that with `"yy-MMMM-d"`:
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05",
#'   format = "yy-MMMM-d"
#' )
#' ```
#' ```
#' #> [1] "18-July-4"
#' ```
#'
#' The output is localizable and so elements will be translated when supplying
#' the appropriate locale code. Let's use `locale = es` to get the month written
#' in Spanish:
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05",
#'   format = "yy-MMMM-d",
#'   locale = "es"
#' )
#' ```
#' ```
#' #> [1] "18-julio-4"
#' ```
#'
#' `POSIXct` or `POSIXlt` datetimes can serve as an `input` to `fdt()`. Let's
#' create a single datetime value where the timezone is set as `"Asia/Tokyo"`.
#'
#' ```r
#' fdt(
#'   input = lubridate::ymd_hms("2020-03-15 19:09:12", tz = "Asia/Tokyo"),
#'   format = "EEEE, MMMM d, y 'at' h:mm:ss B (VVVV)"
#' )
#' ```
#' ```
#' #> [1] "Sunday, March 15, 2020 at 7:09:12 in the evening (Tokyo Time)"
#' ```
#'
#' If you're going minimal, it's possible to supply an input datetime string
#' without a `format` directive. What this gives us is formatted datetime
#' output that conforms to ISO 8601. Note that the implied time zone is UTC.
#'
#' ```r
#' fdt(input = "2018-07-04 22:05")
#' ```
#' ```
#' #> [1] "2018-07-04T22:05:00Z"
#' ````
#'
#' ## Using locales stored in the [fdt_locales_lst] list
#'
#' The [fdt_locales_lst] object is provided in **bigD** to make it easier to
#' choose one of supported locales. You can avoid typing errors and every
#' element of the list is meant to work. For example, we can use the `"it"`
#' locale by accessing it from [fdt_locales_lst] (autocomplete makes this
#' relatively simple).
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05",
#'   format = "yy-MMMM-d",
#'   locale = fdt_locales_lst$it
#' )
#' ```
#' ```
#' #> [1] "18-luglio-4"
#' ````
#'
#' ## Omission of date or time in `input`
#'
#' You don't have to supply a full datetime to `input`. Just supplying the date
#' portion implies midnight (and is just fine if you're only going to present
#' the date anyway).
#'
#' ```r
#' fdt(input = "2018-07-04")
#' ```
#' ```
#' #> [1] "2018-07-04T00:00:00Z"
#' ````
#'
#' If you omit the date and just supply a time, `fdt()` will correctly parse
#' this. The current date on the user system will be used because we need to
#' create some sort of datetime value internally. Again, this is alright if
#' you just intend to present a formatted time value.
#'
#' ```r
#' fdt(input = "22:05")
#' ```
#' ```
#' #> [1] "2022-08-16T22:05:00Z"
#' ````
#'
#' To see all of the supported locales, we can look at the vector supplied by
#' the [fdt_locales_vec()] function.
#'
#' ## Using standardized forms with the `standard_*()` helper functions
#'
#' With an input datetime of `"2018-07-04 22:05(America/Vancouver)"`, we can
#' format the date and time in a standardized way with `standard_date_time()`
#' providing the correct formatting string. This function is invoked in the
#' `format` argument of `fdt()`:
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date_time(type = "full")
#' )
#' ```
#' ```
#' #> [1] "Wednesday, July 4, 2018 at 10:05:00 PM Pacific Daylight Time"
#' ```
#'
#' The locale can be changed and we don't have to worry about the particulars
#' of the formatting string (they are standardized across locales).
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date_time(type = "full"),
#'   locale = fdt_locales_lst$nl
#' )
#' ```
#' ```
#' #> [1] "woensdag 4 juli 2018 om 22:05:00 Pacific-zomertijd"
#' ```
#'
#' We can use different `type` values to control the output datetime string. The
#' default is `"short"`.
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date_time()
#' )
#' ```
#' ```
#' #> [1] "7/4/18, 10:05 PM"
#' ```
#'
#' After that, it's `"medium"`:
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date_time(type = "medium")
#' )
#' ```
#' ```
#' #> [1] "Jul 4, 2018, 10:05:00 PM"
#' ```
#'
#' The `"short"` and `"medium"` types don't display time zone information in the
#' output. Beginning with `"long"`, the tz is shown.
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date_time(type = "long")
#' )
#' ```
#' ```
#' #> [1] "July 4, 2018 at 10:05:00 PM PDT"
#' ```
#'
#' If you don't include time zone information in the input, the `"UTC"` time
#' zone will be assumed:
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05",
#'   format = standard_date_time(type = "full")
#' )
#' ```
#' ```
#' #> [1] "Wednesday, July 4, 2018 at 10:05:00 PM GMT+00:00"
#' ```
#'
#' ## Using flexible date and time (12- and 24-hour) formatting
#'
#' The **bigD** package supplies a set of lists to allow for flexible date and
#' time formatting ([flex_d_lst], [flex_t24_lst], and [flex_t12_lst]). These
#' are useful when you need a particular format that works across all locales.
#' Here's an example that uses the `"yMMMEd"` flexible date type by accessing it
#' from the `flex_d_lst` object, yielding a formatted date.
#'
#' ```r
#' fdt(
#'   input = "2021-01-09 16:32(America/Toronto)",
#'   format = flex_d_lst$yMMMEd,
#' )
#' ```
#' ```
#' #> [1] "Sat, Jan 9, 2021"
#' ```
#'
#' If we wanted this in a different locale, the locale-specific `format` pattern
#' behind the flexible date identifier would ensure consistency while moving to
#' that locale.
#'
#' ```r
#' fdt(
#'   input = "2021-01-09 16:32(America/Toronto)",
#'   format = flex_d_lst$yMMMEd,
#'   locale = "fr_CA"
#' )
#' ```
#' ```
#' #> [1] "sam. 9 janv. 2021"
#' ```
#'
#' Formatting as a 12-hour time with the [flex_t12_lst] list and using the
#' `"hms"` flexible type:
#'
#' ```r
#' fdt(
#'   input = "2021-01-09 16:32(America/Toronto)",
#'   format = flex_t12_lst$hms
#' )
#' ```
#' ```
#' #> [1] "4:32:00 PM"
#' ```
#'
#' The 24-hour variant, [flex_t24_lst], has a similar `"Hms"` flexible type that
#' will give us a 24-hour version of the same clock time:
#'
#' ```r
#' fdt(
#'   input = "2021-01-09 16:32(America/Toronto)",
#'   format = flex_t24_lst$Hms
#' )
#' ```
#' ```
#' #> [1] "16:32:00"
#' ```
#'
#' A flexible date and time can be used together by enveloping the two in a
#' list (**bigD** will handle putting the date and time together in a sensible
#' manner).
#'
#' ```r
#' fdt(
#'   input = "2021-01-09 16:32(America/Toronto)",
#'   format = list(flex_d_lst$yMMMEd, flex_t24_lst$Hmv)
#' )
#' ```
#' ```
#' #> "Sat, Jan 9, 2021, 16:32 ET"
#' ```
#'
#' @export
fdt <- function(
    input,
    format = NULL,
    use_tz = NULL,
    locale = NULL
) {

  # TODO: Validate locale

  if (!is.null(locale)) {
    locale <- gsub("_", "-", locale, fixed = TRUE)
  }

  locale <- locale %||% "en"

  if (inherits(input, "POSIXlt")) {
    input <- as.POSIXct(input)
  }

  if (inherits(input, "Date")) {
    input <- as.POSIXct(as.POSIXlt(input))
  }


  if (is.null(format)) {
    format <- "yyyy-MM-dd'T'HH:mm:ssXXX"
  } else if (
    length(format) == 1 &&
    is.character(format) &&
    !inherits(format, "date_time_pattern")
  ) {
    NULL
  } else if (
    length(format) == 1 &&
    is.character(format) &&
    inherits(format, "date_time_pattern") &&
    !inherits(format, "standard")
  ) {

    format <- amend_week_pattern(format = format, input = input)

    format <- dates[dates$locale == locale, ][["date_time_available_formats"]][["value"]][[format]]

  } else if (
    length(format) == 1 &&
    is.character(format) &&
    inherits(format, "date_time_pattern") &&
    inherits(format, "standard")
  ) {

    dt_types <- base::setdiff(class(format), c("date_time_pattern", "standard"))
    type_date_time <- dt_types[1]
    type_size <- dt_types[2]

    date_format <-
      dates[dates$locale == locale, ][["date_formats"]][["value"]][[type_size]]

    time_format <-
      dates[dates$locale == locale, ][["time_formats"]][["value"]][[type_size]]

    combining_format <-
      dates[dates$locale == locale, ][["date_time_patterns"]][["value"]][[type_size]]

    if (type_date_time == "date_time") {
      format <- gsub("{1}", date_format, combining_format, fixed = TRUE)
      format <- gsub("{0}", time_format, format, fixed = TRUE)
    } else if (type_date_time == "date") {
      format <- date_format
    } else {
      format <- time_format
    }

  } else if (
    is.list(format) &&
    length(format) == 2 &&
    inherits(format[[1]], "date_time_pattern") &&
    inherits(format[[2]], "date_time_pattern")
  ) {
    if (inherits(format[[1]], "flex_d")) {
      format_1 <- "date"
    } else {
      format_1 <- "time"
    }

    if (inherits(format[[2]], "flex_d")) {
      format_2 <- "date"
    } else {
      format_2 <- "time"
    }

    if (identical(format_1, format_2)) {
      stop(
        "When supplying two flexible formats, one must represent a date and ",
        "the other a time",
        call. = FALSE
      )
    }

    date_format <- if (format_1 == "date") format[[1]] else format[[2]]
    time_format <- if (format_2 == "time") format[[2]] else format[[1]]

    date_format <- amend_week_pattern(format = date_format, input = input)

    combine_length <- "medium"

    date_format <-
      dates[dates$locale == locale, ][["date_time_available_formats"]][["value"]][[date_format]]

    time_format <-
      dates[dates$locale == locale, ][["date_time_available_formats"]][["value"]][[time_format]]

    combining_format <-
      dates[dates$locale == locale, ][["date_time_patterns"]][["value"]][[combine_length]]

    format <- gsub("{1}", date_format, combining_format, fixed = TRUE)
    format <- gsub("{0}", time_format, format, fixed = TRUE)
  }

  dt_out <- rep_len(NA_character_, length(input))

  # Verify if each entry has timezone (checking before loop for performance)
  if (is.character(input)) {
    tz_present <-  is_tz_present(input = input)
  }

  for (i in seq_along(input)) {

    input_i <- input[i]

    # Initialize an empty `tz_info` list
    tz_info <-
      list(
        tz_str = NA_character_,
        tz_offset = NA_real_,
        long_tzid = NA_character_,
        tz_short_specific = NA_character_,
        tz_long_specific = NA_character_
      )

    # Modify the `format` string so it can be more precisely formatted
    pattern_list <- dt_format_pattern(format = format)

    if (is.character(input_i)) {

      date_present <- is_date_present(input = input_i)
      time_present <- is_time_present(input = input_i)

      # Determine if tz information is present, either as:
      # [1] a tz offset in hours from GMT (defined above before loop)
      # [2] a long tz identifier (either canonical or an alias)
      tz_present_i <- tz_present[i]
      long_tzid_present <- is_long_tzid_present(input = input_i)

      # Strip away tz information from the input and return as `input_str`
      if (tz_present_i) {
        input_str <- strip_tz(input = input_i)
      } else if (long_tzid_present) {
        input_str <- strip_long_tzid(input = input_i)
      } else {
        input_str <- input_i
      }

      # Obtain the date and time
      if (date_present) {

        input_dt <- as.POSIXct(gsub("T", " ", input_str, fixed = TRUE), tz = "UTC")

      } else if (!date_present && time_present) {

        date_now <- as.character(Sys.Date())
        input_str <- paste0(date_now, "T", input_str)
        input_dt <- as.POSIXct(gsub("T", " ", input_str, fixed = TRUE), tz = "UTC")

      } else if (!date_present && !time_present) {

        if (nchar(input_str) == 4 && grepl("^[0-9]{4}$", input_str)) {

          # Case where only a four-digit C.E. year is provided
          input_str <- paste0(input_str, "-01-01")
          input_dt <- as.POSIXct(input_str, tz = "UTC")
        }

        if (nchar(input_str) == 7 && grepl("^[0-9]{4}-[0-9]{2}$", input_str)) {

          # Case where the year and month are provided in the YYYY-MM format
          input_str <- paste0(input_str, "-01")
          input_dt <- as.POSIXct(input_str, tz = "UTC")
        }

        if (grepl("W", input_str, fixed = TRUE)) {

          # Case where week dates are possibly provided in one of
          # several permitted ISO 8601 formats

          # Perform check of `input_str` to confirm that week dates
          # conform to one of four ISO 8601 specifications:
          # (1) YYYY-Www
          # (2) YYYYWww
          # (3) YYYY-Www-D
          # (4) YYYYWwwD
          if (!grepl("^([0-9]{4}-?W[0-9]{2}|[0-9]{4}-?W[0-9]{2}-?[1-7])$", input_str)) {
            stop(
              "The provided input '", input_str, "' does not conform to a week ",
              "date representation.",
              call. = FALSE
            )
          }

          # Remove hyphens if any are present
          input_str <- gsub("-", "", input_str, fixed = TRUE)

          input_dt <- week_date_as_datetime(input_str = input_str)
        }

        if (startsWith(input_str, "--")) {

          if (!grepl("^--[0-9]{2}-?[0-9]{2}$", input_str)) {
            stop(
              "The provided input '", input_str, "' does not conform to a ",
              "month-day representation.",
              call. = FALSE
            )
          }

          # Remove all hyphens from `input_str`
          input_str <- gsub("-", "", input_str, fixed = TRUE)

          input_dt <- month_day_as_datetime(input_str = input_str)
        }

        if (grepl("[0-9]+\\s?(BC|AD|BCE|CE|B\\.C\\.E\\.|C\\.E\\.|EV)$", input_str)) {

          # Remove any spaces or periods from `input_str`
          input_str <- gsub(" ", "", input_str, fixed = TRUE)
          input_str <- gsub(".", "", input_str, fixed = TRUE)

          year <- as.integer(gsub("([0-9]+).*", "\\1", input_str))
          era <- gsub("[0-9]+", "", input_str)

          if (any(c("BC", "BCE") %in% era)) {
            year <- year * (-1)
          }

          adjustment <- -1900L

          posixlt_dt <- as.POSIXlt("0000-01-01", tz = "UTC")
          posixlt_dt$year <- year + adjustment
          input_dt <- as.POSIXct(posixlt_dt)
        }

        # TODO: Add parsers for different date/time cases
      }

      # Derive more detailed time zone information from the `long_tzid` value
      if (!is.null(use_tz)) {

        # Validate and then normalize the provided time zone
        validate_long_tzid(long_tzid = use_tz)
        long_tzid <- normalize_long_tzid(long_tzid = use_tz)

        tz_info$long_tzid <- long_tzid
        tz_info$tz_str <- long_tzid_to_tz_str(long_tzid = tz_info$long_tzid, input_dt = input_dt)
        tz_info$tz_offset <- get_tz_offset_val_from_tz_str(tz_str = tz_info$tz_str)
        tz_info$tz_short_specific <- get_tz_short_specific(long_tzid = tz_info$long_tzid, input_dt = input_dt)
        tz_info$tz_long_specific <- get_tz_long_specific(long_tzid = tz_info$long_tzid, input_dt = input_dt, locale = locale)

      } else if (long_tzid_present) {

        tz_info$long_tzid <- get_long_tzid_str(input = input_i)
        tz_info$tz_str <- long_tzid_to_tz_str(long_tzid = tz_info$long_tzid, input_dt = input_dt)
        tz_info$tz_offset <- get_tz_offset_val_from_tz_str(tz_str = tz_info$tz_str)
        tz_info$tz_short_specific <- get_tz_short_specific(long_tzid = tz_info$long_tzid, input_dt = input_dt)
        tz_info$tz_long_specific <- get_tz_long_specific(long_tzid = tz_info$long_tzid, input_dt = input_dt, locale = locale)

      } else if (tz_present_i) {

        tz_info$tz_str <- get_tz_str(input = input_i)
        tz_info$tz_offset <- get_tz_offset_val(input = input_i)

      } else {

        tz_info$long_tzid <- "UTC"
        tz_info$tz_str <- "GMT"
        tz_info$tz_offset <- 0L
        tz_info$tz_short_specific <- get_tz_short_specific(long_tzid = tz_info$long_tzid, input_dt = input_dt)
      }
    }

    if (inherits(input, "POSIXct")) {

      input_dt <- input_i

      # Extract the input time zone
      long_tzid <- attr(input_dt, which = "tzone", exact = TRUE)

      offset_val <- as.POSIXlt(input_dt)$gmtoff / 3600

      if (!is.null(long_tzid) && long_tzid == "UTC") {

        tz_info$long_tzid <- long_tzid
        tz_info$tz_str <- "GMT"
        tz_info$tz_offset <- 0L
        tz_info$tz_short_specific <- get_tz_short_specific(long_tzid = tz_info$long_tzid, input_dt = input_dt)
        tz_info$tz_long_specific <- get_tz_long_specific(long_tzid = tz_info$long_tzid, input_dt = input_dt, locale = locale)

      } else if (is.null(long_tzid)) {

        tz_info$long_tzid <- NA_character_
        tz_info$tz_str <- "GMT"
        tz_info$tz_offset <- offset_val

      } else {

        tz_info$long_tzid <- long_tzid
        tz_info$tz_str <- long_tzid_to_tz_str(long_tzid = tz_info$long_tzid, input_dt = input_dt)
        tz_info$tz_offset <- get_tz_offset_val_from_tz_str(tz_str = tz_info$tz_str)
        tz_info$tz_short_specific <- get_tz_short_specific(long_tzid = tz_info$long_tzid, input_dt = input_dt)
        tz_info$tz_long_specific <- get_tz_long_specific(long_tzid = tz_info$long_tzid, input_dt = input_dt, locale = locale)
      }
    }

    # Replace format "{yyyy}-{MM}-{dd}'1'{HH}:{mm}:{ss}{XXX}"
    # by the corresponding value
    dt <- dt_replace(
      # example "{yyyy}-{MM}-{dd}'1'{HH}:{mm}:{ss}{XXX}"
      dt =  pattern_list$format,
      input_dt = input_dt,
      dt_lett = pattern_list$dt_letters,
      locale = locale,
      tz_info = tz_info
    )

    dt <- gsub("\\{(.*?)\\}", "\\1", dt)

    # Replace string literal markers `'<#>'` with captured literal values
    for (j in seq_along(pattern_list$literals)) {

      dt <-
        gsub(
          paste0("'", j, "'"),
          pattern_list$literals[j],
          dt,
          fixed = TRUE
        )
    }

    # Replace each instance of `''` with `'`
    dt <- gsub("''", "'", dt, fixed = TRUE)

    dt_out[i] <- dt
  }

  dt_out
}

sub_letters <- function() {
  c(
    "G",
    "y",
    "Y",
    "u",
    "U",
    "r",
    "Q",
    "q",
    "M",
    "L",
    "w",
    "W",
    "d",
    "D",
    "F",
    "g",
    "E",
    "e",
    "c",
    "a",
    "b",
    "B",
    "h",
    "H",
    "K",
    "k",
    "m",
    "s",
    "S",
    "A",
    "z",
    "Z",
    "O",
    "v",
    "V",
    "X",
    "x"
  )
}

extract_literals_from_pattern <- function(string) {

  gsub(
    "(^'|'$)", "",
    unlist(regmatches(string, gregexpr(pattern = "'.*?'", text = string)))
  )
}

dt_format_pattern <- function(format) {

  literals <- extract_literals_from_pattern(string = format)

  for (i in seq_along(literals)) {
    if (nzchar(literals[i])) {
      format <- sub(paste0("'", literals[i], "'"), paste0("'", i, "'"), format)
    }
  }

  dt_letters <-
    base::intersect(
      unique(unlist(strsplit(format, "", fixed = TRUE))),
      sub_letters()
    )
  # Bad error if length(dt_letters) == 0
  pattern <- paste0("(", paste0("[", dt_letters, "]+", collapse = "|"), ")")

  format <- gsub(pattern, "\\{\\1\\}", format)

  literals <- gsub("(^'|'$)", "", literals)

  list(
    format = format,
    literals = literals,
    dt_letters = dt_letters
  )
}

amend_week_pattern <- function(format, input) {

  if (format == "MMMMW" || format == "yw") {

    year_week <- format_yearweek(input = input)

    if (grepl("W01", year_week, fixed = TRUE)) {
      format <- paste0(format, "-count-one")
    } else {
      format <- paste0(format, "-count-other")
    }
  }

  format
}

dt_replace <- function(dt, input_dt, dt_lett, locale, tz_info) {


  # Verify that there is no duplication in dt_lett
  stopifnot(anyDuplicated(dt_lett) == 0)

  # This means we can use sub() instead of gsub()

  if ("G" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{G}", dt_G(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{GG}", dt_G(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{GGG}", dt_G(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{GGGG}", dt_GGGG(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{GGGGG}", dt_GGGGG(input_dt, locale), dt, fixed = TRUE)
  }


  if ("y" %in% dt_lett) {
    # Like stringr::str_extract()
    pattern <- regmatches(dt, m = regexpr("\\{y+\\}", dt))
    # Pattern will be of shape "{y+}" and will call the appropriate function
    # For replacing the year in the final value
    # Will call dt_year$`{yy}`(input_dt) and replace the resulting value
    dt <- sub(pattern, dt_year[[pattern]](input_dt), dt, fixed = TRUE)
  }

  if ("Y" %in% dt_lett) {
    pattern <- regmatches(dt, m = regexpr("\\{Y+\\}", dt))
    dt <- sub(pattern, dt_year_full[[pattern]](input_dt), dt, fixed = TRUE)
  }

  if ("u" %in% dt_lett) {
    dt <- sub("{u}", dt_u(input_dt), dt, fixed = TRUE)
  }

  if ("U" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{U}", dt_U(input_dt), dt, fixed = TRUE)
    dt <- sub("{UU}", dt_U(input_dt), dt, fixed = TRUE)
    dt <- sub("{UUU}", dt_U(input_dt), dt, fixed = TRUE)
    dt <- sub("{UUUU}", dt_UUUU(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{UUUUU}", dt_UUUUU(input_dt, locale), dt, fixed = TRUE)
  }

  if ("Q" %in% dt_lett) {
    pattern <- regmatches(dt, m = regexpr("\\{Q+\\}", dt))
    dt <- sub(pattern, dt_quarter_full[[pattern]](input_dt, locale), dt, fixed = TRUE)
  }

  if ("q" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{q}", dt_q(input_dt), dt, fixed = TRUE)
    dt <- sub("{qq}", dt_qq(input_dt), dt, fixed = TRUE)
    dt <- sub("{qqq}", dt_qqq(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{qqqq}", dt_qqqq(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{qqqqq}", dt_qqqqq(input_dt, locale), dt, fixed = TRUE)
  }

  if ("M" %in% dt_lett) {
    # Like stringr::str_extract()
    pattern <- regmatches(dt, m = regexpr("\\{M+\\}", dt))
    # Pattern will be of shape "{y+}" and will call the appropriate function
    # For replacing the year in the final value
    # Will call dt_year$`{MM}`(input_dt) and replace the resulting value
    dt <- sub(pattern, dt_Month[[pattern]](input_dt, locale), dt, fixed = TRUE)
  }

  if ("L" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{L}", dt_L(input_dt), dt, fixed = TRUE)
    dt <- sub("{LL}", dt_LL(input_dt), dt, fixed = TRUE)
    dt <- sub("{LLL}", dt_LLL(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{LLLL}", dt_LLLL(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{LLLLL}", dt_LLLLL(input_dt, locale), dt, fixed = TRUE)
  }

  if ("w" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{w}", dt_w(input_dt), dt, fixed = TRUE)
    dt <- sub("{ww}", dt_ww(input_dt), dt, fixed = TRUE)
  }

  if ("W" %in% dt_lett) {
    dt <- sub("{W}", dt_W(input_dt, locale), dt, fixed = TRUE)
  }

  if ("d" %in% dt_lett) {
    pattern <- regmatches(dt, m = regexpr("\\{d+\\}", dt))
    dt <- sub(pattern, dt_day[[pattern]](input_dt), dt, fixed = TRUE)
  }

  if ("D" %in% dt_lett) {
    pattern <- regmatches(dt, m = regexpr("\\{D+\\}", dt))
    dt <- sub(pattern, dt_day_full[[pattern]](input_dt), dt, fixed = TRUE)
  }

  if ("F" %in% dt_lett) {
    dt <- sub("{F}", dt_F(input_dt), dt, fixed = TRUE)
  }

  if ("E" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{E}", dt_E(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{EE}", dt_E(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{EEE}", dt_E(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{EEEE}", dt_EEEE(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{EEEEE}", dt_EEEEE(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{EEEEEE}", dt_EEEEEE(input_dt, locale), dt, fixed = TRUE)
  }

  if ("e" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{e}", dt_e(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{ee}", dt_ee(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{eee}", dt_eee(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{eeee}", dt_eeee(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{eeeee}", dt_eeeee(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{eeeeee}", dt_eeeeee(input_dt, locale), dt, fixed = TRUE)
  }

  if ("c" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{c}", dt_c(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{cc}", dt_cc(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{ccc}", dt_ccc(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{cccc}", dt_cccc(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{ccccc}", dt_ccccc(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{cccccc}", dt_cccccc(input_dt, locale), dt, fixed = TRUE)
  }

  if ("a" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{a}", dt_a(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{aa}", dt_a(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{aaa}", dt_a(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{aaaa}", dt_aaaa(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{aaaaa}", dt_aaaaa(input_dt, locale), dt, fixed = TRUE)
  }

  if ("b" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{b}", dt_b(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{bb}", dt_b(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{bbb}", dt_b(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{bbbb}", dt_bbbb(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{bbbbb}", dt_bbbbb(input_dt, locale), dt, fixed = TRUE)
  }

  if ("B" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{B}", dt_B(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{BB}", dt_B(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{BBB}", dt_B(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{BBBB}", dt_BBBB(input_dt, locale), dt, fixed = TRUE)
    dt <- sub("{BBBBB}", dt_BBBBB(input_dt, locale), dt, fixed = TRUE)
  }

  if ("h" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{h}", dt_h(input_dt), dt, fixed = TRUE)
    dt <- sub("{hh}", dt_hh(input_dt), dt, fixed = TRUE)
  }

  if ("H" %in% dt_lett) {
    pattern <- regmatches(dt, m = regexpr("\\{H+\\}", dt))
    # Pattern will be of shape "{y+}" and will call the appropriate function
    # For replacing the year in the final value
    # Will call dt_year$`{MM}`(input_dt) and replace the resulting value
    dt <- sub(pattern, dt_hour_full[[pattern]](input_dt), dt, fixed = TRUE)
  }

  if ("K" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{K}", dt_K(input_dt), dt, fixed = TRUE)
    dt <- sub("{KK}", dt_KK(input_dt), dt, fixed = TRUE)
  }

  if ("k" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{k}", dt_k(input_dt), dt, fixed = TRUE)
    dt <- sub("{kk}", dt_kk(input_dt), dt, fixed = TRUE)
  }

  if ("m" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{m}", dt_m(input_dt), dt, fixed = TRUE)
    dt <- sub("{mm}", dt_mm(input_dt), dt, fixed = TRUE)
  }

  if ("s" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{s}", dt_s(input_dt), dt, fixed = TRUE)
    dt <- sub("{ss}", dt_ss(input_dt), dt, fixed = TRUE)
  }

  if ("S" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{S}", dt_S(input_dt), dt, fixed = TRUE)
    dt <- sub("{SS}", dt_SS(input_dt), dt, fixed = TRUE)
    dt <- sub("{SSS}", dt_SSS(input_dt), dt, fixed = TRUE)
    dt <- sub("{SSSS}", dt_SSSS(input_dt), dt, fixed = TRUE)
    dt <- sub("{SSSSS}", dt_SSSSS(input_dt), dt, fixed = TRUE)
    dt <- sub("{SSSSSS}", dt_SSSSSS(input_dt), dt, fixed = TRUE)
    dt <- sub("{SSSSSSS}", dt_SSSSSSS(input_dt), dt, fixed = TRUE)
    dt <- sub("{SSSSSSSS}", dt_SSSSSSSS(input_dt), dt, fixed = TRUE)
    dt <- sub("{SSSSSSSSS}", dt_SSSSSSSSS(input_dt), dt, fixed = TRUE)
  }

  if ("A" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{A}", dt_A(input_dt), dt, fixed = TRUE)
    dt <- sub("{AA}", dt_AA(input_dt), dt, fixed = TRUE)
    dt <- sub("{AAA}", dt_AAA(input_dt), dt, fixed = TRUE)
    dt <- sub("{AAAA}", dt_AAAA(input_dt), dt, fixed = TRUE)
    dt <- sub("{AAAAA}", dt_AAAAA(input_dt), dt, fixed = TRUE)
    dt <- sub("{AAAAAA}", dt_AAAAAA(input_dt), dt, fixed = TRUE)
    dt <- sub("{AAAAAAA}", dt_AAAAAAA(input_dt), dt, fixed = TRUE)
    dt <- sub("{AAAAAAAA}", dt_AAAAAAAA(input_dt), dt, fixed = TRUE)
    dt <- sub("{AAAAAAAAA}", dt_AAAAAAAAA(input_dt), dt, fixed = TRUE)
  }

  if ("z" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{z}", dt_z(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{zz}", dt_z(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{zzz}", dt_z(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{zzzz}", dt_zzzz(input_dt, tz_info, locale), dt, fixed = TRUE)
  }

  if ("Z" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{Z}", dt_Z(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{ZZ}", dt_Z(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{ZZZ}", dt_Z(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{ZZZZ}", dt_ZZZZ(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{ZZZZZ}", dt_ZZZZZ(input_dt, tz_info, locale), dt, fixed = TRUE)
  }

  if ("O" %in% dt_lett) {
    dt <- sub("{O}", dt_O(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{OOOO}", dt_OOOO(input_dt, tz_info, locale), dt, fixed = TRUE)
  }

  if ("v" %in% dt_lett) {
    dt <- sub("{v}", dt_v(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{vvvv}", dt_vvvv(input_dt, tz_info, locale), dt, fixed = TRUE)
  }

  if ("V" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{V}", dt_V(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{VV}", dt_VV(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{VVV}", dt_VVV(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{VVVV}", dt_VVVV(input_dt, tz_info, locale), dt, fixed = TRUE)
  }

  if ("X" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{X}", dt_X(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{XX}", dt_XX(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{XXX}", dt_XXX(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{XXXX}", dt_XXXX(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{XXXXX}", dt_XXXXX(input_dt, tz_info, locale), dt, fixed = TRUE)
  }

  if ("x" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{x}", dt_x(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{xx}", dt_xx(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{xxx}", dt_xxx(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{xxxx}", dt_xxxx(input_dt, tz_info, locale), dt, fixed = TRUE)
    dt <- sub("{xxxxx}", dt_xxxxx(input_dt, tz_info, locale), dt, fixed = TRUE)
  }

  if ("g" %in% dt_lett) {
    # TODO extract the exact substitution to do to avoid problematic
    dt <- sub("{g}", dt_g(input_dt), dt, fixed = TRUE)
    dt <- sub("{gg}", dt_gg(input_dt), dt, fixed = TRUE)
    dt <- sub("{ggg}", dt_ggg(input_dt), dt, fixed = TRUE)
    dt <- sub("{gggg}", dt_gggg(input_dt), dt, fixed = TRUE)
    dt <- sub("{ggggg}", dt_ggggg(input_dt), dt, fixed = TRUE)
    dt <- sub("{gggggg}", dt_gggggg(input_dt), dt, fixed = TRUE)
    dt <- sub("{ggggggg}", dt_ggggggg(input_dt), dt, fixed = TRUE)
    dt <- sub("{gggggggg}", dt_gggggggg(input_dt), dt, fixed = TRUE)
    dt <- sub("{ggggggggg}", dt_ggggggggg(input_dt), dt, fixed = TRUE)
  }

  # Return modified output
  dt
}


# Era // abbreviated (G..GGG) (AD, variant: CE)
dt_G <- function(input, locale = NULL) {

  year <- as.integer(format(input, format = "%Y"))

  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$eras_abbrev
  )[[if (year < 0) "0" else "1"]]
}

# Era // wide (Anno Domini, variant: Common Era)
dt_GGGG <- function(input, locale = NULL) {

  year <- as.integer(format(input, format = "%Y"))

  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$eras_names
  )[[if (year < 0) "0" else "1"]]
}

# Era // narrow (A)
dt_GGGGG <- function(input, locale = NULL) {

  year <- as.integer(format(input, format = "%Y"))

  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$eras_narrow
  )[[if (year < 0) "0" else "1"]]
}


dt_year <- list(
  # Calendar year
  `{y}` = function(input) {
    as.character(abs(as.integer(format(input, format = "%Y"))))
  },
  # Abbreviated year (final 2 digits)
  `{yy}` = function(input) {
    format(input, format = "%y")
  },
  # Calendar year (min 3-9 digits wide)
  `{yyy}` = function(input) dt_yyy_plus(input = input, length = 3),
  `{yyyy}` = function(input) dt_yyy_plus(input = input, length = 4),
  `{yyyyy}` = function(input) dt_yyy_plus(input = input, length = 5),
  `{yyyyyy}` = function(input) dt_yyy_plus(input = input, length = 6),
  `{yyyyyyy}` = function(input) dt_yyy_plus(input = input, length = 7),
  `{yyyyyyyy}` = function(input) dt_yyy_plus(input = input, length = 8),
  `{yyyyyyyyy}` = function(input) dt_yyy_plus(input = input, length = 9)
)

# Full year with minimum zero padding (yyy -> 2, yyyy -> 3, etc.)
dt_yyy_plus <- function(input, length) {
  zero_pad_to_width(
    value = abs(as.integer(format(input, format = "%Y"))),
    width = length
  )
}

dt_year_full <- list(
  # Full year (week in year calendar)
  `{Y}` = function(input) {
    yearweek <- format_yearweek(input = input)
    unlist(strsplit(yearweek, "-W", fixed = TRUE))[[1]]
  },
  # Abbreviated year (2 digit) (week in year calendar)
  `{YY}` = function(input) {
    year_str <- dt_Y(input = input)
    substr(year_str, nchar(year_str) - 1, nchar(year_str))
  },
  # Calendar year (week in year calendar, min 3-9 digits wide)
  # appear unused?
  `{YYY}` = function(input) dt_YYY_plus(input = input, length = 3),
  `{YYYY}` = function(input) dt_YYY_plus(input = input, length = 4),
  `{YYYYY}` = function(input) dt_YYY_plus(input = input, length = 5),
  `{YYYYYY}` = function(input) dt_YYY_plus(input = input, length = 6),
  `{YYYYYYY}` = function(input) dt_YYY_plus(input = input, length = 7),
  `{YYYYYYYY}` = function(input) dt_YYY_plus(input = input, length = 8),
  `{YYYYYYYYY}` = function(input) dt_YYY_plus(input = input, length = 9)
)

# Full year with minimum zero padding (yyy -> 2, yyyy -> 3, etc.)
# (week in year calendar)
dt_YYY_plus <- function(input, length) {
  zero_pad_to_width(
    value = as.integer(dt_year_full$`{Y}`(input = input)),
    width = length
  )
}

# Extended year (numeric)
dt_u <- function(input) {
  "u"
}

# Cyclic year name (abbreviated) U..UUU
dt_U <- function(input) {
  "U"
}

# Cyclic year name (wide)
dt_UUUU <- function(input, locale = NULL) {
  "UUUU"
}

# Cyclic year name (narrow)
dt_UUUUU <- function(input, locale = NULL) {
  "UUUUU"
}

# Related Gregorian year (numeric)
dt_r_plus <- function(input, length) {
  "r_plus"
}


dt_quarter_full <- list(
  `{Q}` = function(input, ...) {
    # Quarter (formatting), numeric (1 digit-wide)
    yearquarter <- format_quarter(input = input)
    unlist(strsplit(yearquarter, "-Q", fixed = TRUE))[[2]]
  },
  # Quarter (formatting), numeric (2 digit-wide, zero padded)
  `{QQ}` = function(input, ...) {
    yearquarter <- format_quarter(input = input)

    zero_pad_to_width(
      value = as.integer(unlist(strsplit(yearquarter, "-Q", fixed = TRUE))[[2]]),
      width = 2
    )
  },
  # Quarter (formatting), abbreviated
  `{QQQ}` = function(input, locale = NULL) {
    yearquarter <- format_quarter(input = input)
    quarter <- as.integer(unlist(strsplit(yearquarter, "-Q", fixed = TRUE))[[2]])

    cldr_dates_bigd(
      locale = locale,
      element = dates_elements_bigd$quarters_format_abbrev
    )[[quarter]]
  },
  # Quarter (formatting), wide
  `{QQQQ}` = function(input, locale = NULL) {
    yearquarter <- format_quarter(input = input)
    quarter <- as.integer(unlist(strsplit(yearquarter, "-Q", fixed = TRUE))[[2]])

    cldr_dates_bigd(
      locale = locale,
      element = dates_elements_bigd$quarters_format_wide
    )[[quarter]]
  },
  # Quarter (formatting), narrow
  `{QQQQQ}` = function(input, locale = NULL) {
    yearquarter <- format_quarter(input = input)
    quarter <- as.integer(unlist(strsplit(yearquarter, "-Q", fixed = TRUE))[[2]])

    cldr_dates_bigd(
      locale = locale,
      element = dates_elements_bigd$quarters_format_narrow
    )[[quarter]]
  }
)

# Quarter (standalone), numeric (1 digit)
dt_q <- function(input) {
  dt_quarter_full$`{Q}`(input = input)
}

# Quarter (standalone), numeric (2 digit, zero padded)
dt_qq <- function(input) {
  dt_quarter_full$`{QQ}`(input = input)
}

# Quarter (standalone), abbreviated
dt_qqq <- function(input, locale = NULL) {

  yearquarter <- format_quarter(input = input)
  quarter <- as.integer(unlist(strsplit(yearquarter, "-Q", fixed = TRUE))[[2]])

  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$quarters_standalone_abbrev
  )[[quarter]]
}

# Quarter (standalone), wide
dt_qqqq <- function(input, locale = NULL) {

  yearquarter <- format_quarter(input = input)
  quarter <- as.integer(unlist(strsplit(yearquarter, "-Q", fixed = TRUE))[[2]])

  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$quarters_standalone_wide
  )[[quarter]]
}

# Quarter (standalone), narrow
dt_qqqqq <- function(input, locale = NULL) {

  yearquarter <- format_quarter(input = input)
  quarter <- as.integer(unlist(strsplit(yearquarter, "-Q", fixed = TRUE))[[2]])

  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$quarters_standalone_narrow
  )[[quarter]]
}

# List of function to transform "{MM}" to the correct output
dt_Month <- list(
  # Month (format), numeric form (1 digit) ("9")
  `{M}` = function(input, ...) {
    as.character(as.integer(format(input, format = "%m")))
  },
  # Month (format), numeric form (2 digit, zero padded) ("09")
  `{MM}` = function(input, ...) {
    format(input, format = "%m")
  },
  # Month (format), abbreviated ("Sep")
  `{MMM}` = function(input, locale = NULL) {
    cldr_dates_bigd(
      locale = locale,
      element = dates_elements_bigd$months_format_abbrev
    )[[as.integer(format(input, format = "%m"))]]
  },
  # Month (format), full ("September")
  `{MMMM}` = function(input, locale = NULL) {
    cldr_dates_bigd(
      locale = locale,
      element = dates_elements_bigd$months_format_wide
    )[[as.integer(format(input, format = "%m"))]]
  },
  # Month (format), narrow ("S")
  `{MMMMM}` = function(input, locale = NULL) {
    cldr_dates_bigd(
      locale = locale,
      element = dates_elements_bigd$months_format_narrow
    )[[as.integer(format(input, format = "%m"))]]
  }
)

# Month (standalone), numeric form (1 digit) ("9")
dt_L <- function(input) {
  dt_Month$`{M}`(input)
}

# Month (standalone), numeric form (2 digit, zero padded) ("09")
dt_LL <- function(input) {
  dt_Month$`{MM}`(input)
}

# Month (standalone), abbreviated ("Sep")
dt_LLL <- function(input, locale = NULL) {
  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$months_standalone_abbrev
  )[[as.integer(format(input, format = "%m"))]]
}

# Month (standalone), full ("September")
dt_LLLL <- function(input, locale = NULL) {
  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$months_standalone_wide
  )[[as.integer(format(input, format = "%m"))]]
}

# Month (standalone), narrow ("S")
dt_LLLLL <- function(input, locale = NULL) {
  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$months_standalone_narrow
  )[[as.integer(format(input, format = "%m"))]]
}

# Week of year, numeric, 1-2 digits
dt_w <- function(input) {
  yearweek <- format_yearweek(input = input)
  as.character(as.integer(unlist(strsplit(yearweek, "-W", fixed = TRUE))[[2]]))
}

# Week of year, numeric, 2 digits zero-padded
dt_ww <- function(input) {

  yearweek <- format_yearweek(input = input)

  zero_pad_to_width(
    value = as.integer(unlist(strsplit(yearweek, "-W", fixed = TRUE))[[2]]),
    width = 2
  )
}

# Week of month, numeric, 1 digit
dt_W <- function(input, locale) {
  as.character(get_week_in_month(input = input, locale = locale))
}


dt_day <- list(
  # Day of month, numeric, 1-2 digits
  `{d}` = function(input) {
    as.character(as.integer(format(input, format = "%d")))
  },
  # Day of month, numeric, 2 digits zero padded
  `{dd}` =  function(input) {
    zero_pad_to_width(value = as.integer(format(input, format = "%d")), width = 2)
  }
)

dt_day_full <- list(
  # Day of year, numeric, 1-3 digits
  `{D}` = function(input) {
    j_day <- as.integer(format(input, format = "%j"))
    zero_pad_to_width(value = j_day, width = 1)
  },

  # Day of year, numeric, 2-3 digits zero-padded
  `{DD}` =  function(input) {
    j_day <- as.integer(format(input, format = "%j"))
    zero_pad_to_width(value = j_day, width = 2)
  },

  # Day of year, numeric, 3 digits zero-padded
  `{DDD}` = function(input) {
    j_day <- as.integer(format(input, format = "%j"))
    zero_pad_to_width(value = j_day, width = 3)
  }
)

# Day of week in month, numeric, 1 digit ("2", as in the 2nd Wed in July")
dt_F <- function(input) {
  as.character(get_dow_n_in_month(input = input))
}

# Modified Julian day, numeric ("2451334")
dt_g_plus <- function(input, length) {
  zero_pad_to_width(
    value = get_modified_julian_day(input = input),
    width = length
  )
}

# Modified Julian day (min 1-9 digits wide)
dt_g <- function(input) dt_g_plus(input = input, length = 1)
dt_gg <- function(input) dt_g_plus(input = input, length = 2)
dt_ggg <- function(input) dt_g_plus(input = input, length = 3)
dt_gggg <- function(input) dt_g_plus(input = input, length = 4)
dt_ggggg <- function(input) dt_g_plus(input = input, length = 5)
dt_gggggg <- function(input) dt_g_plus(input = input, length = 6)
dt_ggggggg <- function(input) dt_g_plus(input = input, length = 7)
dt_gggggggg <- function(input) dt_g_plus(input = input, length = 8)
dt_ggggggggg <- function(input) dt_g_plus(input = input, length = 9)

# Day of Week Name // abbreviated (E..EEE) ("Tue")
dt_E <- function(input, locale = NULL) {

  dow <- as.integer(format(input, format = "%w")) + 1L

  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$days_standalone_abbrev
  )[[names_wkdays()[dow]]]
}

# Day of Week Name // wide ("Tuesday")
dt_EEEE <- function(input, locale = NULL) {

  dow <- as.integer(format(input, format = "%w")) + 1L

  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$days_standalone_wide
  )[[names_wkdays()[dow]]]
}

# Day of Week Name // narrow ("T")
dt_EEEEE <- function(input, locale = NULL) {

  dow <- as.integer(format(input, format = "%w")) + 1L

  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$days_standalone_narrow
  )[[names_wkdays()[dow]]]
}

# Day of Week Name // short ("Tu")
dt_EEEEEE <- function(input, locale = NULL) {

  dow <- as.integer(format(input, format = "%w")) + 1L

  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$days_standalone_short
  )[[names_wkdays()[dow]]]
}

# Local Day of Week Name/Number // 1 digit
dt_e <- function(input, locale = NULL) {

  # TODO: get starting day of the week for locale

  "e"
}

# Local Day of Week Name/Number // 2 digit, zero padded
dt_ee <- function(input, locale = NULL) {

  # TODO: get starting day of the week for locale

  "ee"
}

# Local Day of Week Name/Number // abbreviated ("Tue")
dt_eee <- function(input, locale = NULL) {

  # TODO: get starting day of the week for locale

  "eee"
}

# Local Day of Week Name/Number // wide ("Tuesday")
dt_eeee <- function(input, locale = NULL) {

  # TODO: get starting day of the week for locale

  "eeee"
}

# Local Day of Week Name/Number // narrow ("T")
dt_eeeee <- function(input, locale = NULL) {

  # TODO: get starting day of the week for locale

  "eeeee"
}

# Local Day of Week Name/Number // short ("Tu")
dt_eeeeee <- function(input, locale = NULL) {

  # TODO: get starting day of the week for locale

  "eeeeee"
}

# Standalone Local Day of Week Name/Number // 1 digit
dt_c <- function(input, locale = NULL) {

  # TODO: get starting day of the week for locale

  "c"
}

# Standalone Local Day of Week Name/Number // 2 digit, zero padded
dt_cc <- function(input, locale = NULL) {

  # TODO: get starting day of the week for locale

  "cc"
}

# Standalone Local Day of Week Name/Number // abbreviated ("Tue")
dt_ccc <- function(input, locale = NULL) {

  # TODO: get starting day of the week for locale

  "ccc"
}

# Standalone Local Day of Week Name/Number // wide ("Tuesday")
dt_cccc <- function(input, locale = NULL) {

  # TODO: get starting day of the week for locale

  "cccc"
}

# Standalone Local Day of Week Name/Number // narrow ("T")
dt_ccccc <- function(input, locale = NULL) {

  # TODO: get starting day of the week for locale

  "ccccc"
}

# Standalone Local Day of Week Name/Number // short ("Tu")
dt_cccccc <- function(input, locale = NULL) {

  # TODO: get starting day of the week for locale

  "cccccc"
}

# Period: am, pm // abbreviated (a..aaa)
dt_a <- function(input, locale = NULL) {
  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$dayperiods_format_abbrev
  )[[tolower(format(input, format = "%p"))]]
}

# Period: am, pm // wide
dt_aaaa <- function(input, locale = NULL) {
  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$dayperiods_format_wide
  )[[tolower(format(input, format = "%p"))]]
}

# Period: am, pm // narrow
dt_aaaaa <- function(input, locale = NULL) {
  cldr_dates_bigd(
    locale = locale,
    element = dates_elements_bigd$dayperiods_format_narrow
  )[[tolower(format(input, format = "%p"))]]
}

# Period: am, pm, noon, midnight // abbreviated (b..bbb)
dt_b <- function(input, locale = NULL) {

  period <- get_noon_midnight_period(input = input, locale = locale)

  if (is.na(period)) {
    return(dt_a(input = input, locale = locale))
  }

  day_periods_locale <-
    cldr_dates_bigd(
      locale = locale,
      element = dates_elements_bigd$dayperiods_format_abbrev
    )

  if (!(period %in% names(day_periods_locale))) {
    return(dt_a(input = input, locale = locale))
  }

  day_periods_locale[[period]]
}

# Period: am, pm, noon, midnight // wide
dt_bbbb <- function(input, locale = NULL) {

  period <- get_noon_midnight_period(input = input, locale = locale)

  if (is.na(period)) {
    return(dt_aaaa(input = input, locale = locale))
  }

  day_periods_locale <-
    cldr_dates_bigd(
      locale = locale,
      element = dates_elements_bigd$dayperiods_format_wide
    )

  if (!(period %in% names(day_periods_locale))) {
    return(dt_aaaa(input = input, locale = locale))
  }

  day_periods_locale[[period]]
}

# Period: am, pm, noon, midnight // narrow
dt_bbbbb <- function(input, locale = NULL) {

  period <- get_noon_midnight_period(input = input, locale = locale)

  if (is.na(period)) {
    return(dt_aaaaa(input = input, locale = locale))
  }

  day_periods_locale <-
    cldr_dates_bigd(
      locale = locale,
      element = dates_elements_bigd$dayperiods_format_narrow
    )

  if (!(period %in% names(day_periods_locale))) {
    return(dt_aaaaa(input = input, locale = locale))
  }

  day_periods_locale[[period]]
}

# Flexible day periods // abbreviated (B..BBB)
dt_B <- function(input, locale = NULL) {

  period <- get_flexible_day_period(input = input, locale = locale)

  if (is.na(period)) {
    return(dt_a(input = input, locale = locale))
  }

  day_periods_locale <-
    cldr_dates_bigd(
      locale = locale,
      element = dates_elements_bigd$dayperiods_format_abbrev
    )

  if (!(period %in% names(day_periods_locale))) {
    return(dt_a(input = input, locale = locale))
  }

  day_periods_locale[[period]]
}

# Flexible day periods // wide
dt_BBBB <- function(input, locale = NULL) {

  period <- get_flexible_day_period(input = input, locale = locale)

  if (is.na(period)) {
    return(dt_aaaa(input = input, locale = locale))
  }

  day_periods_locale <-
    cldr_dates_bigd(
      locale = locale,
      element = dates_elements_bigd$dayperiods_format_wide
    )

  if (!(period %in% names(day_periods_locale))) {
    return(dt_aaaa(input = input, locale = locale))
  }

  day_periods_locale[[period]]
}

# Flexible day periods // narrow
dt_BBBBB <- function(input, locale = NULL) {

  period <- get_flexible_day_period(input = input, locale = locale)

  if (is.na(period)) {
    return(dt_aaaaa(input = input, locale = locale))
  }

  day_periods_locale <-
    cldr_dates_bigd(
      locale = locale,
      element = dates_elements_bigd$dayperiods_format_narrow
    )

  if (!(period %in% names(day_periods_locale))) {
    return(dt_aaaaa(input = input, locale = locale))
  }

  day_periods_locale[[period]]
}

# Hour [1-12] // numeric, 1-2 digits
dt_h <- function(input) {
  zero_pad_to_width(value = as.integer(format(input, format = "%I")), width = 1)
}

# Hour [1-12] // numeric, 2 digits, zero padded
dt_hh <- function(input) {
  zero_pad_to_width(value = as.integer(format(input, format = "%I")), width = 2)
}


dt_hour_full <- list(
  # Hour [0-23] // numeric, 1-2 digits
  `{H}` = function(input) {
    zero_pad_to_width(value = as.integer(format(input, format = "%H")), width = 1)
  },
  # Hour [0-23] // numeric, 2 digits, zero padded
  `{HH}` = function(input) {
    zero_pad_to_width(value = as.integer(format(input, format = "%H")), width = 2)
  }
)


# Hour [0-11] // numeric, 1-2 digits
dt_K <- function(input) {
  zero_pad_to_width(value = as.integer(format(input, format = "%I")) - 1, width = 1)
}

# Hour [0-11] // numeric, 2 digits, zero padded
dt_KK <- function(input) {
  zero_pad_to_width(value = as.integer(format(input, format = "%I")) - 1, width = 2)
}

# Hour [1-24] // numeric, 1-2 digits
dt_k <- function(input) {
  zero_pad_to_width(value = as.integer(format(input, format = "%H")) + 1, width = 1)
}

# Hour [1-24] // numeric, 2 digits, zero padded
dt_kk <- function(input) {
  zero_pad_to_width(value = as.integer(format(input, format = "%H")) + 1, width = 2)
}

# Hour (Input Skeleton Symbol) // numeric, 1-2 digits + abbrev day period
dt_j <- function(input) {
  "j"
}

# Hour (Input Skeleton Symbol) // numeric, 2 digits, zero padded + abbrev day period
dt_jj <- function(input) {
  "jj"
}

# Hour (Input Skeleton Symbol) // numeric, 1-2 digits + wide day period
dt_jjj <- function(input) {
  "jjj"
}

# Hour (Input Skeleton Symbol) // numeric, 2 digits, zero padded + wide day period
dt_jjjj <- function(input) {
  "jjjj"
}

# Hour (Input Skeleton Symbol) // numeric, 1-2 digits + narrow day period
dt_jjjjj <- function(input) {
  "jjjjj"
}

# Hour (Input Skeleton Symbol) // numeric, 2 digits, zero padded + narrow day period
dt_jjjjjj <- function(input) {
  "jjjjjj"
}

# Hour (Input Skeleton Symbol) // numeric, 2 digits, zero padded (no AM/PM)
dt_J <- function(input) {
  "J"
}

# Hour (Input Skeleton Symbol) // numeric, 1-2 digits (no AM/PM)
dt_JJ <- function(input) {
  "JJ"
}

# Hour (Input Skeleton Symbol) // numeric, 1-2 digits + abbrev day period (no AM/PM)
dt_C <- function(input) {
  "C"
}

# Hour (Input Skeleton Symbol) // numeric, 2 digits, zero padded + abbrev day period (no AM/PM)
dt_CC <- function(input) {
  "CC"
}

# Hour (Input Skeleton Symbol) // numeric, 1-2 digits + wide day period (no AM/PM)
dt_CCC <- function(input) {
  "CCC"
}

# Hour (Input Skeleton Symbol) // numeric, 2 digits, zero padded + wide day period (no AM/PM)
dt_CCCC <- function(input) {
  "CCCC"
}

# Hour (Input Skeleton Symbol) // numeric, 1-2 digits + narrow day period (no AM/PM)
dt_CCCCC <- function(input) {
  "CCCCC"
}

# Hour (Input Skeleton Symbol) // numeric, 2 digits, zero padded + narrow day period (no AM/PM)
dt_CCCCCC <- function(input) {
  "CCCCCC"
}

# Minute // numeric, 1-2 digits
dt_m <- function(input) {
  zero_pad_to_width(value = as.integer(format(input, format = "%M")), width = 1)
}

# Minute // numeric, 2 digits, zero padded
dt_mm <- function(input) {
  zero_pad_to_width(value = as.integer(format(input, format = "%M")), width = 2)
}

# Second // numeric, 1-2 digits
dt_s <- function(input) {
  zero_pad_to_width(value = as.integer(format(input, format = "%S")), width = 1)
}

# Second // numeric, 2 digits, zero padded
dt_ss <- function(input) {
  zero_pad_to_width(value = as.integer(format(input, format = "%S")), width = 2)
}

# Fractional second
dt_S_plus <- function(input, length) {
  format_fractional_seconds(input = input, digits = length)
}

# Fractional second (min 1-9 digits wide)
dt_S <- function(input) dt_S_plus(input = input, length = 1)
dt_SS <- function(input) dt_S_plus(input = input, length = 2)
dt_SSS <- function(input) dt_S_plus(input = input, length = 3)
dt_SSSS <- function(input) dt_S_plus(input = input, length = 4)
dt_SSSSS <- function(input) dt_S_plus(input = input, length = 5)
dt_SSSSSS <- function(input) dt_S_plus(input = input, length = 6)
dt_SSSSSSS <- function(input) dt_S_plus(input = input, length = 7)
dt_SSSSSSSS <- function(input) dt_S_plus(input = input, length = 8)
dt_SSSSSSSSS <- function(input) dt_S_plus(input = input, length = 9)

# Milliseconds in day
dt_A_plus <- function(input, length) {
  zero_pad_to_width(
    value = get_milliseconds_in_day(input = input),
    width = length
  )
}

# Milliseconds in day (min 1-9 digits wide)
dt_A <- function(input) dt_A_plus(input = input, length = 1)
dt_AA <- function(input) dt_A_plus(input = input, length = 2)
dt_AAA <- function(input) dt_A_plus(input = input, length = 3)
dt_AAAA <- function(input) dt_A_plus(input = input, length = 4)
dt_AAAAA <- function(input) dt_A_plus(input = input, length = 5)
dt_AAAAAA <- function(input) dt_A_plus(input = input, length = 6)
dt_AAAAAAA <- function(input) dt_A_plus(input = input, length = 7)
dt_AAAAAAAA <- function(input) dt_A_plus(input = input, length = 8)
dt_AAAAAAAAA <- function(input) dt_A_plus(input = input, length = 9)

# TZ // short specific non-location format ("PDT") (z..zzz)
# Fallback to "O"
dt_z <- function(input, tz_info, locale = NULL) {

  tz_short_specific <- tz_info$tz_short_specific
  tz_offset <- tz_info$tz_offset

  # If the short specific non-location format tz is present, return that
  if (!is.na(tz_short_specific)) {
    return(tz_short_specific)
  }

  # If we don't have the former tz type but have an offset, use the short
  # localized GMT format
  if (is.na(tz_short_specific) && !is.na(tz_offset)) {
    out <-
      dt_O(
        input = input,
        tz_info = tz_info,
        locale = locale
      )
    return(out)
  }

  # If we don't have either, return an empty string`
  ""
}

# TZ // long specific non-location format ("Pacific Daylight Time") (Z..ZZZ)
# Fallback to "OOOO"
dt_zzzz <- function(input, tz_info, locale = NULL) {

  tz_long_specific <- tz_info$tz_long_specific

  if (is.na(tz_long_specific) || length(tz_long_specific) != 1) {
    return(dt_OOOO(input = input, tz_info = tz_info, locale = locale))
  }

  tz_long_specific
}

# TZ // short specific non-location format ("-0800") (Z..ZZZ)
# Equivalent to "xxxx"
dt_Z <- function(input, tz_info, locale = NULL) {

  tz_offset <- tz_info$tz_offset
  if (is.na(tz_offset)) tz_offset <- 0

  format_tz_offset_min_sec(
    tz_offset = tz_offset,
    use_colon = FALSE,
    optional_min = FALSE,
    use_z = FALSE
  )
}

# TZ // long localized GMT format ("GMT-8:00")
# Equivalent to "OOOO"
dt_ZZZZ <- function(input, tz_info, locale = NULL) {

  tz_offset <- tz_info$tz_offset
  if (is.na(tz_offset)) tz_offset <- 0

  format_tz_offset_min_sec(
    tz_offset = tz_offset,
    use_colon = TRUE,
    optional_min = FALSE,
    use_z = FALSE,
    hours_width = 1,
    prepend_with = "GMT"
  )
}

# TZ // ISO8601 extended format ("-08:00")
# Equivalent to "XXXXX"
dt_ZZZZZ <- function(input, tz_info, locale = NULL) {

  tz_offset <- tz_info$tz_offset
  if (is.na(tz_offset)) tz_offset <- 0

  format_tz_offset_min_sec(
    tz_offset = tz_offset,
    use_colon = TRUE,
    optional_min = FALSE,
    use_z = TRUE
  )
}

# TZ // short localized GMT format ("GMT-8")
dt_O <- function(input, tz_info, locale = NULL) {

  tz_formats_i <- tz_formats[tz_formats$locale == locale, ]

  tz_offset <- tz_info$tz_offset

  if (is.na(tz_offset)) tz_offset <- 0

  if (tz_offset == 0) {
    return(tz_formats_i[["gmt_zero_format"]])
  }

  gmt_format <- tz_formats_i[["gmt_format"]]
  hour_format_combined <- tz_formats_i[["hour_format"]]

  if (tz_offset < 0) {
    hour_format <- unlist(strsplit(hour_format_combined, split = ";", fixed = TRUE))[2]
  } else {
    hour_format <- unlist(strsplit(hour_format_combined, split = ";", fixed = TRUE))[1]
  }

  hours_val <- zero_pad_to_width(abs(trunc(tz_offset)), 1)
  minutes_val <- zero_pad_to_width(abs((tz_offset - trunc(tz_offset))) * 60, 2)

  if (minutes_val == "00") {
    hour_format <- gsub("mm", "", hour_format, fixed = TRUE)
    hour_format <- gsub(":", "", hour_format, fixed = TRUE)
  }

  hours_minutes <- gsub("(H|HH)", hours_val, hour_format)
  hours_minutes <- gsub("mm", minutes_val, hours_minutes, fixed = TRUE)

  gsub("{0}", hours_minutes, gmt_format, fixed = TRUE)
}

# TZ // long localized GMT format ("GMT-08:00")
dt_OOOO <- function(input, tz_info, locale = NULL) {

  tz_formats_i <- tz_formats[tz_formats$locale == locale, ]

  tz_offset <- tz_info$tz_offset

  if (is.na(tz_offset)) tz_offset <- 0

  if (tz_offset == 0) {
    return(tz_formats_i[["gmt_zero_format"]])
  }

  gmt_format <- tz_formats_i[["gmt_format"]]
  hour_format_combined <- tz_formats_i[["hour_format"]]

  if (tz_offset < 0) {
    hour_format <- unlist(strsplit(hour_format_combined, split = ";", fixed = TRUE))[2]
  } else {
    hour_format <- unlist(strsplit(hour_format_combined, split = ";", fixed = TRUE))[1]
  }

  if (grepl("HH", hour_format, fixed = TRUE)) {
    hours_val <- zero_pad_to_width(abs(trunc(tz_offset)), 2)
  } else {
    hours_val <- zero_pad_to_width(abs(trunc(tz_offset)), 1)
  }

  minutes_val <- zero_pad_to_width(abs((tz_offset - trunc(tz_offset))) * 60, 2)

  hours_minutes <- gsub("(H|HH)", hours_val, hour_format)
  hours_minutes <- gsub("mm", minutes_val, hours_minutes, fixed = TRUE)

  gsub("{0}", hours_minutes, gmt_format, fixed = TRUE)
}

# TZ // short generic non-location format ("PT")
#
# Where that is unavailable, falls back to the generic location
# format ("VVVV"), then the short localized GMT format as the
# final fallback.
dt_v <- function(input, tz_info, locale = NULL) {

  long_tzid <- tz_info$long_tzid

  tz_generic_non_location_short <-
    get_tz_non_location(
      long_tzid = long_tzid,
      locale = locale,
      short_long = "short",
      type = "generic"
    )

  if (is.na(tz_generic_non_location_short)) {
    return(dt_VVVV(input = input, tz_info = tz_info, locale = locale))
  }

  tz_generic_non_location_short
}

# TZ // long generic non-location format ("Pacific Time")
#
# Where that is unavailable, falls back to generic location
# format ("VVVV").
dt_vvvv <- function(input, tz_info, locale = NULL) {

  long_tzid <- tz_info$long_tzid

  tz_generic_non_location_long <-
    get_tz_non_location(
      long_tzid = long_tzid,
      locale = locale,
      short_long = "long",
      type = "generic"
    )

  if (is.na(tz_generic_non_location_long)) {
    return(dt_VVVV(input = input, tz_info = tz_info, locale = locale))
  }

  tz_generic_non_location_long
}

# TZ // short time zone ID ("uslax")
dt_V <- function(input, tz_info, locale = NULL) {

  long_tzid <- tz_info$long_tzid

  bcp_id <- get_tz_bcp_id(long_tzid = long_tzid)

  if (is.na(bcp_id)) {
    return("unk")
  }

  bcp_id
}

# TZ // long time zone ID ("America/Los_Angeles")
dt_VV <- function(input, tz_info, locale = NULL) {

  long_tzid <- tz_info$long_tzid

  # If a long time zone ID is available, return it
  if (!is.na(long_tzid)) {
    return(long_tzid)
  }

  # Otherwise return an empty string
  ""
}

# TZ // exemplar city location ("Los_Angeles")
#
# The exemplar city (location) for the time zone. Where that is
# unavailable, the localized exemplar city name for the special
# zone `Etc/Unknown` is used as the fallback (for example, "Unknown City").
dt_VVV <- function(input, tz_info, locale = NULL) {

  long_tzid <- tz_info$long_tzid

  # TODO: Validate `long_tzid`

  # create the `get_localized_examplar_city()` function
  examplar_city <-
    get_localized_exemplar_city(long_tzid = long_tzid, locale = locale)

  examplar_city
}

# TZ // generic location format ("Los Angeles Time")
#
# Where that is unavailable, falls back to the long localized
# GMT format ("OOOO"; Note: Fallback is only necessary with a
# GMT-style Time Zone ID, like Etc/GMT-830.)
# This is especially useful when presenting possible timezone
# choices for user selection, since the naming is more uniform
# than the "v" format.
dt_VVVV <- function(input, tz_info, locale = NULL) {

  long_tzid <- tz_info$long_tzid

  # TODO: Validate `long_tzid`

  # Get the localized variant of the exemplar city
  exemplar_city_localized <-
    get_localized_exemplar_city(
      long_tzid = long_tzid,
      locale = locale,
      yield_unknown = FALSE
    )

  if (is.na(exemplar_city_localized)) {
    return(dt_OOOO(input = input, tz_info = tz_info, locale = locale))
  }

  tz_generic_location_pattern <-
    tz_formats[tz_formats$locale == locale, "region_format"]

  gsub("{0}", exemplar_city_localized, tz_generic_location_pattern, fixed = TRUE)
}

# TZ // ISO 8601 basic format ("-08"), optional minutes, uses "Z"
dt_X <- function(input, tz_info, locale = NULL) {

  tz_offset <- tz_info$tz_offset
  if (is.na(tz_offset)) tz_offset <- 0

  format_tz_offset_min_sec(
    tz_offset = tz_offset,
    use_colon = FALSE,
    optional_min = TRUE,
    use_z = TRUE
  )
}

# TZ // ISO 8601 basic format ("-0800"), uses minutes, uses "Z"
dt_XX <- function(input, tz_info, locale = NULL) {

  tz_offset <- tz_info$tz_offset
  if (is.na(tz_offset)) tz_offset <- 0

  format_tz_offset_min_sec(
    tz_offset = tz_offset,
    use_colon = FALSE,
    optional_min = FALSE,
    use_z = TRUE
  )
}

# TZ // ISO 8601 extended format ("-08:00"), uses minutes, uses "Z"
dt_XXX <- function(input, tz_info, locale = NULL) {

  tz_offset <- tz_info$tz_offset
  if (is.na(tz_offset)) tz_offset <- 0

  format_tz_offset_min_sec(
    tz_offset = tz_offset,
    use_colon = TRUE,
    optional_min = FALSE,
    use_z = TRUE
  )
}

# TZ // ISO 8601 basic format ("-0800"), uses minutes, optional seconds,
# uses "Z"
dt_XXXX <- function(input, tz_info, locale = NULL) {

  tz_offset <- tz_info$tz_offset
  if (is.na(tz_offset)) tz_offset <- 0

  format_tz_offset_min_sec(
    tz_offset = tz_offset,
    use_colon = FALSE,
    optional_min = FALSE,
    use_z = TRUE
  )
}

# TZ // ISO 8601 extended format ("-08:00"), uses minutes, optional seconds,
# uses "Z"
dt_XXXXX <- function(input, tz_info, locale = NULL) {

  tz_offset <- tz_info$tz_offset
  if (is.na(tz_offset)) tz_offset <- 0

  format_tz_offset_min_sec(
    tz_offset = tz_offset,
    use_colon = TRUE,
    optional_min = FALSE,
    use_z = TRUE
  )
}

# TZ // ISO 8601 basic format ("-08"), optional minutes,
# doesn't use "Z"
dt_x <- function(input, tz_info, locale = NULL) {

  tz_offset <- tz_info$tz_offset
  if (is.na(tz_offset)) tz_offset <- 0

  format_tz_offset_min_sec(
    tz_offset = tz_offset,
    use_colon = FALSE,
    optional_min = TRUE,
    use_z = FALSE
  )
}

# TZ // ISO 8601 basic format ("-0800"), uses minutes,
# doesn't use "Z"
dt_xx <- function(input, tz_info, locale = NULL) {

  tz_offset <- tz_info$tz_offset
  if (is.na(tz_offset)) tz_offset <- 0

  format_tz_offset_min_sec(
    tz_offset = tz_offset,
    use_colon = FALSE,
    optional_min = FALSE,
    use_z = FALSE
  )
}

# TZ // ISO 8601 extended format ("-08:00"), uses minutes,
# doesn't use "Z"
dt_xxx <- function(input, tz_info, locale = NULL) {

  tz_offset <- tz_info$tz_offset
  if (is.na(tz_offset)) tz_offset <- 0

  format_tz_offset_min_sec(
    tz_offset = tz_offset,
    use_colon = TRUE,
    optional_min = FALSE,
    use_z = FALSE
  )
}

# TZ // ISO 8601 basic format ("-0800"), uses minutes, optional seconds,
# doesn't use "Z"
dt_xxxx <- function(input, tz_info, locale = NULL) {

  tz_offset <- tz_info$tz_offset
  if (is.na(tz_offset)) tz_offset <- 0

  format_tz_offset_min_sec(
    tz_offset = tz_offset,
    use_colon = FALSE,
    optional_min = FALSE,
    use_z = FALSE
  )
}

# TZ // ISO 8601 extended format ("-08:00"), uses minutes, optional seconds,
# doesn't use "Z"
dt_xxxxx <- function(input, tz_info, locale = NULL) {

  tz_offset <- tz_info$tz_offset
  if (is.na(tz_offset)) tz_offset <- 0

  format_tz_offset_min_sec(
    tz_offset = tz_offset,
    use_colon = TRUE,
    optional_min = FALSE,
    use_z = FALSE
  )
}

zero_pad_to_width <- function(value, width) {
  formatC(value, width = width, flag = "0", format = "d")
}

format_tz_offset_min_sec <- function(
    tz_offset,
    use_colon = TRUE,
    optional_min = TRUE,
    use_z = FALSE,
    hours_width = 2,
    prepend_with = NULL
) {

  if (use_z && tz_offset == 0) {
    return("Z")
  }

  hours_val <- zero_pad_to_width(abs(trunc(tz_offset)), hours_width)
  minutes_val <- zero_pad_to_width(abs((tz_offset - trunc(tz_offset))) * 60, 2)

  paste0(
    prepend_with, # can be NULL (will not print)
    ifelse(tz_offset >= 0, "+", "-"),
    hours_val,
    if (optional_min && minutes_val == "00") {
      ""
    } else if (use_colon) {
      paste0(":", minutes_val)
    } else {
      minutes_val
    }
  )
}

get_iso_week <- function(input) {
  as.integer(strftime(as.Date(input, format = "%Y-%m-%d"), format = "%V"))
}

format_yearweek <- function(input) {

  date_input <- as.Date(input, format = "%Y-%m-%d")
  day_input <- as.integer(format(date_input, format = "%d"))
  month_input <- as.integer(format(date_input, format = "%m"))

  yearweek_int <- as.integer(strftime(date_input, format = "%Y%V"))
  week_int <- get_iso_week(input = date_input)

  yearweek_int_res <-
    ifelse(
      day_input > 7 & week_int == 1,
      yearweek_int + 100,
      ifelse(
        as.integer(format(date_input, format = "%m")) == 1 & week_int > 51,
        yearweek_int - 100,
        yearweek_int
      )
    )

  yearweek_int_res_char <- as.character(yearweek_int_res)

  week_part <- substr(yearweek_int_res_char, nchar(yearweek_int_res_char) - 1, nchar(yearweek_int_res_char))
  year_part <- substr(yearweek_int_res_char, 1, nchar(yearweek_int_res_char) - 2)

  paste0(year_part, "-W", week_part)
}

format_fractional_seconds <- function(input, digits) {

  frac_s <- gsub("0.", "", as.character(as.POSIXlt(input)$sec %% 1), fixed = TRUE)

  if (frac_s == "0") {
    return(strrep("0", digits))
  }

  n_char <- nchar(frac_s)

  if (n_char > 3) {
    frac_s <- substr(frac_s, 1, 3)
  }

  if (digits < 3) {
    frac_s <- substr(frac_s, 1, digits)
  }

  if (digits > 3) {
    frac_s <- paste0(frac_s, strrep("0", digits - 3))
  }

  frac_s
}

get_milliseconds_in_day <- function(input) {

  input_lt <- as.POSIXlt(input)

  # Get the number of milliseconds elapsed in day
  as.integer(
    (
      (
        (input_lt$hour * 3600) +
          (input_lt$min * 60)
      ) * 1000
    ) +
      trunc(input_lt$sec * 1000, digits = 0)
  )
}

get_locale_territory <- function(locale) {

  locale_has_territory <- grepl("^[a-zA-Z_-]+?(-|_)?([A-Z]{2}|001|419|150)", locale)

  if (locale_has_territory) {

    territory <- gsub(".*([A-Z]{2}|001|419|150).*", "\\1", locale)
    return(territory)
  }

  default_locale_name <-
    default_locales[default_locales$base_locale == gsub("_", "-", locale, fixed = TRUE), "default_locale"]

  if (length(default_locale_name) == 0L) {

    default_locale_name <-
      default_locales[startsWith(default_locales$base_locale, locale), "default_locale"][1]
  }

  territory <- gsub(".*([A-Z]{2}|001|419|150).*", "\\1", default_locale_name)

  if (is.na(territory)) {
    territory <- "001"
  }

  territory
}

get_week_in_month <- function(input, locale) {

  date_input <- as.Date(input, format = "%Y-%m-%d")

  territory <- get_locale_territory(locale = locale)

  start_of_week_territory <- start_of_week[start_of_week$territory == territory, "day_of_week"]

  if (length(start_of_week_territory) == 0L) {
    start_of_week <- "mon"
  } else {
    start_of_week <- start_of_week_territory[[1]]
  }

  if (start_of_week == "mon") {
    week_number <- get_iso_week(input = input)
  } else {
    week_number <- as.integer(format(date_input, format = "%U"))
  }

  min_date_in_month <- as.Date(paste0(format(date_input, "%Y-%m"), "-01"))
  week_num_mininmum <- as.integer(format(min_date_in_month, format = "%U"))

  week_number - (week_num_mininmum - 1)
}

get_dow_n_in_month <- function(input) {

  date_input <- as.Date(input, format = "%Y-%m-%d")
  day_input <- as.integer(format(date_input, format = "%d"))

  length(seq(from = day_input, to = 1, by = -7))
}

zero_weekday <- function(input) {
  as.integer((as.integer(format(input, "%w")) + 6) %% 7 + 1L)
}

zero_thursday <- function(date) {
  date - zero_weekday(input = as.Date(date)) + 3L
}

week_date_as_datetime <- function(input_str) {

  year <- as.integer(substr(input_str, 1, 4))
  week <- as.integer(substr(input_str, 6, 7))

  if (nchar(input_str) == 8) {
    weekday <- as.integer(substr(input_str, 8, 8))
  } else {
    weekday <- 1L
  }

  january_fourth <- as.Date(paste(year, "01", "04", sep = "-"), tz = "UTC")
  first_thursday <- zero_thursday(date = january_fourth)
  nearest_thursday <- first_thursday + 7 * (week - 1)
  final_date <- nearest_thursday - 4 + weekday + 1
  as.POSIXct(as.character(final_date), tz = "UTC")
}

month_day_as_datetime <- function(input_str) {

  month_str <- substr(input_str, 1, 2)
  day_str <- substr(input_str, 3, 4)

  date_now <- as.character(Sys.Date())
  year_str <- substr(date_now, 1, 4)

  as.POSIXct(paste(year_str, month_str, day_str, sep = "-"), tz = "UTC")
}

get_modified_julian_day <- function(input) {

  date_input <- as.Date(input, format = "%Y-%m-%d")
  day_input <- as.integer(format(date_input, format = "%d"))
  month_input <- as.integer(format(date_input, format = "%m"))
  year_input <- as.integer(format(date_input, format = "%Y"))

  if (month_input <= 2) {
    m_x <- 1
  } else {
    m_x <- 0
  }

  m_m <- (12 * m_x) + month_input
  y_x <- year_input - m_x
  y_y <- floor(0.01 * y_x)

  floor(365.25 * y_x) +
    floor(30.6001 * (1 + m_m)) +
    day_input +
    1720995 +
    (2 - (y_y - floor(0.25 * y_y))) -
    2400001
}

get_flexible_day_period <- function(input, locale) {

  locale_in_day_periods_tbl <- locale %in% day_periods[["locale"]]

  # Verify that the supplied locale is defined within the `day_periods`
  # table (use the base locale if necessary); return NA if not found
  if (!locale_in_day_periods_tbl) {

    locale <- sub("^([a-z]*).*", "\\1", locale)

    if (!locale %in% day_periods[["locale"]]) {
      return(NA_character_)
    }
  }

  input_lt <- as.POSIXlt(input)

  time_str <-
    paste0(
      zero_pad_to_width(
        value = input_lt$hour,
        width = 2
      ),
      ":",
      zero_pad_to_width(
        value = input_lt$min,
        width = 2
      )
    )

  day_periods_locale <- day_periods[day_periods$locale == locale, ]

  if (time_str == "00:00" && "00:00" %in% day_periods_locale$at) {

    period <-
      day_periods_locale[
        !is.na(day_periods_locale$at) &
          day_periods_locale$at == "00:00",
        "period",
        drop = TRUE
      ]

  } else if (time_str == "12:00" && "12:00" %in% day_periods_locale$at) {

    period <-
      day_periods_locale[
        !is.na(day_periods_locale$at) &
          day_periods_locale$at == "12:00", "period", drop = TRUE]

  } else {

    day_periods_locale_from_to <-
      day_periods_locale[
        !is.na(day_periods_locale$from), c("period", "from", "to")
      ]

    to_next_day_idx <-
      which(day_periods_locale_from_to$to < day_periods_locale_from_to$from)

    if (length(to_next_day_idx) == 1) {

      row_1 <- row_2 <- day_periods_locale_from_to[to_next_day_idx, ]

      row_1$to <- "24:00"
      row_2$from <- "00:00"

      day_periods_locale_from_to <-
        rbind(
          day_periods_locale_from_to[-to_next_day_idx, ],
          row_1,
          row_2
        )
    }

    period <-
      day_periods_locale_from_to[
        day_periods_locale_from_to$from <= time_str &
          time_str < day_periods_locale_from_to$to, "period", drop = TRUE]
  }

  period
}

get_noon_midnight_period <- function(input, locale) {

  locale_in_day_periods_tbl <- locale %in% day_periods[["locale"]]

  # Verify that the supplied locale is defined within the `day_periods`
  # table (use the base locale if necessary); return NA if not found
  if (!locale_in_day_periods_tbl) {

    # Modify locale
    locale <- sub("^([a-z]*).*", "\\1", locale)

    if (!locale %in% day_periods[["locale"]]) {
      return(NA_character_)
    }
  }

  input_lt <- as.POSIXlt(input, tz = "UTC")

  time_str <-
    paste0(
      zero_pad_to_width(
        value = input_lt$hour,
        width = 2
      ),
      ":",
      zero_pad_to_width(
        value = input_lt$min,
        width = 2
      )
    )

  day_periods_locale <- day_periods[day_periods$locale == locale, ]

  if (time_str == "00:00" && "00:00" %in% day_periods_locale$at) {

    period <-
      day_periods_locale[
        !is.na(day_periods_locale$at) &
          day_periods_locale$at == "00:00", "period", drop = TRUE]

  } else if (time_str == "12:00" && "12:00" %in% day_periods_locale$at) {

    period <-
      day_periods_locale[
        !is.na(day_periods_locale$at) &
          day_periods_locale$at == "12:00", "period", drop = TRUE]

  } else {

    period <- NA_character_
  }

  period
}

format_quarter <- function(input) {

  date_input <- as.Date(input, format = "%Y-%m-%d")

  month_input <- as.integer(format(date_input, format = "%m"))
  year_input <- as.integer(format(date_input, format = "%Y"))

  if (month_input < 4L) {
    quarter <- 1
  } else if (month_input >= 4L && month_input < 7L) {
    quarter <- 2
  } else if (month_input >= 7L && month_input < 10L) {
    quarter <- 3
  } else if (month_input >= 10L) {
    quarter <- 4
  }

  paste0(year_input, "-Q", quarter)
}

#' Obtain a standard datetime format that works across locales
#'
#' The `standard_date_time()` function can be invoked in the `format` argument
#' of the `fdt()` function to help generate a locale-specific formatting string
#' of a certain 'type' of formatted datetime. The `type` value is a keyword that
#' represents precision and verbosity; the available keywords are `"short"` (the
#' default), `"medium"`, `"long"`, and `"full"`.
#'
#' @param type One of four standardized types for the resulting datetime that
#'   range in precision and verbosity. These are `"short"` (the default),
#'   `"medium"`, `"long"`, and `"full"`.
#'
#' @return A vector of class `date_time_pattern`.
#'
#' @section Examples:
#'
#' With an input datetime of `"2018-07-04 22:05(America/Vancouver)"`, we can
#' format the date and time in a standardized way with `standard_date_time()`
#' providing the correct formatting string. This function is invoked in the
#' `format` argument of `fdt()`:
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date_time(type = "full")
#' )
#' ```
#' ```
#' #> [1] "Wednesday, July 4, 2018 at 10:05:00 PM Pacific Daylight Time"
#' ```
#'
#' The locale can be changed and we don't have to worry about the particulars
#' of the formatting string (they are standardized across locales).
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date_time(type = "full"),
#'   locale = fdt_locales_lst$nl
#' )
#' ```
#' ```
#' #> [1] "woensdag 4 juli 2018 om 22:05:00 Pacific-zomertijd"
#' ```
#'
#' We can use different `type` values to control the output datetime string. The
#' default is `"short"`.
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date_time()
#' )
#' ```
#' ```
#' #> [1] "7/4/18, 10:05 PM"
#' ```
#'
#' After that, it's `"medium"`:
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date_time(type = "medium")
#' )
#' ```
#' ```
#' #> [1] "Jul 4, 2018, 10:05:00 PM"
#' ```
#'
#' The `"short"` and `"medium"` types don't display time zone information in the
#' output. Beginning with `"long"`, the tz is shown.
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date_time(type = "long")
#' )
#' ```
#' ```
#' #> [1] "July 4, 2018 at 10:05:00 PM PDT"
#' ```
#'
#' If you don't include time zone information in the input, the `"UTC"` time
#' zone will be assumed:
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05",
#'   format = standard_date_time(type = "full")
#' )
#' ```
#' ```
#' #> [1] "Wednesday, July 4, 2018 at 10:05:00 PM GMT+00:00"
#' ```
#'
#' @export
standard_date_time <- function(
    type = c("short", "medium", "long", "full")
) {
  type <- match.arg(type)
  x <- "sdt"
  class(x) <- c("date_time_pattern", "standard", "date_time", type)
  x
}

#' Obtain a standard date format that works across locales
#'
#' The `standard_date()` function can be invoked in the `format` argument of the
#' `fdt()` function to help generate a locale-specific formatting string of a
#' certain 'type' of formatted date. The `type` value is a keyword that
#' represents precision and verbosity; the available keywords are `"short"` (the
#' default), `"medium"`, `"long"`, and `"full"`.
#'
#' @param type One of four standardized types for the resulting date that range
#'   in precision and verbosity. These are `"short"` (the default), `"medium"`,
#'   `"long"`, and `"full"`.
#'
#' @return A vector of class `date_time_pattern`.
#'
#' @section Examples:
#'
#' With an input datetime of `"2018-07-04 22:05(America/Vancouver)"`, we can
#' format as a date in a standardized way with `standard_date()` providing the
#' correct formatting string. This function is invoked in the `format` argument
#' of `fdt()`:
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date(type = "full")
#' )
#' ```
#' ```
#' #> [1] "Wednesday, July 4, 2018"
#' ```
#'
#' The locale can be changed and we don't have to worry about the particulars
#' of the formatting string (they are standardized across locales).
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date(type = "full"),
#'   locale = fdt_locales_lst$nl
#' )
#' ```
#' ```
#' #> [1] "woensdag 4 juli 2018"
#' ```
#'
#' We can use different `type` values to control the output date string. The
#' default is `"short"`.
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date()
#' )
#' ```
#' ```
#' #> [1] "7/4/18"
#' ```
#'
#' After that, it's `"medium"`:
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date(type = "medium")
#' )
#' ```
#' ```
#' #> [1] "Jul 4, 2018"
#' ```
#'
#' Then, `"long"`:
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_date(type = "long")
#' )
#' ```
#' ```
#' #> [1] "July 4, 2018"
#' ```
#'
#' And finally up to `"full"`, which was demonstrated in the first example.
#'
#' @export
standard_date <- function(
    type = c("short", "medium", "long", "full")
) {
  type <- match.arg(type)
  x <- "sdt"
  class(x) <- c("date_time_pattern", "standard", "date", type)
  x
}

#' Obtain a standard time format that works across locales
#'
#' The `standard_time()` function can be invoked in the `format` argument of the
#' `fdt()` function to help generate a locale-specific formatting string of a
#' certain 'type' of formatted time. The `type` value is a keyword that
#' represents precision and verbosity; the available keywords are `"short"` (the
#' default), `"medium"`, `"long"`, and `"full"`.
#'
#' @param type One of four standardized types for the resulting time that range
#'   in precision and verbosity. These are `"short"` (the default), `"medium"`,
#'   `"long"`, and `"full"`.
#'
#' @return A vector of class `date_time_pattern`.
#'
#' @section Examples:
#'
#' With an input datetime of `"2018-07-04 22:05(America/Vancouver)"`, we can
#' format as a time in a standardized way with `standard_time()` providing the
#' correct formatting string. This function is invoked in the `format` argument
#' of `fdt()`:
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_time(type = "full")
#' )
#' ```
#' ```
#' #> [1] "10:05:00 PM Pacific Daylight Time"
#' ```
#'
#' The locale can be changed and we don't have to worry about the particulars
#' of the formatting string (they are standardized across locales).
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_time(type = "full"),
#'   locale = fdt_locales_lst$nl
#' )
#' ```
#' ```
#' #> [1] "22:05:00 Pacific-zomertijd"
#' ```
#'
#' We can use different `type` values to control the output date string. The
#' default is `"short"`.
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_time()
#' )
#' ```
#' ```
#' #> [1] "10:05 PM"
#' ```
#'
#' After that, it's `"medium"`:
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_time(type = "medium")
#' )
#' ```
#' ```
#' #> [1] "10:05:00 PM"
#' ```
#'
#' Then, `"long"`:
#'
#' ```r
#' fdt(
#'   input = "2018-07-04 22:05(America/Vancouver)",
#'   format = standard_time(type = "long")
#' )
#' ```
#' ```
#' #> [1] "10:05:00 PM PDT"
#' ```
#'
#' And finally up to `"full"`, which was demonstrated in the first example.
#'
#' @export
standard_time <- function(
    type = c("short", "medium", "long", "full")
) {
  type <- match.arg(type)
  x <- "sdt"
  class(x) <- c("date_time_pattern", "standard", "time", type)
  x
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/standard_date_time.R
\name{standard_date}
\alias{standard_date}
\title{Obtain a standard date format that works across locales}
\usage{
standard_date(type = c("short", "medium", "long", "full"))
}
\arguments{
\item{type}{One of four standardized types for the resulting date that range
in precision and verbosity. These are \code{"short"} (the default), \code{"medium"},
\code{"long"}, and \code{"full"}.}
}
\value{
A vector of class \code{date_time_pattern}.
}
\description{
The \code{standard_date()} function can be invoked in the \code{format} argument of the
\code{fdt()} function to help generate a locale-specific formatting string of a
certain 'type' of formatted date. The \code{type} value is a keyword that
represents precision and verbosity; the available keywords are \code{"short"} (the
default), \code{"medium"}, \code{"long"}, and \code{"full"}.
}
\section{Examples}{


With an input datetime of \code{"2018-07-04 22:05(America/Vancouver)"}, we can
format as a date in a standardized way with \code{standard_date()} providing the
correct formatting string. This function is invoked in the \code{format} argument
of \code{fdt()}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date(type = "full")
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "Wednesday, July 4, 2018"
}\if{html}{\out{</div>}}

The locale can be changed and we don't have to worry about the particulars
of the formatting string (they are standardized across locales).

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date(type = "full"),
  locale = fdt_locales_lst$nl
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "woensdag 4 juli 2018"
}\if{html}{\out{</div>}}

We can use different \code{type} values to control the output date string. The
default is \code{"short"}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date()
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "7/4/18"
}\if{html}{\out{</div>}}

After that, it's \code{"medium"}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date(type = "medium")
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "Jul 4, 2018"
}\if{html}{\out{</div>}}

Then, \code{"long"}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date(type = "long")
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "July 4, 2018"
}\if{html}{\out{</div>}}

And finally up to \code{"full"}, which was demonstrated in the first example.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/days_months.R
\name{names_months}
\alias{names_months}
\title{Get a vector of all the short month names}
\usage{
names_months()
}
\value{
A character vector of short month names.
}
\description{
The \code{names_months()} function produces a vector of all short month names
used by the \strong{bigD} package.
}
\examples{
# Let's get all the short month names with
# the `names_months()` function
names_months()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fdf_ftf.R
\docType{data}
\name{flex_t12_lst}
\alias{flex_t12_lst}
\title{A list of all 12-hour flexible time types}
\format{
An object of class \code{list} of length 12.
}
\usage{
flex_t12_lst
}
\value{
A list where each element corresponds to a classifier for a 12-hour
flexible time type.
}
\description{
The \code{flex_t12_lst} object is a list of the 12-hour flexible time types which
are widely supported. Flexible time types are classes of time formatting
which can be translated across locales. There are 12
flexible time types of the 12-hour variety in \code{flex_t12_lst}.
}
\section{Examples}{


The \code{flex_t12_lst} object can be incredibly useful when you need to get a
format for 12-hour time formatting that works across all locales. You can
avoid typing errors by using this list and every flexible time type from this
list is guaranteed to work across all supported locales. In this example,
we'll use the \code{"Ehms"} flexible time type by accessing it from the
\code{flex_t12_lst} object.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05",
  format = flex_t12_lst$Bhms,
  locale = "en"
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "10:05:00 at night"
}\if{html}{\out{</div>}}

If we wanted this in a different locale, the locale-specific \code{format} pattern
behind the flexible date identifier would ensure consistency while moving to
that locale. Let's use the \code{fdt_locales_lst} object in the same spirit to
specify the German (Austria) locale.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05",
  format = flex_t12_lst$Bhms,
  locale = fdt_locales_lst$de_AT
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "10:05:00 abends"
}\if{html}{\out{</div>}}
}

\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/standard_date_time.R
\name{standard_date_time}
\alias{standard_date_time}
\title{Obtain a standard datetime format that works across locales}
\usage{
standard_date_time(type = c("short", "medium", "long", "full"))
}
\arguments{
\item{type}{One of four standardized types for the resulting datetime that
range in precision and verbosity. These are \code{"short"} (the default),
\code{"medium"}, \code{"long"}, and \code{"full"}.}
}
\value{
A vector of class \code{date_time_pattern}.
}
\description{
The \code{standard_date_time()} function can be invoked in the \code{format} argument
of the \code{fdt()} function to help generate a locale-specific formatting string
of a certain 'type' of formatted datetime. The \code{type} value is a keyword that
represents precision and verbosity; the available keywords are \code{"short"} (the
default), \code{"medium"}, \code{"long"}, and \code{"full"}.
}
\section{Examples}{


With an input datetime of \code{"2018-07-04 22:05(America/Vancouver)"}, we can
format the date and time in a standardized way with \code{standard_date_time()}
providing the correct formatting string. This function is invoked in the
\code{format} argument of \code{fdt()}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date_time(type = "full")
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "Wednesday, July 4, 2018 at 10:05:00 PM Pacific Daylight Time"
}\if{html}{\out{</div>}}

The locale can be changed and we don't have to worry about the particulars
of the formatting string (they are standardized across locales).

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date_time(type = "full"),
  locale = fdt_locales_lst$nl
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "woensdag 4 juli 2018 om 22:05:00 Pacific-zomertijd"
}\if{html}{\out{</div>}}

We can use different \code{type} values to control the output datetime string. The
default is \code{"short"}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date_time()
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "7/4/18, 10:05 PM"
}\if{html}{\out{</div>}}

After that, it's \code{"medium"}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date_time(type = "medium")
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "Jul 4, 2018, 10:05:00 PM"
}\if{html}{\out{</div>}}

The \code{"short"} and \code{"medium"} types don't display time zone information in the
output. Beginning with \code{"long"}, the tz is shown.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date_time(type = "long")
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "July 4, 2018 at 10:05:00 PM PDT"
}\if{html}{\out{</div>}}

If you don't include time zone information in the input, the \code{"UTC"} time
zone will be assumed:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05",
  format = standard_date_time(type = "full")
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "Wednesday, July 4, 2018 at 10:05:00 PM GMT+00:00"
}\if{html}{\out{</div>}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/days_months.R
\name{first_day_of_week}
\alias{first_day_of_week}
\title{Get a named vector of all first-day-of-the-week names for different regions}
\usage{
first_day_of_week()
}
\value{
A character vector of short month names.
}
\description{
The \code{names_months()} function produces a vector of all short month names
used by the \strong{bigD} package.
}
\examples{
# Let's get a vector of regions where the
# first day of the week is Saturday
names(first_day_of_week()[first_day_of_week() == "sat"])

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fdt_locales.R
\docType{data}
\name{fdt_locales_lst}
\alias{fdt_locales_lst}
\title{A list of all supported locales}
\format{
An object of class \code{list} of length 574.
}
\usage{
fdt_locales_lst
}
\value{
A list where each element corresponds to a supported locale ID.
}
\description{
The \code{fdt_locales_lst} object is a list of all supported locales. This is
useful when used with the \code{\link[=fdt]{fdt()}} function as the list can be auto-completed
with a locale identifier and this generates valid input for the \code{locale}
argument.
}
\section{Examples}{


The \code{fdt_locales_lst} object can be incredibly useful when choosing one of
supported locales. You can avoid typing errors and every element of the list
is meant to work. In this example, we'll use the \code{"da"} locale through
use of the list.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05",
  format = "yy-MMMM-d",
  locale = fdt_locales_lst$da
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "18-juli-4"
}\if{html}{\out{</div>}}
}

\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fdf_ftf.R
\name{flex_d_vec}
\alias{flex_d_vec}
\title{Get a vector of all flexible date types}
\usage{
flex_d_vec()
}
\value{
A character vector of supported flexible date types.
}
\description{
The \code{flex_d_vec()} function produces a vector of all supported flexible date
types in the \strong{bigD} package. These types are essentially identifiers for
classes of cross-locale date formatting, so, none of these should be used
directly in the \code{format} argument of the \code{fdt()} function (use the
\link{flex_d_lst} object for that).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fdf_ftf.R
\docType{data}
\name{flex_d_lst}
\alias{flex_d_lst}
\title{A list of all flexible date types}
\format{
An object of class \code{list} of length 26.
}
\usage{
flex_d_lst
}
\value{
A list where each element corresponds to a classifier for a flexible
date type.
}
\description{
The \code{flex_d_lst} object is a list of widely supported flexible date types.
Flexible date types are classes of date formatting which can be translated
across locales. There are 26 flexible date types in
\code{flex_d_lst}.
}
\section{Examples}{


The \code{flex_d_lst} object can be incredibly useful when you need to get a
format for date formatting that works across all locales. You can avoid
typing errors by using this list and every flexible date type from this list
is guaranteed to work across all supported locales. In this example, we'll
use the \code{"yMMMEd"} flexible date type by accessing it from the \code{flex_d_lst}
object.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05",
  format = flex_d_lst$yMMMEd,
  locale = "en"
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "Wed, Jul 4, 2018"
}\if{html}{\out{</div>}}

If we wanted this in a different locale, the locale-specific \code{format} pattern
behind the flexible date identifier would ensure consistency while moving to
that locale. Let's use the \code{fdt_locales_lst} object in the same spirit to
specify the French (Canada) locale.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05",
  format = flex_d_lst$yMMMEd,
  locale = fdt_locales_lst$fr_CA
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "mer. 4 juill. 2018"
}\if{html}{\out{</div>}}
}

\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fdf_ftf.R
\name{flex_t24_vec}
\alias{flex_t24_vec}
\title{Get a vector of all 24-hour flexible time types}
\usage{
flex_t24_vec()
}
\value{
A character vector of supported 24-hour flexible time types.
}
\description{
The \code{flex_t24_vec()} function produces a vector of all supported flexible
24-hour time types in the \strong{bigD} package. These types are essentially
identifiers for classes of cross-locale time formatting, so, none of these
should be used directly in the \code{format} argument of the \code{fdt()} function (use
the \link{flex_t24_lst} object for that).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fdf_ftf.R
\docType{data}
\name{flex_t24_lst}
\alias{flex_t24_lst}
\title{A list of all 24-hour flexible time types}
\format{
An object of class \code{list} of length 8.
}
\usage{
flex_t24_lst
}
\value{
A list where each element corresponds to a classifier for a 24-hour
flexible time type.
}
\description{
The \code{flex_t24_lst} object is a list of the 24-hour flexible time types which
are widely supported. Flexible time types are classes of time formatting
which can be translated across locales. There are 8
flexible time types of the 24-hour variety in \code{flex_t24_lst}.
}
\section{Examples}{


The \code{flex_t24_lst} object can be incredibly useful when you need to get a
format for 24-hour time formatting that works across all locales. You can
avoid typing errors by using this list and every flexible time type from this
list is guaranteed to work across all supported locales. In this example,
we'll use the \code{"EHms"} flexible time type by accessing it from the
\code{flex_t24_lst} object.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05",
  format = flex_t24_lst$EHms,
  locale = "en"
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "Wed 22:05:00"
}\if{html}{\out{</div>}}

If we wanted this in a different locale, the locale-specific \code{format} pattern
behind the flexible date identifier would ensure consistency while moving to
that locale. Let's use the \code{fdt_locales_lst} object in the same spirit to
specify the German locale.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05",
  format = flex_t24_lst$EHms,
  locale = fdt_locales_lst$de
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "Mi, 22:05:00"
}\if{html}{\out{</div>}}
}

\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/days_months.R
\name{names_wkdays}
\alias{names_wkdays}
\title{Get a vector of all the short weekday names}
\usage{
names_wkdays()
}
\value{
A character vector of short weekday names.
}
\description{
The \code{names_wkdays()} function produces a vector of all short weekday names
used by the \strong{bigD} package.
}
\examples{
# Let's get all the short weekday names with
# the `names_wkdays()` function
names_wkdays()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fdt_locales.R
\name{fdt_locales_vec}
\alias{fdt_locales_vec}
\title{Get a vector of all supported locales}
\usage{
fdt_locales_vec()
}
\value{
A character vector of supported locale IDs.
}
\description{
The \code{fdt_locales_vec()} function produces a vector of all supported locale
IDs in the \strong{bigD} package.
}
\examples{
# Let's get all the `ar` locales that exist
# in the vector produced by `fdt_locales_vec()`
grep("^ar", fdt_locales_vec(), value = TRUE)

# Let's get all the locales that pertain to the
# `CH` territory in the vector produced by
# `fdt_locales_vec()`
grep("CH", fdt_locales_vec(), value = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fdt.R
\name{fdt}
\alias{fdt}
\title{Format a datetime with a formatting string}
\usage{
fdt(input, format = NULL, use_tz = NULL, locale = NULL)
}
\arguments{
\item{input}{A vector of date, time, or datetime values. Several
representations are acceptable here including strings, \code{Date} objects, or
\code{POSIXct} objects. Refer to the \emph{Valid Input Values} section for more
information.}

\item{format}{The formatting string to apply to all \code{input} values. If not
provided, the inputs will be formatted to ISO 8601 datetime strings. The
\emph{Date/Time Format Syntax} section has detailed information on how to create
a formatting string.}

\item{use_tz}{A tzid (e.g., \code{"America/New_York"}) time-zone designation for
precise formatting of related outputs. This overrides any time zone
information available in \code{character}-based input values and is applied to
all vector components.}

\item{locale}{The output locale to use for formatting the input value
according to the specified locale's rules. Example locale names include
\code{"en"} for English (United States) and \code{"es-EC"} for Spanish (Ecuador). If
a locale isn't provided the \code{"en"} locale will be used. The
\link{fdt_locales_vec} vector contains the valid locales and \link{fdt_locales_lst}
list provides an easy way to obtain a valid locale.}
}
\value{
A character vector of formatted dates, times, or datetimes.
}
\description{
With \code{fdt()}, we can format datetime values with the greatest of ease, and,
with great power. There is a lot of leniency in what types of input
date/time/datetime values can be passed in. The formatting string allows for
a huge array of possibilities when formatting. Not only that, we can set a
\code{locale} value and get the formatted values localized in the language/region
of choice. There's plenty of ways to represent time zone information, and
this goes along with the option to enrich the input values with a precise
time zone identifier (like \code{"America/Los_Angeles"}). The choices are ample
here, with the goal being a comprehensiveness and ease-of-use in date/time
formatting.
}
\section{Valid Input Values}{


The \code{input} argument of the \code{fdt()} function allows for some flexibility on
what can be passed in. This section describes the kinds of inputs that are
understandable by \code{fdt()}. A vector of strings is allowed, as are vectors of
\code{Date} or \code{POSIXct} values.

If using strings, a good option is to use those that adhere to the ISO
8601:2004 standard. For a datetime this can be of the form
\verb{YYYY-MM-DDThh:mm:ss.s<TZD>}. With this, \code{YYYY-MM-DD} corresponds to the
date, the literal \code{"T"} is optional, \code{hh:mm:ss} is the time (where seconds,
\code{ss}, is optional as are \code{.s} for fractional seconds), and \verb{<TZD>} refers to
an optional time-zone designation (more on time zones in the next paragraph).
You can provide just the date part, and this assumes midnight as an implicit
time. It's also possible to provide just the time part, and this internally
assembles a datetime that uses the current date. When formatting standalone
dates or times, you'll probably just format the explicit parts but \code{fdt()}
won't error if you format the complementary parts.

The time zone designation on string-based datetimes is completely optional.
If not provided then \code{"UTC"} is assumed. If you do want to supply time zone
information, it can be given as an offset value with the following
constructions:
\itemize{
\item \verb{<time>Z}
\item \verb{<time>(+/-)hh:mm}
\item \verb{<time>(+/-)hhmm}
\item \verb{<time>(+/-)hh}
}

The first, \verb{<time>Z}, is zone designator for the zero UTC offset; it's
equivalent to \code{"+00:00"}. The next two are formats for providing the time
offsets from UTC with hours and minutes fields. Examples are \code{"-05:00"} (New
York, standard time), \code{"+0200"} (Cairo), and \code{"+05:30"} (Mumbai). Note that
the colon is optional but leading zeros to maintain two-digit widths are
essential. The final format, \verb{<time>(+/-)hh}, omits the minutes field and as
so many offsets have \code{"00"} minutes values, this can be convenient.

We can also supply an Olson/IANA-style time zone identifier (tzid) in
parentheses within the string, or, as a value supplied to \code{use_tz} (should a
tzid apply to all date/time/datetime values in the \code{input} vector). By
extension, this would use the form: \verb{YYYY-MM-DDThh:mm:ss.s<TZD>(<tzid>)}.
Both a \verb{<TZD>} (UTC offset value) and a \verb{<tzid>} shouldn't really be used
together but if that occurs the \verb{<tzid>} overrides the UTC offset. Here are
some examples:
\itemize{
\item \code{"2018-07-04T22:05 (America/Vancouver)"} (preferable)
\item \code{"2018-07-04T22:05-0800(America/Vancouver)"} (redundant, but still okay)
}

A tzid contains much more information about the time zone than a UTC offset
value since it is tied to some geographical location and the timing of
Standard Time (STD) and Daylight Saving Time (DST) is known. In essence we
can derive UTC offset values from a tzid and also a host of other identifiers
(time zone names, their abbreviations, etc.). Here are some examples of valid
tzid values that can be used:
\itemize{
\item \code{"America/Jamaica"} (the official time in Jamaica, or, \code{"Jamaica Time"})
\item \code{"Australia/Perth"} (\code{"+08:00"} year round in Western Australia)
\item \code{"Europe/Dublin"} (IST/GMT time: \code{"+01:00"}/\code{"+00:00"})
}

The tz database (a compilation of information about the world's time zones)
consists of canonical zone names (those that are primary and preferred) and
alternative names (less preferred in modern usage, and was either discarded
or more commonly replaced by a canonical zone name). The \code{fdt()} function can
handle both types and what occurs is that non-canonical tzid values are
internally mapped onto canonical zone names. Here's a few examples:
\itemize{
\item \code{"Africa/Luanda"} (in Angola) maps to \code{"Africa/Lagos"}
\item \code{"America/Indianapolis"} maps to \code{"America/Indiana/Indianapolis"}
\item \code{"Asia/Calcutta"} maps to \code{"Asia/Kolkata"}
\item \code{"Pacific/Midway"} maps to \code{"Pacific/Pago_Pago"}
\item \code{"Egypt"} maps to \code{"Africa/Cairo"}
}

For the most part, the Olson-format tzid follows the form \code{"{region}/{city}"}
where the region is usually a continent, the city is considered an 'exemplar
city', and the exemplar city itself belongs in a country.
}

\section{Date/Time Format Syntax}{


A formatting pattern as used in \strong{bigD} consists of a string of characters,
where certain strings are replaced with date and time data that are derived
from the parsed input.

The characters used in patterns are tabulated below to show which specific
strings produce which outputs (e.g., \code{"y"} for the year). A common pattern is
characters that are used consecutively to produce variations on a date, time,
or timezone output. Say that the year in the input is 2015. If using \code{"yy"}
you'll get \code{"15"} but with \code{"yyyy"} the output becomes \code{"2015"}. There's a
whole lot of this, so the following subsections try to illustrate as best as
possible what each string will produce. All of the examples will use this
string-based datetime input unless otherwise indicated:

\code{"2018-07-04T22:05:09.2358(America/Vancouver)"}
\subsection{Year}{
\subsection{Calendar Year (little y)}{

This yields the calendar year, which is always numeric. In most cases the
length of the \code{"y"} field specifies the minimum number of digits to display,
zero-padded as necessary. More digits will be displayed if needed to show the
full year. There is an exception: \code{"yy"} gives use just the two low-order
digits of the year, zero-padded as necessary. For most use cases, \code{"y"} or
\code{"yy"} should be good enough.\tabular{ll}{
   Field Patterns \tab Output \cr
   \code{"y"} \tab \code{"2018"} \cr
   \code{"yy"} \tab \code{"18"} \cr
   \code{"yyy"} to \code{"yyyyyyyyy"} \tab \code{"2018"} to \code{"000002018"} \cr
}

}

\subsection{Year in the Week in Year Calendar (big Y)}{

This is the year in 'Week of Year' based calendars in which the year
transition occurs on a week boundary. This may differ from calendar year
\code{"y"} near a year transition. This numeric year designation is used in
conjunction with pattern character \code{"w"} in the ISO year-week calendar as
defined by ISO 8601.\tabular{ll}{
   Field Patterns \tab Output \cr
   \code{"Y"} \tab \code{"2018"} \cr
   \code{"YY"} \tab \code{"18"} \cr
   \code{"YYY"} to \code{"YYYYYYYYY"} \tab \code{"2018"} to \code{"000002018"} \cr
}

}

}

\subsection{Quarter}{
\subsection{Quarter of the Year: formatting (big Q) and standalone (little q)}{

The quarter names are identified numerically, starting at \code{1} and ending at
\code{4}. Quarter names may vary along two axes: the width and the context. The
context is either 'formatting' (taken as a default), which the form used
within a complete date format string, or, 'standalone', the form for date
elements used independently (such as in calendar headers). The standalone
form may be used in any other date format that shares the same form of the
name. Here, the formatting form for quarters of the year consists of some run
of \code{"Q"} values whereas the standalone form uses \code{"q"}.\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"Q"}/\code{"q"} \tab \code{"3"} \tab Numeric, one digit \cr
   \code{"QQ"}/\code{"qq"} \tab \code{"03"} \tab Numeric, two digits (zero padded) \cr
   \code{"QQQ"}/\code{"qqq"} \tab \code{"Q3"} \tab Abbreviated \cr
   \code{"QQQQ"}/\code{"qqqq"} \tab \code{"3rd quarter"} \tab Wide \cr
   \code{"QQQQQ"}/\code{"qqqqq"} \tab \code{"3"} \tab Narrow \cr
}

}

}

\subsection{Month}{
\subsection{Month: formatting (big M) and standalone (big L)}{

The month names are identified numerically, starting at \code{1} and ending at
\code{12}. Month names may vary along two axes: the width and the context. The
context is either 'formatting' (taken as a default), which the form used
within a complete date format string, or, 'standalone', the form for date
elements used independently (such as in calendar headers). The standalone
form may be used in any other date format that shares the same form of the
name. Here, the formatting form for months consists of some run of \code{"M"}
values whereas the standalone form uses \code{"L"}.\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"M"}/\code{"L"} \tab \code{"7"} \tab Numeric, minimum digits \cr
   \code{"MM"}/\code{"LL"} \tab \code{"07"} \tab Numeric, two digits (zero padded) \cr
   \code{"MMM"}/\code{"LLL"} \tab \code{"Jul"} \tab Abbreviated \cr
   \code{"MMMM"}/\code{"LLLL"} \tab \code{"July"} \tab Wide \cr
   \code{"MMMMM"}/\code{"LLLLL"} \tab \code{"J"} \tab Narrow \cr
}

}

}

\subsection{Week}{
\subsection{Week of Year (little w)}{

Values calculated for the week of year range from \code{1} to \code{53}. Week \code{1} for a
year is the first week that contains at least the specified minimum number of
days from that year. Weeks between week \code{1} of one year and week \code{1} of the
following year are numbered sequentially from \code{2} to \code{52} or \code{53} (if
needed).

There are two available field lengths. Both will display the week of year
value but the \code{"ww"} width will always show two digits (where weeks \code{1} to
\code{9} are zero padded).\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"w"} \tab \code{"27"} \tab Minimum digits \cr
   \code{"ww"} \tab \code{"27"} \tab Two digits (zero padded) \cr
}

}

\subsection{Week of Month (big W)}{

The week of a month can range from \code{1} to \code{5}. The first day of every month
always begins at week \code{1} and with every transition into the beginning of a
week, the week of month value is incremented by \code{1}.\tabular{ll}{
   Field Pattern \tab Output \cr
   \code{"W"} \tab \code{"1"} \cr
}

}

}

\subsection{Day}{
\subsection{Day of Month (little d)}{

The day of month value is always numeric and there are two available field
length choices in its formatting. Both will display the day of month value
but the \code{"dd"} formatting will always show two digits (where days \code{1} to \code{9}
are zero padded).\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"d"} \tab \code{"4"} \tab Minimum digits \cr
   \code{"dd"} \tab \code{"04"} \tab Two digits, zero padded \cr
}

}

\subsection{Day of Year (big D)}{

The day of year value ranges from \code{1} (January 1) to either \code{365} or \code{366}
(December 31), where the higher value of the range indicates that the year is
a leap year (29 days in February, instead of 28). The field length specifies
the minimum number of digits, with zero-padding as necessary.\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"D"} \tab \code{"185"} \tab  \cr
   \code{"DD"} \tab \code{"185"} \tab Zero padded to minimum width of 2 \cr
   \code{"DDD"} \tab \code{"185"} \tab Zero padded to minimum width of 3 \cr
}

}

\subsection{Day of Week in Month (big F)}{

The day of week in month returns a numerical value indicating the number of
times a given weekday had occurred in the month (e.g., '2nd Monday in
March'). This conveniently resolves to predicable case structure where ranges
of day of the month values return predictable day of week in month values:
\itemize{
\item days \code{1} - \code{7} -> \code{1}
\item days \code{8} - \code{14} -> \code{2}
\item days \code{15} - \code{21} -> \code{3}
\item days \code{22} - \code{28} -> \code{4}
\item days \code{29} - \code{31} -> \code{5}
}\tabular{ll}{
   Field Pattern \tab Output \cr
   \code{"F"} \tab \code{"1"} \cr
}

}

\subsection{Modified Julian Date (little g)}{

The modified version of the Julian date is obtained by subtracting
2,400,000.5 days from the Julian date (the number of days since January 1,
4713 BC). This essentially results in the number of days since midnight
November 17, 1858. There is a half day offset (unlike the Julian date, the
modified Julian date is referenced to midnight instead of noon).\tabular{ll}{
   Field Patterns \tab Output \cr
   \code{"g"} to \code{"ggggggggg"} \tab \code{"58303"} -> \code{"000058303"} \cr
}

}

}

\subsection{Weekday}{
\subsection{Day of Week Name (big E)}{

The name of the day of week is offered in four different widths.\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"E"}, \code{"EE"}, or \code{"EEE"} \tab \code{"Wed"} \tab Abbreviated \cr
   \code{"EEEE"} \tab \code{"Wednesday"} \tab Wide \cr
   \code{"EEEEE"} \tab \code{"W"} \tab Narrow \cr
   \code{"EEEEEE"} \tab \code{"We"} \tab Short \cr
}

}

}

\subsection{Periods}{
\subsection{AM/PM Period of Day (little a)}{

This denotes before noon and after noon time periods. May be upper or
lowercase depending on the locale and other options. The wide form may be
the same as the short form if the 'real' long form (e.g. 'ante meridiem') is
not customarily used. The narrow form must be unique, unlike some other
fields.\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"a"}, \code{"aa"}, or \code{"aaa"} \tab \code{"PM"} \tab Abbreviated \cr
   \code{"aaaa"} \tab \code{"PM"} \tab Wide \cr
   \code{"aaaaa"} \tab \code{"p"} \tab Narrow \cr
}

}

\subsection{AM/PM Period of Day Plus Noon and Midnight (little b)}{

Provide AM and PM as well as phrases for exactly noon and midnight. May be
upper or lowercase depending on the locale and other options. If the locale
doesn't have the notion of a unique 'noon' (i.e., 12:00), then the PM form
may be substituted. A similar behavior can occur for 'midnight' (00:00) and
the AM form. The narrow form must be unique, unlike some other fields.

(a) \code{input_midnight}: \code{"2020-05-05T00:00:00"}
(b) \code{input_noon}: \code{"2020-05-05T12:00:00"}\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"b"}, \code{"bb"}, or \code{"bbb"} \tab (a) \code{"midnight"} \tab Abbreviated \cr
    \tab (b) \code{"noon"} \tab  \cr
   \code{"bbbb"} \tab (a) \code{"midnight"} \tab Wide \cr
    \tab (b) \code{"noon"} \tab  \cr
   \code{"bbbbb"} \tab (a) \code{"mi"} \tab Narrow \cr
    \tab (b) \code{"n"} \tab  \cr
}

}

\subsection{Flexible Day Periods (big B)}{

Flexible day periods denotes things like 'in the afternoon', 'in the
evening', etc., and the flexibility comes from a locale's language and
script. Each locale has an associated rule set that specifies when the day
periods start and end for that locale.

(a) \code{input_morning}: \code{"2020-05-05T00:08:30"}
(b) \code{input_afternoon}: \code{"2020-05-05T14:00:00"}\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"B"}, \code{"BB"}, or \code{"BBB"} \tab (a) \code{"in the morning"} \tab Abbreviated \cr
    \tab (b) \code{"in the afternoon"} \tab  \cr
   \code{"BBBB"} \tab (a) \code{"in the morning"} \tab Wide \cr
    \tab (b) \code{"in the afternoon"} \tab  \cr
   \code{"BBBBB"} \tab (a) \code{"in the morning"} \tab Narrow \cr
    \tab (b) \code{"in the afternoon"} \tab  \cr
}

}

}

\subsection{Hours, Minutes, and Seconds}{
\subsection{Hour 0-23 (big H)}{

Hours from \code{0} to \code{23} are for a standard 24-hour clock cycle (midnight plus
1 minute is \code{00:01}) when using \code{"HH"} (which is the more common width that
indicates zero-padding to 2 digits).

Using \code{"2015-08-01T08:35:09"}:\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"H"} \tab \code{"8"} \tab Numeric, minimum digits \cr
   \code{"HH"} \tab \code{"08"} \tab Numeric, 2 digits (zero padded) \cr
}

}

\subsection{Hour 1-12 (little h)}{

Hours from \code{1} to \code{12} are for a standard 12-hour clock cycle (midnight plus
1 minute is \code{12:01}) when using \code{"hh"} (which is the more common width that
indicates zero-padding to 2 digits).

Using \code{"2015-08-01T08:35:09"}:\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"h"} \tab \code{"8"} \tab Numeric, minimum digits \cr
   \code{"hh"} \tab \code{"08"} \tab Numeric, 2 digits (zero padded) \cr
}

}

\subsection{Hour 1-24 (little k)}{

Using hours from \code{1} to \code{24} is a less common way to express a 24-hour clock
cycle (midnight plus 1 minute is \code{24:01}) when using \code{"kk"} (which is the
more common width that indicates zero-padding to 2 digits).

Using \code{"2015-08-01T08:35:09"}:\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"k"} \tab \code{"9"} \tab Numeric, minimum digits \cr
   \code{"kk"} \tab \code{"09"} \tab Numeric, 2 digits (zero padded) \cr
}

}

\subsection{Hour 0-11 (big K)}{

Using hours from \code{0} to \code{11} is a less common way to express a 12-hour clock
cycle (midnight plus 1 minute is \code{00:01}) when using \code{"KK"} (which is the
more common width that indicates zero-padding to 2 digits).

Using \code{"2015-08-01T08:35:09"}:\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"K"} \tab \code{"7"} \tab Numeric, minimum digits \cr
   \code{"KK"} \tab \code{"07"} \tab Numeric, 2 digits (zero padded) \cr
}

}

\subsection{Minute (little m)}{

The minute of the hour which can be any number from \code{0} to \code{59}. Use \code{"m"} to
show the minimum number of digits, or \code{"mm"} to always show two digits
(zero-padding, if necessary).\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"m"} \tab \code{"5"} \tab Numeric, minimum digits \cr
   \code{"mm"} \tab \code{"06"} \tab Numeric, 2 digits (zero padded) \cr
}

}

\subsection{Seconds (little s)}{

The second of the minute which can be any number from \code{0} to \code{59}. Use \code{"s"}
to show the minimum number of digits, or \code{"ss"} to always show two digits
(zero-padding, if necessary).\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"s"} \tab \code{"9"} \tab Numeric, minimum digits \cr
   \code{"ss"} \tab \code{"09"} \tab Numeric, 2 digits (zero padded) \cr
}

}

\subsection{Fractional Second (big S)}{

The fractional second truncates (like other time fields) to the width
requested (i.e., count of letters). So using pattern \code{"SSSS"} will display
four digits past the decimal (which, incidentally, needs to be added manually
to the pattern).\tabular{ll}{
   Field Patterns \tab Output \cr
   \code{"S"} to \code{"SSSSSSSSS"} \tab \code{"2"} -> \code{"235000000"} \cr
}

}

\subsection{Milliseconds Elapsed in Day (big A)}{

There are 86,400,000 milliseconds in a day and the \code{"A"} pattern will provide
the whole number. The width can go up to nine digits with \code{"AAAAAAAAA"} and
these higher field widths will result in zero padding if necessary.

Using \code{"2011-07-27T00:07:19.7223"}:\tabular{ll}{
   Field Patterns \tab Output \cr
   \code{"A"} to \code{"AAAAAAAAA"} \tab \code{"439722"} -> \code{"000439722"} \cr
}

}

}

\subsection{Era}{
\subsection{The Era Designator (big G)}{

This provides the era name for the given date. The Gregorian calendar has two
eras: AD and BC. In the AD year numbering system, AD 1 is immediately
preceded by 1 BC, with nothing in between them (there was no year zero).\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"G"}, \code{"GG"}, or \code{"GGG"} \tab \code{"AD"} \tab Abbreviated \cr
   \code{"GGGG"} \tab \code{"Anno Domini"} \tab Wide \cr
   \code{"GGGGG"} \tab \code{"A"} \tab Narrow \cr
}

}

}

\subsection{Time Zones}{
\subsection{TZ // Short and Long Specific non-Location Format (little z)}{

The short and long specific non-location formats for time zones are suggested
for displaying a time with a user friendly time zone name. Where the short
specific format is unavailable, it will fall back to the short localized GMT
format (\code{"O"}). Where the long specific format is unavailable, it will fall
back to the long localized GMT format (\code{"OOOO"}).\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"z"}, \code{"zz"}, or \code{"zzz"} \tab \code{"PDT"} \tab Short Specific \cr
   \code{"zzzz"} \tab \code{"Pacific Daylight Time"} \tab Long Specific \cr
}

}

\subsection{TZ // Common UTC Offset Formats (big Z)}{

The ISO8601 basic format with hours, minutes and optional seconds fields is
represented by \code{"Z"}, \code{"ZZ"}, or \code{"ZZZ"}. The format is equivalent to RFC 822
zone format (when the optional seconds field is absent). This is equivalent
to the \code{"xxxx"} specifier. The field pattern \code{"ZZZZ"} represents the long
localized GMT format. This is equivalent to the \code{"OOOO"} specifier. Finally,
\code{"ZZZZZ"} pattern yields the ISO8601 extended format with hours, minutes and
optional seconds fields. The ISO8601 UTC indicator \code{Z} is used when local
time offset is \code{0}. This is equivalent to the \code{"XXXXX"} specifier.\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"Z"}, \code{"ZZ"}, or \code{"ZZZ"} \tab \code{"-0700"} \tab ISO 8601 basic format \cr
   \code{"ZZZZ"} \tab \code{"GMT-7:00"} \tab Long localized GMT format \cr
   \code{"ZZZZZ"} \tab \code{"-07:00"} \tab ISO 8601 extended format \cr
}

}

\subsection{TZ // Short and Long Localized GMT Formats (big O)}{

The localized GMT formats come in two widths \code{"O"} (which removes the minutes
field if it's \code{0}) and \code{"OOOO"} (which always contains the minutes field).
The use of the \code{GMT} indicator changes according to the locale.\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"O"} \tab \code{"GMT-7"} \tab Short localized GMT format \cr
   \code{"OOOO"} \tab \code{"GMT-07:00"} \tab Long localized GMT format \cr
}

}

\subsection{TZ // Short and Long Generic non-Location Formats (little v)}{

The generic non-location formats are useful for displaying a recurring wall
time (e.g., events, meetings) or anywhere people do not want to be overly
specific. Where either of these is unavailable, there is a fallback to the
generic location format (\code{"VVVV"}), then the short localized GMT format as
the final fallback.\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"v"} \tab \code{"PT"} \tab Short generic non-location format \cr
   \code{"vvvv"} \tab \code{"Pacific Time"} \tab Long generic non-location format \cr
}

}

\subsection{TZ // Short Time Zone IDs and Exemplar City Formats (big V)}{

These formats provide variations of the time zone ID and often include the
exemplar city. The widest of these formats, \code{"VVVV"}, is useful for
populating a choice list for time zones, because it supports 1-to-1 name/zone
ID mapping and is more uniform than other text formats.\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"V"} \tab \code{"cavan"} \tab Short time zone ID \cr
   \code{"VV"} \tab \code{"America/Vancouver"} \tab Long time zone ID \cr
   \code{"VVV"} \tab \code{"Vancouver"} \tab The tz exemplar city \cr
   \code{"VVVV"} \tab \code{"Vancouver Time"} \tab Generic location format \cr
}

}

\subsection{TZ // ISO 8601 Formats with Z for +0000 (big X)}{

The \code{"X"}-\code{"XXX"} field patterns represent valid ISO 8601 patterns for time
zone offsets in datetimes. The final two widths, \code{"XXXX"} and \code{"XXXXX"} allow
for optional seconds fields. The seconds field is \emph{not} supported by the ISO
8601 specification. For all of these, the ISO 8601 UTC indicator \code{Z} is used
when the local time offset is \code{0}.\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"X"} \tab \code{"-07"} \tab ISO 8601 basic format (h, optional m) \cr
   \code{"XX"} \tab \code{"-0700"} \tab ISO 8601 basic format (h & m) \cr
   \code{"XXX"} \tab \code{"-07:00"} \tab ISO 8601 extended format (h & m) \cr
   \code{"XXXX"} \tab \code{"-0700"} \tab ISO 8601 basic format (h & m, optional s) \cr
   \code{"XXXXX"} \tab \code{"-07:00"} \tab ISO 8601 extended format (h & m, optional s) \cr
}

}

\subsection{TZ // ISO 8601 Formats (no use of Z for +0000) (little x)}{

The \code{"x"}-\code{"xxxxx"} field patterns represent valid ISO 8601 patterns for time
zone offsets in datetimes. They are similar to the \code{"X"}-\code{"XXXXX"} field
patterns except that the ISO 8601 UTC indicator \code{Z} \emph{will not} be used when
the local time offset is \code{0}.\tabular{lll}{
   Field Patterns \tab Output \tab Notes \cr
   \code{"x"} \tab \code{"-07"} \tab ISO 8601 basic format (h, optional m) \cr
   \code{"xx"} \tab \code{"-0700"} \tab ISO 8601 basic format (h & m) \cr
   \code{"xxx"} \tab \code{"-07:00"} \tab ISO 8601 extended format (h & m) \cr
   \code{"xxxx"} \tab \code{"-0700"} \tab ISO 8601 basic format (h & m, optional s) \cr
   \code{"xxxxx"} \tab \code{"-07:00"} \tab ISO 8601 extended format (h & m, optional s) \cr
}

}

}
}

\section{Examples}{

\subsection{Basics with \code{input} datetimes, formatting strings, and localization}{

With an input datetime of \code{"2018-07-04 22:05"} supplied as a string, we can
format to get just a date with the full year first, the month abbreviation
second, and the day of the month last (separated by hyphens):

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05",
  format = "y-MMM-dd"
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "2018-Jul-04"
}\if{html}{\out{</div>}}

There are sometimes many options for each time part. Instead of using
\code{"y-MMM-dd"}, let's try a variation on that with \code{"yy-MMMM-d"}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05",
  format = "yy-MMMM-d"
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "18-July-4"
}\if{html}{\out{</div>}}

The output is localizable and so elements will be translated when supplying
the appropriate locale code. Let's use \code{locale = es} to get the month written
in Spanish:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05",
  format = "yy-MMMM-d",
  locale = "es"
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "18-julio-4"
}\if{html}{\out{</div>}}

\code{POSIXct} or \code{POSIXlt} datetimes can serve as an \code{input} to \code{fdt()}. Let's
create a single datetime value where the timezone is set as \code{"Asia/Tokyo"}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = lubridate::ymd_hms("2020-03-15 19:09:12", tz = "Asia/Tokyo"),
  format = "EEEE, MMMM d, y 'at' h:mm:ss B (VVVV)"
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "Sunday, March 15, 2020 at 7:09:12 in the evening (Tokyo Time)"
}\if{html}{\out{</div>}}

If you're going minimal, it's possible to supply an input datetime string
without a \code{format} directive. What this gives us is formatted datetime
output that conforms to ISO 8601. Note that the implied time zone is UTC.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(input = "2018-07-04 22:05")
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "2018-07-04T22:05:00Z"
}\if{html}{\out{</div>}}
}

\subsection{Using locales stored in the \link{fdt_locales_lst} list}{

The \link{fdt_locales_lst} object is provided in \strong{bigD} to make it easier to
choose one of supported locales. You can avoid typing errors and every
element of the list is meant to work. For example, we can use the \code{"it"}
locale by accessing it from \link{fdt_locales_lst} (autocomplete makes this
relatively simple).

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05",
  format = "yy-MMMM-d",
  locale = fdt_locales_lst$it
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "18-luglio-4"
}\if{html}{\out{</div>}}
}

\subsection{Omission of date or time in \code{input}}{

You don't have to supply a full datetime to \code{input}. Just supplying the date
portion implies midnight (and is just fine if you're only going to present
the date anyway).

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(input = "2018-07-04")
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "2018-07-04T00:00:00Z"
}\if{html}{\out{</div>}}

If you omit the date and just supply a time, \code{fdt()} will correctly parse
this. The current date on the user system will be used because we need to
create some sort of datetime value internally. Again, this is alright if
you just intend to present a formatted time value.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(input = "22:05")
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "2022-08-16T22:05:00Z"
}\if{html}{\out{</div>}}

To see all of the supported locales, we can look at the vector supplied by
the \code{\link[=fdt_locales_vec]{fdt_locales_vec()}} function.
}

\subsection{Using standardized forms with the \verb{standard_*()} helper functions}{

With an input datetime of \code{"2018-07-04 22:05(America/Vancouver)"}, we can
format the date and time in a standardized way with \code{standard_date_time()}
providing the correct formatting string. This function is invoked in the
\code{format} argument of \code{fdt()}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date_time(type = "full")
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "Wednesday, July 4, 2018 at 10:05:00 PM Pacific Daylight Time"
}\if{html}{\out{</div>}}

The locale can be changed and we don't have to worry about the particulars
of the formatting string (they are standardized across locales).

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date_time(type = "full"),
  locale = fdt_locales_lst$nl
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "woensdag 4 juli 2018 om 22:05:00 Pacific-zomertijd"
}\if{html}{\out{</div>}}

We can use different \code{type} values to control the output datetime string. The
default is \code{"short"}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date_time()
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "7/4/18, 10:05 PM"
}\if{html}{\out{</div>}}

After that, it's \code{"medium"}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date_time(type = "medium")
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "Jul 4, 2018, 10:05:00 PM"
}\if{html}{\out{</div>}}

The \code{"short"} and \code{"medium"} types don't display time zone information in the
output. Beginning with \code{"long"}, the tz is shown.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05(America/Vancouver)",
  format = standard_date_time(type = "long")
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "July 4, 2018 at 10:05:00 PM PDT"
}\if{html}{\out{</div>}}

If you don't include time zone information in the input, the \code{"UTC"} time
zone will be assumed:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2018-07-04 22:05",
  format = standard_date_time(type = "full")
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "Wednesday, July 4, 2018 at 10:05:00 PM GMT+00:00"
}\if{html}{\out{</div>}}
}

\subsection{Using flexible date and time (12- and 24-hour) formatting}{

The \strong{bigD} package supplies a set of lists to allow for flexible date and
time formatting (\link{flex_d_lst}, \link{flex_t24_lst}, and \link{flex_t12_lst}). These
are useful when you need a particular format that works across all locales.
Here's an example that uses the \code{"yMMMEd"} flexible date type by accessing it
from the \code{flex_d_lst} object, yielding a formatted date.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2021-01-09 16:32(America/Toronto)",
  format = flex_d_lst$yMMMEd,
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "Sat, Jan 9, 2021"
}\if{html}{\out{</div>}}

If we wanted this in a different locale, the locale-specific \code{format} pattern
behind the flexible date identifier would ensure consistency while moving to
that locale.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2021-01-09 16:32(America/Toronto)",
  format = flex_d_lst$yMMMEd,
  locale = "fr_CA"
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "sam. 9 janv. 2021"
}\if{html}{\out{</div>}}

Formatting as a 12-hour time with the \link{flex_t12_lst} list and using the
\code{"hms"} flexible type:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2021-01-09 16:32(America/Toronto)",
  format = flex_t12_lst$hms
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "4:32:00 PM"
}\if{html}{\out{</div>}}

The 24-hour variant, \link{flex_t24_lst}, has a similar \code{"Hms"} flexible type that
will give us a 24-hour version of the same clock time:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2021-01-09 16:32(America/Toronto)",
  format = flex_t24_lst$Hms
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> [1] "16:32:00"
}\if{html}{\out{</div>}}

A flexible date and time can be used together by enveloping the two in a
list (\strong{bigD} will handle putting the date and time together in a sensible
manner).

\if{html}{\out{<div class="sourceCode r">}}\preformatted{fdt(
  input = "2021-01-09 16:32(America/Toronto)",
  format = list(flex_d_lst$yMMMEd, flex_t24_lst$Hmv)
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> "Sat, Jan 9, 2021, 16:32 ET"
}\if{html}{\out{</div>}}
}
}
